////////////////////////////////////////////////////////////////////////////////////
// PROYECTO FIN DE CARRERA															  											 	//
// TITULO: Disenyo de nuevas arquitecturas y algoritmos de gesti�n de recursos en	//
//		  	 redes de acceso FiWi  												  												//
// AUTOR: Ana Emma Lopez Mato																											//
// TUTOR: Noemi Merayo Alvarez																										//
// INGENIERIA TECNICA DE TELECOMUNICACIONES, SISTEMAS DE TELECOMUNICACION					//
// UNIVERSIDAD DE VALLADOLID													      											//
////////////////////////////////////////////////////////////////////////////////////

#include "BS_mac.h"
#include "BS_table.h"
#include "BS_queue.h"
#include "User_Interface.h"
#include "DATA_m.h"
#include <string.h>
/*
 * MODULO BS_mac:
 * Elabora la tabla de encaminamiento para todos los usuarios wireless, estableciendo la
 * ruta optima en funcion de un criterio elegido en el fichero omnetpp.ini, de entre los
 * tres posibles:
 * 		0->Camino de menor longitud
 * 		1->Camino de menor retardo
 * 		2->Camino de menos saltos
 * Asimismo, gestiona la multiplexacion en tiempo de la capacidad de cada enlace de radio
 * saliente del modulo BS (cada enlace correspondiente a un submodulo BS_queue).
 */
Define_Module(BS_mac);

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION INITIALIZE()--> ESTA FUNCION SE INVOCA DESPUES DE QUE OMNET++ HA PUESTO EN MARCHA LA RED, EN LA CUAL SE LEEN LOS	//
//						  PARIMETROS DEL MODULO Y SE INICIALIZAN TODAS DAS LAS VARIABLES DECLARADAS PARA ESTE MODULO SIMPLE, SE 			//
//						  ASIGNAN ESTRUCTURAS DE DATOS DINIMICOS Y SE ASIGNAN E INICIALIZAN LOS AUTOMENSAJES SI SON NECESARIOS	  		//
//						  PARA EL FUNCIONAMIENTO DE ESTE MODULO.			   																															//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void BS_mac::initialize()
{
	// Se incluyen en la topologia los nodos que contienen el parametro 'aux' (Users, BSs y ONUs):
	top.extractByParameter("aux");
	// NOTA: La clase cTopology es muy util para el manejo sencillo de los modulos compuestos (BS, ONU, User)
	//			 Cuando se crea un objeto de esta clase, se esta definiendo una representacion abstracta de la red
	// 			 muy util para funciones de encaminamiento.
	//			 Los objetos cTopoloogy son grafos formados por nodos y enlaces. Los nodos se incluyen con la linea
	//			 'extractByParameter("")' y los enlaces son las conexiones entre ellos, que estan definidas en l .ned
	//			 como canales.
	//			 En este caso, para que el grafo incluya a todas las BSs, las ONUs y los Users, se ha definido en sus
	//			 respectivos ficheros de configuracion *.ned un parametro entero auxiliar llamado "aux", y se ha dicho
	//			 al grafo que incluya todos los nodos que tienen ese parametro.
	// Se extrae el nodo correspondiente al modulo compuesto que contiene este submodulo:
	top_node = top.getNodeFor((cModule*)this->getParentModule());
	// Si el enlace num.1 conecta con una ONU, es una pasarela:
	// NOTA: Tal como se han definido las conexiones en los ficheros *.ned de configuracion de red,
	//			 el enlace num.0 siempre es el que conecta BS con User (canal 'wifi'); y en el caso de
	//		   que una BS sea pasarela, los enlaces num.1 a num.<numcolas> corresponden a conexiones
	//			 con la ONU pasarela (canal 'pasarela'). Los dem�s enlaces conectan las BSs entre si (canal 'hops').
	if(strcmp(top_node->getLinkOut(1)->getRemoteNode()->getModule()->getName(),"onu")==0){	esPasarela=1;	}
	else{	esPasarela=0;	}

	// INICIALIZACION DE PARAMETROS //
	int indice_esta_bs = (int)this->getParentModule()->par("indice"); 	// Indice de este modulo BS
	//int indice_esta_bs = (int)this->getParentModule()->getIndex(); /////// CAMBIO_RED ///////

	prioridades = (int)this->getParentModule()->par("numcolas"); // Numero de clases de servicio de la parte inalambrica de la red
	ind_primero = 0; // Indice del primer submodulo BS_queue que esta activo (uso = usado = 1).
	int estaciones = (int) this->getParentModule()->par("numBS"); // Numero de estaciones base de la red
	int num_pasarelas = (int) this->getParentModule()->par("numGW"); // Numero de BSs pasarela
	int usuarios = (int) this->getParentModule()->par("numUser"); // Corresponde a las filas de la tabla rutas_optimas[][]
	int criterio = (int) this->getParentModule()->par("criterio"); // Criterio seguido para encaminar (algoritmos SPRA, DARA o MHRA)

	extraction_method = (int) this->getParentModule()->par("user_extractionmethod_StrictPQ0_Centralized1"); // Metodo de extraccion de paquetes de las colas BS_queue[]
	media_saltos = 0; // Numero medio de saltos de las rutas elegidas por el algoritmo de encaminamiento implementado
	nodos_actualizados = 0; // Numero de BSs que han actualizado los pesos de sus enlaces, tras haber recibido todos los LSAs que esperaban, en cada periodo entre envios de LSAs en el algoritmo DARA (esta variable se actualiza desde el modulo BS_rx_tx).
	actualizaciones = 0;
	txrate = (double) this->getParentModule()->par("tasabinaria"); // Tasa de transmision wireless (bps)
	rxrate = (double) this->getParentModule()->par("tasabinaria"); // Tasa de recepcion wireless (bps)
	// NOTA: La definicion de dos parametros con el mismo valor (txrate=rxrate="tasabinaria") tiene su explicacion en que las BSs pasarela tienen diferente txrate,
	//			 pues solo transmiten trafico en sentido ascendente, hacia las ONUs pasarela, mediante un enlace optico, luego su tasa de transmision a las ONUs es la de la red EPON y no la del estandar WiFi.
	//			 Ahora bien, los paquetes que reciben las BSs pasarela de las otras BSs es a la tasa de transmision definida para el estandar wireless, luego por eso en este caso se distinguen dos variables distintas.
	enlaces_intermedios = (int) top_node->getNumOutLinks() - 1; // Numero de conexiones entre esta BS y otras BSs de la parte inalambrica de la red.
	if(esPasarela == 1) // Si es una BS pasarela...
	{
		double opt_txrate = (double) this->getParentModule()->par("txrate");
		txrate = opt_txrate/10; // ...la tasa de transmision es la misma que la de las ONUs...
		enlaces_intermedios = enlaces_intermedios - prioridades; // ...y al numero de conexiones entre BSs hay que descontarle las conexiones con la ONU pasarela ("numqueue").
	}

	C_i = (double) rxrate/enlaces_intermedios; // Capacidad de enlace = Capacidad del canal entre numero de enlaces con otras BSs.

	// CALCULO DEL PESO INICIAL DE CADA ENLACE INALAMBRICO:
	double init_weight = 0.0;
	int long_pckt = (int) this->getParentModule()->par("longitud_media_paquete");
	int bits_pckt = 8*long_pckt; // Promedio de bits en cada paquete de las fuentes inalambricas
															 // (parametro inverso a mhu, la tasa de servicio).
	int primer_indice; // Para cada BS de la topologia, indice del primer enlace cuyo peso se actualiza.
	BS_mac *aux_control; // Variable para acceder remotamente a la capa MAC de cada BS y sus parametros.
	if(indice_esta_bs == estaciones-1)
	{
		// Se actualiza el peso de cada enlace con su valor inicial (en el ".ned" se pone '0.0' por defecto):
		for(int i=0; i<(int)top.getNumNodes(); i++)
		{ // Para cada nodo de la topologia...
			if(strpbrk(top.getNode(i)->getModule()->getFullName(),"b")!=NULL) // Si es una BS...
			{
				primer_indice = 1; // El primer enlace es siempre con un modulo User, no interesa aqui.
				if(strcmp(top.getNode(i)->getLinkOut(1)->getRemoteNode()->getModule()->getName(),"onu")==0)
				{ // Si es una BS-pasarela, los 'prioridades' enlaces posteriores al primero tampoco interesan en esta
					// parte, porque conectan con la correspondiente ONU pasarela.
					primer_indice = primer_indice + prioridades;
				}
				for(int a=primer_indice; a<(int)top.getNode(i)->getNumOutLinks(); a++)
				{
					aux_control = check_and_cast<BS_mac *>(top.getNode(i)->getModule()->getSubmodule("bs_mac"));
					switch(criterio) // Segun el criterio de encaminamiento elegido en 'omnetpp.ini'...
					{
					case 0: // SPRA: Shortest Path Routing Algorithm
						init_weight = (double) 1/aux_control->C_i; // peso = 1/Ci
						// Ci = Capacidad efectiva de cada enlace = Capacidad de radio/Numero de enlaces de salida de cada nodo.
						break;
					case 1: // DARA: Delay Aware Routing Algorithm
						//aux_control = check_and_cast<BS_mac *>(top.getNode(i)->getLinkOut(a)->getRemoteNode()->getModule()->getSubmodule("bs_mac"));
						init_weight = (double) (3*bits_pckt)/(2*aux_control->C_i); // peso = (3/2)*(1/uCi)
						break;
					case 2: // MHRA: Minimum Hop Routing Algorithm
						init_weight = (double) 1; // peso = 1
						break;
					}
					(check_and_cast<cDelayChannel*>(top.getNode(i)->getLinkOut(a)->getLocalGate()->getChannel()))->par(3)
							= (double)init_weight;
				}
			}
		}
	}
	// DEFINIMOS EL PROCESO PARA PODER ENTRAR DESDE ESTE MODULO AL MODULO 'BS_table':
	Tabla = check_and_cast<BS_table *>(getParentModule()->getSubmodule("bs_table")); // ENTRAMOS Y CHEQUEAMOS EL MODULO BS_TABLE

	filas = Tabla->filas; // Filas de las tablas del modulo BS_Table
	columnas = Tabla->columnas; // Columnas de las tablas del modulo BS_Table

	// Vector con los nombres de todas las BS (para acceder mas facilmente a ellas dentro de bucles):
	nombres_bs.resize(estaciones);
	nombres_bs[0] = "bs0";
	nombres_bs[1] = "bs1";
	nombres_bs[2] = "bs2";
	nombres_bs[3] = "bs3";
	nombres_bs[4] = "bs4";
	nombres_bs[5] = "bs5";
	nombres_bs[6] = "bs6";
	nombres_bs[7] = "bs7";
	nombres_bs[8] = "bs8";
	nombres_bs[9] = "bs9";
	/*
	nombres_bs[10] = "bs10";
	nombres_bs[11] = "bs11";
	nombres_bs[12] = "bs12";
	nombres_bs[13] = "bs13";
	nombres_bs[14] = "bs14";
	nombres_bs[15] = "bs15";
	nombres_bs[16] = "bs16";
	nombres_bs[17] = "bs17";
	nombres_bs[18] = "bs18";
	nombres_bs[19] = "bs19";
	nombres_bs[20] = "bs20";
	nombres_bs[21] = "bs21";
	nombres_bs[22] = "bs22";
	nombres_bs[23] = "bs23";
	nombres_bs[24] = "bs24";
	*/
	// Vector con los indices de las "numGW" BSs pasarela:
	indices_pasarelas.resize(num_pasarelas);
	/*
	indices_pasarelas[0] = 7;
	indices_pasarelas[1] = 9;
	indices_pasarelas[2] = 13;
	indices_pasarelas[3] = 19;
	indices_pasarelas[4] = 24;
	*/
	indices_pasarelas[0] = 5;
	indices_pasarelas[1] = 7;
	indices_pasarelas[2] = 9;

	for(int i=0; i<num_pasarelas; i++)
	{
		Tabla->rutas_optimas_pasarelas[i].resize(usuarios);
		for(int n=0;n<usuarios;n++)
		{
			Tabla->rutas_optimas_pasarelas[i][n].resize(columnas," ");
		}
		////ev << " Dimensiones tabla K-rutas optimas[" << i << "]: " << Tabla->rutas_optimas_pasarelas[i].size() << "x" << Tabla->rutas_optimas_pasarelas[i][0].size() << endl;
	}

	// Invocacion del metodo 'routing()' para cada BS, una vez que todas han sido inicializadas (desde la ultima BS de la subred):
	if(indice_esta_bs == estaciones-1)
	{
		BS_mac *route;
		for(int i=0; i<estaciones; i++)
		{
			route = check_and_cast<BS_mac *>(this->getParentModule()->getParentModule()->getSubmodule(nombres_bs[i])->getSubmodule("bs_mac"));
			//route = check_and_cast<BS_mac *>(this->getParentModule()->getParentModule()->getSubmodule("bs",i)->getSubmodule("bs_mac")); /////// CAMBIO_RED ///////
			route->routing();
		}
	}

	/////// GESTION TESTIGO PROTOCOLO TDMA ///////
	sumcolas = 0; // Total bytes en las colas de esta BS.
	int conex = (int) this->getParentModule()->par("conexiones"); // Longitud del array de modulos BS_queue[] (numero de potenciales conexiones con otras BSs).
	int colas = conex-1;
	index_used_queues.resize(colas,-1);
	//slot_tx_delay.resize(colas, (simtime_t)0);
	t_frame = (double) this->getParentModule()->par("delta_time"); // Duracion del periodo de trama TDM en los canales inalambricos
	get_num_used_queues(); // Invocacion a la funcion que calcula el numero de modulos BS_queue[] que realmente estan conectados a enlaces que conectan con otras BSs intermedias.
	//ev << "Enlaces activos en esta BS: " << used_queues << endl;
	time_frame = (simtime_t) t_frame;
	//ev << "Tiempo de trama TDMA: " << t_frame << "s" << endl;
	//ev << "Tiempo de slot TDMA: " << t_slot << "s" << endl;
	timepacket = 0;
	//cMessage *msg = new cMessage ("TDMA",1); /////// ACTIVAR ///////
	//scheduleAt((simtime_t)0, msg); /////// ACTIVAR ///////
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION HANDLEMESSAGE(CMESSAGE *MSG)--> ESTA FUNCION SE INVOCA CON EL MENSAJE COMO PARAMETRO CADA VEZ QUE EL MODULO RECIBE UN //
//										  UN MENSAJE O PAQUETE. SE RECORRE EL CODIGO IMPLEMENTADO PARA DEVOLVER UN VALOR O  											//
//										  EJECUTAR UNA FUNCION DENTRO DEL MODULO SIMPLE. EL TIEMPO DE SIMULACION NO TRANSCURRE  									//
//										  DENTRO DE LA LLAMADA DE LA FUNCION HANDLEMESSAGE(CMESSAGE *MSG) MIENTRAS RECORRE EL											//
//										  CODIGO IMPLEMENTADO.																																										//                                                                         //
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void BS_mac::handleMessage(cMessage *msg)
{
	int type = msg->getKind();
	switch(type)
	{
		case 1:
			//ev << "New TDMA frame" << endl;
			scheduleAt(simTime()+time_frame, msg);
			get_num_used_queues(); // Re-obtencion de los indices de los modulos BS_queue[] activos
			timepacket = 0; // Reseteo de variables
			//slot_tx_delay.clear();
			for(int q=0; q<used_queues; q++)
			{
				queue = check_and_cast<BS_queue *>(this->getParentModule()->getSubmodule("bs_queue",index_used_queues[q]));
				queue->sumcheckedpacket = 0;
				queue->sumextractedpacket = 0;
			}
			extract_packets();
		break;
		default:
			delete msg;
			break;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION IS_AN_EMPTY_ROW(INT ROW)--> COMPRUEBA SI LA FILA DE INDICE 'row' DE LA TABLA CON TODAS LAS RUTAS POSIBLES ESTA VACIA.	//
//									  DEVUELVE 0 SI ESTA LIBRE, Y 1 SI ESTA OCUPADA.											 																			//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int BS_mac::is_an_empty_row(int row)
{
	int ocupado = 0;
	if (strcmp(" ",Tabla->array_de_rutas[row][0])!=0)
	{
		ocupado = 1;
	}
	return ocupado;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION GET_INDEX_FIRST_EMPTY_ROW()--> DEVUELVE EL INDICE DE LA PRIMERA FILA VACIA EN LA TABLA 'ARRAY_DE_RUTAS'. //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int BS_mac::get_index_first_empty_row()
{
	int i=0, indice=0;
	int coincide = 1;
	while(i<filas && coincide !=0)
	{
		coincide = strcmp(" ", Tabla->array_de_rutas[i][0]);
		indice = i;
		i++;
	}
	return indice;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION GET_INDEX_LAST_ELEMENT(INT ROW)--> DEVUELVE EL INDICE DE LA ULTIMA COLUMNA OCUPADA EN UNA DETERMINADA	//
//											 FILA DE LA TABLA 'ARRAY_DE_RUTAS'. 								 																		//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int BS_mac::get_index_last_element(int row)
{
	// El primer elemento es el usuario, luego podriamos empezar por i=1
	int i=1, encontrado = 0, ocupado, vacio, indice=1;
	while(encontrado == 0 && i<columnas)
	{
		ocupado = strcmp(" ", Tabla->array_de_rutas[row][i]);
		vacio = strcmp(" ", Tabla->array_de_rutas[row][i + 1]);
		if(ocupado !=0 && vacio ==0)
		{
			encontrado = 1;
			indice = i;
		}
		i++;
	}
	return indice;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION YA_APARECE(CHAR* NOMBRE_NODO, INT FILA, INT INDICE)--> COMPRUEBA SI EL NOMBRE DE UN DETERMINADO NODO YA APARECE	//
//																 EN UNA RUTA DADA DE LA TABLA 'ARRAY_DE_RUTAS'. 																				//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int BS_mac::ya_aparece(char* nombre_nodo, int fila, int indice)
{
	int iguales=0, b=indice, aparece=0;
	while(b>=0 && aparece==0)
	{
		iguales = strcmp(nombre_nodo,Tabla->array_de_rutas[fila][b]);
		if(iguales==0)
		{
			aparece = 1;
		}
		b--;
	}
	return aparece;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION ES_PASARELA(CMODULE *NODE)--> COMPRUEBA SI UN MODULO BS ES O NO PASARELA CON LA PARTE OPTICA.	//
//										DEVUELVE EL INDICE DE LA PUERTA DE LA BS QUE CONECTA CON LA ONU (1),						 	//
//										SI ES UNA PASARELA, Y -1 EN CASO CONTRARIO.							 													//
//////////////////////////////////////////////////////////////////////////////////////////////////////////
int BS_mac::es_pasarela(cModule *node)
{
	cModule *nodo_ = check_and_cast<cModule*>(node);
	cTopology::Node * nodo = top.getNodeFor(nodo_);
	char* nombre_enlace = (char*) nodo->getLinkOut(1)->getRemoteNode()->getModule()->getName();
	int encontrado = -1;
	if (strcmp(nombre_enlace,"onu")==0)
	{
		encontrado = 1;
	}
	return encontrado;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION CONSEGUIR_SIGUIENTE_PUERTA(INT INIT_IND, INT PUERTAS, CMODULE *NODE)--> DEVUELVE EL INDICE DE LA SIGUIENTE PUERTA DE SALIDA /
// 																				  QUE EST� CONECTADA DE LA BS. 						  /
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int BS_mac::conseguir_siguiente_puerta(int init_ind, int puertas, cModule *node)
{
	cModule *nodo = check_and_cast<cModule*>(node);
	int i = init_ind, encontrado = 0;
	int indice = init_ind;
	while(i<puertas && encontrado ==0)
	{
		if(nodo->gate("physicalBS$o",i)->isConnected())
		{
			if(strcmp("user",nodo->gate("physicalBS$o",i)->getNextGate()->getOwnerModule()->getName())!=0)
			{
				encontrado = 1;
				indice = i;
			}
		}
		i++;
	}
	return indice;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION PUERTA_ENTRADA_LIBRE(CMODULE *NODE, INT PUERTAS)--> DEVUELVE EL INDICE DE LA PRIMERA PUERTA DE ENTRADA NO CONECTADA DE LA BS. /
//															  SI NO ENCUENTRA NINGUNA, DEVUELVE -1.									    /
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int BS_mac::puerta_entrada_libre (cModule *node, int puertas)
{
	cModule *nodo = check_and_cast<cModule*>(node);
	int i=0, encontrado = -1;
	while (i<puertas && encontrado ==-1)
	{
		if(!nodo->gate("physicalBS$i",i)->isConnected())
		{
			encontrado = i;
		}
		i++;
	}
	return encontrado;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION PUERTA_SALIDA_LIBRE(CMODULE *NODE, INT PUERTAS)--> DEVUELVE EL INDICE DE LA PRIMERA PUERTA DE SALIDA NO CONECTADA DE LA BS. /
//															 SI NO ENCUENTRA NINGUNA, DEVUELVE -1.									  /
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int BS_mac::puerta_salida_libre (cModule *node, int puertas)
{
	cModule *nodo = check_and_cast<cModule*>(node);
	int i=0, encontrado = -1;
	while (i<puertas && encontrado ==-1)
	{
		if(!nodo->gate("physicalBS$o",i)->isConnected())
		{
			encontrado = i;
		}
		i++;
	}
	return encontrado;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION FINISH()--> ESTA FUNCION SE INVOCA CUANDO LA SIMULACION HA TERMINADO CON EXITO SIN QUE SE PRODUZCA NINGUN ERROR. 		//
//					  LO USAMOS PARA LA RECOGIDA DE ESTADASTICAS Y VISUALIZACION POR PANTALLA O MEDIANTE UN ARCHIVO.            //
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void BS_mac::finish()
{
	if(strcmp(this->getParentModule()->getFullName(),"bs0")==0)
	//if(this->getParentModule()->getIndex()==0) /////// CAMBIO_RED ///////
	{
		retardo_wireless = fopen("results/retardo_wireless.txt", "a+");
		fprintf(retardo_wireless, "TIEMPO SIMULADO: T=%g\n", (double)SIMTIME_DBL(simTime()));
		fprintf(retardo_wireless, "RETARDO MEDIO TOTAL: %g\n", (double)retardo.average());
		fclose(retardo_wireless);
	}
	//int criterio = this->getParentModule()->par("criterio"); // Criterio seguido para encaminar
	//int este_indice = this->getParentModule()->par("indice");
	//int este_indice = this->getParentModule()->getIndex(); /////// CAMBIO_RED ///////
	//int estaciones = this->getParentModule()->par("numBS");
	/*
	if(criterio == 1 && este_indice==estaciones-1)
	{
		ev << endl;
		// Impresion de la tabla de encaminamiento optimo para cada usuario, segun el menor retardo:
		ev << " 	**TABLA CON LA RUTA OPTIMA HASTA UNA ONU (MENOR RETARDO) PARA CADA USUARIO**" << endl;
		for(int a=0; a<(int)this->getParentModule()->par("numUser"); a++)
		{
			for(int b=0; b<columnas; b++)
			{
				ev << " | " << Tabla->rutas_optimas[a][b];
			}
			ev << "  --> Peso (retardo): " << auxiliar_array[a] << " s."<< endl;
		}
		ev << endl;
		//ev << endl;
	//// IMPRESION TABLA 'ARRAY_DE_RETARDOS' (RETARDO DE CADA ENLACE) ////
		//ev << " 	TABLA CON EL PESO (RETARDO) DE CADA ENLACE DE TODAS LAS RUTAS POSIBLES DESDE CADA USUARIO A UNA ONU" << endl;
		for(int a=0; a<max_fila; a++)
		{
			for(int b=0; b<columnas; b++)
			{
				//ev << " | " << Tabla->array_de_pesos[a][b];
			}
			//ev << " --> Total: " << Tabla->peso_total_rutas[a] << endl;
		}
	}
	*/
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION GET_NUM_USED_QUEUES()--> CALCULA CUANTOS MODULOS BS_QUEUE ESTAN TRANSMITIENDO DATOS CON REGULARIDAD, ES DECIR,	/
//								   PUEDEN CONSIDERARSE "ACTIVOS" (USO=1 Y USADO=1). DEVUELVE EL TOTAL DE BS_QUEUE ACTIVAS.	/
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void BS_mac::get_num_used_queues()
{
	used_queues = 0; // Reseteo de la variable
	int num_conexiones = (int) this->getParentModule()->par("conexiones");
	int queues = num_conexiones - 1;
	int auxiliar = 0, contador = 0;
	// Se recorren todos los modulos BS_queue[] de este modulo compuesto BS para comprobar el valor de sus parametros
	// "usado" y "uso". Recordar que cada modulo BS_queue[] corresponde a un enlace entre esta BS y una vecina.
	for(int q=0; q<queues; q++)
	{
		queue = check_and_cast<BS_queue *>(this->getParentModule()->getSubmodule("bs_queue",q));
		if(queue->uso==1 && queue->usado==1)
		{
			used_queues++;
			index_used_queues[contador]=q; // Almacenamiento del indice de cada BS_queue[] 'activa' en este vector.
			contador++;
			if(auxiliar==0)
			{
				ind_primero = q; // Almacenamiento del indice de la primera BS_queue[] del array que esta activa.
				auxiliar = 1;
			}
		}
	}
	// Los posiciones del array que no se han rellenado (porque no hay mas BS_queue[] activos) se rellenan con el valor "-1".
	for(int h=used_queues; h<(int)index_used_queues.size(); h++)
	{
		index_used_queues[h] = -1;
	}
	// Tiempo de slot = Tiempo de trama entre numero de modulos BS_queue[] activos:
	t_slot = t_frame/used_queues;
	// Capacidad efectiva = Capacidad del canal entre numero de modulos BS_queue[] activos:
	eff_cap = (double) txrate/used_queues;
	// Ancho de banda que se puede transmitir en el tiempo de slot que toca a cada BS_queue[]:
	double bw = (double) (txrate*t_slot)/8;
	// Se actualizan los sigquientes parametros en los modulos BS_queue[] activos:
	for(int h=0; h<used_queues; h++)
	{
		// Posicion que ocupa este submodulo BS_queue[] en el vector de colas usadas del BS_mac:
		check_and_cast<BS_queue *>(this->getParentModule()->getSubmodule("bs_queue",index_used_queues[h]))->used_index = h;
		// Ancho de banda disponible por el enlace que representa este BS_queue[] para el siguiente periodo de transmision:
		check_and_cast<BS_queue *>(this->getParentModule()->getSubmodule("bs_queue",index_used_queues[h]))->bandwidth = bw;
		// Tiempo de slot de este BS_queue[]:
		check_and_cast<BS_queue *>(this->getParentModule()->getSubmodule("bs_queue",index_used_queues[h]))->t_slot = t_slot;
		// Numero de modulos BS_queue[] activos en este modulo compuesto BS:
		check_and_cast<BS_queue *>(this->getParentModule()->getSubmodule("bs_queue",index_used_queues[h]))->used_bs_queues = used_queues;
	}

	//ev << " Numero de enlaces activos: " << used_queues << endl;
	//ev << " T_slot = " << t_slot << "s." << endl;
	/*
	ev << " Indice del primer enlace activo: " << ind_primero << endl;
	ev << " Indices de todos los enlaces activos: " << endl;
	for(int h=0; h<(int)index_used_queues.size(); h++)
	{
		ev << "	" << index_used_queues[h];
	}
	ev << endl;
	*/
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION EXTRACT_PACKETS()--> METODO DE EXTRACCION DE PAQUETES EN LAS COLAS TENIENDO EN CUENTA LA PRIORIDAD. /														/
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
void BS_mac::extract_packets()
{
	// Segun el metodo de extraccion de paquetes elegido en 'omnetpp.ini':
	switch(extraction_method)
	{
	case 0: // Metodo de extraccion con prioridad de colas estricta:
		sumcolas = 0;
		for(int w=0; w<used_queues; w++)
		{
			queue = check_and_cast<BS_queue *>(this->getParentModule()->getSubmodule("bs_queue",index_used_queues[w]));
			queue->tamsumcola = 0; // Reseteo de la variable
			for (int h=0; h<prioridades; h++)
			{
				queue->tamsumcola = queue->tamsumcola + queue->tamqueue[h];
				//ev << " Bytes en la cola de prioridad P" << h << " con destino " << queue->getTarget() << ": " << queue->tamqueue[h] << "B" << endl;
			}

			// Si en este modulo BS_queue[], que a priori se considera "activo", no hay ningun paquete durante 2 ciclos,
			// se cambia el calor de su variable "usado" para que en el proximo periodo se considere a este submodulo "inactivo":
			if(queue->tamsumcola==0)
			{
				queue->contador++;
				if(queue->contador>=1)
				{
					queue->usado=0;
					//ev << " El enlace " << this->getParentModule()->getFullName() << "-->" << queue->getTarget() << " esta INACTIVO. " << endl;
				}
			}

			//ev << " Total bytes con destino " << queue->getTarget() << ": " << queue->tamsumcola << "B" << endl;
			sumcolas = sumcolas + queue->tamsumcola;
		}
		//ev << " Total bytes en las colas del nodo " << this->getParentModule()->getFullName() << ": " << sumcolas << "B" << endl;

		// Para cada uno de los modulo BS_queue[] "activos", se invoca su metodo de extraccion de paquetes:
		for(int j=0; j<used_queues; j++)
		{
			queue = check_and_cast<BS_queue *>(this->getParentModule()->getSubmodule("bs_queue",index_used_queues[j]));
			queue->strictpriorityqueue();
			// NOTA: Se envian todos los paquetes "de golpe" de los BS_queue[] al BS_rx_tx, pero este modulo no los pone en la red todos a la vez,
			// 			 sino que los envia al siguiente nodo con el retardo de transmision correspondiente (variable timepacket).
		}
		break;
	case 1: // Metodo de extraccion centralizado:
		sumcolas = 0;
		for(int w=0; w<used_queues; w++)
		{
			queue = check_and_cast<BS_queue *>(this->getParentModule()->getSubmodule("bs_queue",index_used_queues[w]));
			queue->tamsumcola = 0; // Reseteo de la variable
			for (int h=0; h<prioridades; h++)
			{
				queue->tamsumcola = queue->tamsumcola + queue->tamqueue[h];
				//ev << " Bytes en la cola de prioridad P" << h << " con destino " << queue->getTarget() << ": " << queue->tamqueue[h] << "B" << endl;
			}

			// Si en este modulo BS_queue[], que a priori se considera "activo", no hay ningun paquete durante 2 ciclos,
			// se cambia el calor de su variable "usado" para que en el proximo periodo se considere a este submodulo "inactivo":
			if(queue->tamsumcola==0)
			{
				queue->contador++;
				if(queue->contador>=1)
				{
					queue->usado=0;
					//ev << " El enlace " << this->getParentModule()->getFullName() << "-->" << queue->getTarget() << " esta INACTIVO. " << endl;
				}
			}

			//ev << " Total bytes con destino " << queue->getTarget() << ": " << queue->tamsumcola << "B" << endl;
			sumcolas = sumcolas + queue->tamsumcola;
		}
		//ev << " Total bytes en las colas del nodo " << this->getParentModule()->getFullName() << ": " << sumcolas << "B" << endl;

		for(int j=0; j<used_queues; j++)
		{
			queue = check_and_cast<BS_queue *>(this->getParentModule()->getSubmodule("bs_queue",index_used_queues[j]));
			// Para cada uno de los modulo BS_queue[] "activos", se invoca su metodo de extraccion de paquetes:
			queue->centralizedmethod();
			// NOTA: Se envian todos los paquetes "de golpe" de los BS_queue[] al BS_rx_tx, pero este modulo no los pone en la red todos a la vez,
			// 			 sino que los envia al siguiente nodo con el retardo de transmision correspondiente (variable timepacket).
		}
		break;
	default:
		break;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION RELLENO_ARRAYS_RETARDOS()--> RELLENA LA TABLA CON LOS RETARDOS DE CADA ENLACE PARA CADA UNA DE LAS RUTAS POSIBLES, /
//									   ASI COMO EL VECTOR CON EL RETARDO TOTAL DE CADA UNA.									 /
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void BS_mac::relleno_arrays_retardos()
{
	// Se incluyen en la topologia los nodos que contienen el parametro 'aux' (Users, BSs y ONUs):
	top.extractByParameter("aux");
	/////// INICIO RELLENO TABLA CON LOS PESOS (RETARDO) DE CADA ENLACE PARA TODAS LAS RUTAS POSIBLES ///////
	for(int i=0; i<max_fila; i++)
	{
		for(int j=0; j<columnas; j++)
		{
			if(strcmp(Tabla->array_de_rutas[i][j]," ")!=0 && strpbrk(Tabla->array_de_rutas[i][j],"on")==NULL)
			{
				int ind_1=0,ind_2=0;
				for(int p=0; p<(int)top.getNumNodes(); p++)
				{
					if(strcmp(top.getNode(p)->getModule()->getFullName(),Tabla->array_de_rutas[i][j])==0)
					{
						ind_1=p;
					}
					if(strcmp(top.getNode(p)->getModule()->getFullName(),Tabla->array_de_rutas[i][j+1])==0)
					{
						ind_2=p;
					}
				}
				top_node = top.getNode(ind_1);
				////ev << top_node->getModule()->getFullName() << " | ";
				int ind_aux=-1, k=0;
				while(ind_aux<0 && k<(int)top_node->getNumOutLinks())
				{
					////ev << " Otro extremo canal " << k << ": " << top_node->getLinkOut(k)->getRemoteGate()->getOwnerModule()->getFullName() << endl;
					////ev << " Siguiente nodo en la ruta: " << top.getNode(ind_2)->getModule()->getFullName() << " == " << Tabla->array_de_rutas[i][j+1] << endl;
					if(strcmp(top_node->getLinkOut(k)->getRemoteGate()->getOwnerModule()->getFullName(),top.getNode(ind_2)->getModule()->getFullName())==0)
					{
						ind_aux=k;
					}
					k++;
				}
				/*for (int k=0; k<(int)top_node->getNumOutLinks(); k++)
				{
					if(strcmp(top_node->getLinkOut(k)->getRemoteGate()->getOwnerModule()->getFullName(),top.getNode(ind_2)->getModule()->getFullName())==0)
					{
						ind_aux=k;
					}
				}*/
				//AQUI//
				if(strcmp(top_node->getLinkOut(ind_aux)->getRemoteNode()->getModule()->getName(),"onu")!=0)
				{
					Tabla->array_de_pesos[i][j] = (check_and_cast<cDelayChannel*>(top_node->getLinkOut(ind_aux)->getLocalGate()->getChannel()))->par(3);
				}
			}
		}
	}
	/////// FIN RELLENO TABLA CON LOS PESOS (RETARDO) DE CADA ENLACE PARA TODAS LAS RUTAS POSIBLES ///////

	/////// INICIO RELLENO VECTOR CON EL PESO (RETARDO) DE CADA POSIBLE RUTA ///////
	for(int a=0; a<max_fila; a++)
	{
		for(int b=0; b<columnas; b++)
		{
			Tabla->peso_total_rutas[a] = 0;
		}
	}
	for(int a=0; a<max_fila; a++)
	{
		for(int b=0; b<columnas; b++)
		{
			Tabla->peso_total_rutas[a] = Tabla->peso_total_rutas[a] + Tabla->array_de_pesos[a][b];
		}
	}
	/////// FIN RELLENO VECTOR CON EL PESO (RETARDO) DE CADA POSIBLE RUTA ///////
	/*//--->
	ev << " **ACTUALIZACION TABLA DE PESOS DE CADA ENLACE DE TODAS LAS RUTAS POSIBLES DESDE CADA USUARIO A UNA ONU**" << endl;
	for(int x=0; x<max_fila; x++)
	{
		ev << Tabla->peso_total_rutas[x] << " 	== 	";
		int last = get_index_last_element(x);
		for(int b=0; b<last; b++)
		{
			ev << Tabla->array_de_pesos[x][b] << "	+	";
		}
		ev << endl;
	}
	//<---*/
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION GET_ROUTES_FOR_EACH_USER(INT INDEX)--> OBTIENE, A PARTIR DE LA TABLA CON TODAS LAS RUTAS, LOS INDICES DE LAS RUTAS/
//												 QUE CORRESPONDEN AL USUARIO DE INDICE 'INDEX' Y EL NUMERO TOTAL DE RUTAS,	/
//												 CLASIFICADAS SEGUN LA ONU PASARELA DE DESTINO.								/
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void BS_mac::get_routes_for_each_user(int index)
{
	/* Busqueda de todas las rutas posibles desde este usuario y una ONU pasarela */
	cModule *user;//, *bs;
	user = (cModule*) this->getParentModule()->getParentModule()->getSubmodule("user",index);
	int pasarelas = (int) this->getParentModule()->par("numGW"); // Numero de BSs pasarela
	int indice_esta_bs = (int)this->getParentModule()->par("indice");
	//int indice_esta_bs = (int)this->getParentModule()->getIndex(); /////// CAMBIO_RED ///////
	int estaciones = (int) this->getParentModule()->par("numBS"); // Numero de estaciones base de la red (25)
	int connections = (int) this->getParentModule()->par("conexiones");
	int rutas_user = estaciones*pasarelas*connections*pasarelas;
	Tabla->indices_rutas[index].resize(rutas_user);

	for (int w=0; w<rutas_user; w++)
	{
		Tabla->indices_rutas[index][w].resize(pasarelas,-1);
	}
	////ev << " Dimensiones de la tabla 'indices_rutas[" << index << "]': " << Tabla->indices_rutas[index].size() << "x" << Tabla->indices_rutas[index][0].size() << endl;
	char* nombre_user = (char*) user->getFullName(); // Nombre del usuario al que pertenece este modulo
	char* nombre_pasarela; // Nombre de la BS pasarela
	int auxiliar, ultimo;
	for(int i=0; i<pasarelas; i++)
	{
		auxiliar=0; // Entero auxiliar para rellenar cada fila de la matriz con los indices de las rutas de este usuario a cada pasarela.
				//bs = (cModule*) this->getParentModule()->getParentModule()->getSubmodule(nombres_bs[indices_pasarelas[i]]);
		nombre_pasarela = (char*) nombres_bs[indices_pasarelas[i]];
		//bs = (cModule*) this->getParentModule()->getParentModule()->getSubmodule("bs",indices_pasarelas[i]); /////// CAMBIO_RED ///////
		//nombre_pasarela = (char*) bs->getFullName(); /////// CAMBIO_RED ///////

		for(int a=0; a<max_fila; a++)
		{
			ultimo = get_index_last_element(a);
			if(strcmp(nombre_user, Tabla->array_de_rutas[a][0])==0 && strcmp(nombre_pasarela, Tabla->array_de_rutas[a][ultimo-1])==0)
			{
				Tabla->indices_rutas[index][auxiliar][i] = a;
				auxiliar++;
			}
		}
	}
	// SUMA DE LOS ELEMENTOS DEL
	int contador;
	for(int t=0; t<pasarelas; t++)
	{
		contador=0;
		// ... se recorren todos los indices almacenados:
		for(int y=0; y<rutas_user; y++)
		{
			if(Tabla->indices_rutas[index][y][t]!=-1)
			{
				contador++;
			}
		}
		Tabla->num_rutas[index][t] = contador;
		if(indice_esta_bs == estaciones-1)
		{
			//ev << " Numero de rutas desde el usuario " << nombre_user << " y la pasarela " << nombres_bs[indices_pasarelas[t]] << ": " << Tabla->num_rutas[index][t] << endl;
		}
	}
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION CALCULO_RUTA_SHORTEST_PATH(INT IND_USR)--> CALCULA LA RUTA DE MENOR LONGITUD PARA EL USUARIO DE INDICE 'IND_USR'	/
//													 Y ACTUALIZA SU ENTRADA CORRESPONDIENTE EN LA TABLA DE RUTAS OPTIMAS.	/
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void BS_mac::calculo_ruta_shortest_path(int ind_usr)
{
	int indice_esta_bs = (int) this->getParentModule()->par("indice");
	//int indice_esta_bs = (int) this->getParentModule()->getIndex(); /////// CAMBIO_RED ///////
	int estaciones = (int) this->getParentModule()->par("numBS"); // Numero de estaciones base de la red (25)
	int usuarios = (int) this->getParentModule()->par("numUser"); // Numero de usuarios
	int pasarelas = (int) this->getParentModule()->par("numGW"); // Numero de BS-pasarelas
	/////// ENCAMINAMIENTO SEGUN EL CAMINO CON MENOR LONGITUD ///////
	for(int j=0; j<pasarelas; j++)
	{
		int min_index = Tabla->indices_rutas[ind_usr][0][j];
		if(Tabla->num_rutas[ind_usr][j]>1)
		{
			for(int i=0; i<Tabla->num_rutas[ind_usr][j]; i++)
			{
				////ev << " Longitud ruta num. " << i << " del usuario User[" << ind_usr << "]:" << Tabla->longitud_total_rutas[Tabla->indices_rutas[ind_usr][i][j]] << endl;

				//if(Tabla->longitud_total_rutas[Tabla->indices_rutas[ind_usr][i][j]] < Tabla->longitud_total_rutas[min_index])
				if(Tabla->peso_total_rutas[Tabla->indices_rutas[ind_usr][i][j]] < Tabla->peso_total_rutas[min_index])
				{
					min_index=Tabla->indices_rutas[ind_usr][i][j];
				}
			}
			opt_route[j]=min_index;
			////ev << " 	Longitud de la ruta mas corta hasta " << nombres_bs[indices_pasarelas[h]] << ": " << Tabla->longitud_total_rutas[min_index] << endl;
		}
		else if(Tabla->num_rutas[ind_usr][j]==1)
		{
			////ev << " Longitud ruta num. 0 del usuario User[" << ind_usr << "]:" << Tabla->longitud_total_rutas[Tabla->indices_rutas[ind_usr][0][j]] << endl;
			min_index = Tabla->indices_rutas[ind_usr][0][j];
			opt_route[j]=min_index;
			////ev << " 	Longitud de la ruta mas corta hasta " << nombres_bs[indices_pasarelas[h]] << ": " << Tabla->longitud_total_rutas[min_index] << endl;
		}
		else if(Tabla->num_rutas[ind_usr][j]==0)
		{
			opt_route[j]= -1;
		}
	}

	for(int q=0; q<pasarelas; q++)
	{
		if(opt_route[q]!=-1)
		{
			Tabla->rutas_optimas_pasarelas[q][ind_usr] = Tabla->array_de_rutas[opt_route[q]];
			auxiliar_matrix[q][ind_usr] = Tabla->peso_total_rutas[opt_route[q]];
		}
	}
	if(ind_usr == usuarios-1 && indice_esta_bs==estaciones-1)
	{
		for(int k=0; k<pasarelas; k++)
		{
			//ev << endl;
			//ev << " **TABLA CON LA " << k << "-ESIMA RUTA OPTIMA (MENOR DISTANCIA) HASTA LA PASARELA " << nombres_bs[indices_pasarelas[k]]<< " PARA CADA USUARIO**" << endl;
			for(int a=0; a<usuarios; a++)
			{
				for(int b=0; b<columnas; b++)
				{
					//ev << " | " << Tabla->rutas_optimas_pasarelas[k][a][b];
				}
				//ev << " --> Peso: " << auxiliar_matrix[k][a] << endl;
			}
			//ev << endl;
		}
	}

	int aux1=0, first_index=0, first_onu=0;
	for(int h=0; h<pasarelas; h++)
	{
		if(opt_route[h]>0 && aux1==0)
		{
			aux1=1;
			first_index=opt_route[h];
			first_onu=indices_pasarelas[h];
			////ev << " Longitud de la ruta mas corta hasta la " << nombres_bs[indices_pasarelas[h]] << ": " << Tabla->longitud_total_rutas[opt_route[h]] << endl;
		}
		else if(opt_route[h]>0 && aux1!=0)
		{
			////ev << " Longitud de la ruta mas corta hasta la " << nombres_bs[indices_pasarelas[h]] << ": " << Tabla->longitud_total_rutas[opt_route[h]] << endl;

			if(Tabla->peso_total_rutas[opt_route[h]]<Tabla->peso_total_rutas[first_index])
			//if(Tabla->longitud_total_rutas[opt_route[h]]<Tabla->longitud_total_rutas[first_index])
			{
				first_index=opt_route[h];
				first_onu=indices_pasarelas[h];
			}
		}
	}
	////ev << " Ruta mas corta --> " << nombres_bs[indices_pasarelas[h]] << " --> Longitud: " << Tabla->longitud_total_rutas[first_index] << endl;

	Tabla->rutas_optimas[ind_usr] = Tabla->array_de_rutas[first_index];
	media_saltos = (double)(media_saltos + get_index_last_element(first_index)-1);
	auxiliar_array[ind_usr] = Tabla->longitud_total_rutas[first_index];
	//auxiliar_array[ind_usr] = Tabla->peso_total_rutas[first_index];

	if(ind_usr == usuarios-1)
	//if(ind_usr == usuarios-1 && indice_esta_bs==estaciones-1)
	{
		//ev << endl;
		//ev << "		" << this->nombres_bs[indice_esta_bs] << endl;
		//ev << " **TABLA CON LA RUTA OPTIMA HASTA UNA ONU (MENOR DISTANCIA) PARA CADA USUARIO**" << endl;
		for(int a=0; a<usuarios; a++)
		{
			for(int b=0; b<columnas; b++)
			{
				//ev << " | " << Tabla->rutas_optimas[a][b];
			}
			//ev << "  --> Longitud: " << auxiliar_array[a] << " metros" << endl;
		}
		//ev << endl;
	}
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION CALCULO_RUTA_MINIMUM_HOP(INT IND_USR)--> CALCULA LA RUTA DE MENOR NUMERO DE SALTOS PARA EL USUARIO DE INDICE 'IND_USR'/
//												   Y ACTUALIZA SU ENTRADA CORRESPONDIENTE EN LA TABLA DE RUTAS OPTIMAS.			/
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void BS_mac::calculo_ruta_minimum_hop(int ind_usr)
{
	int indice_esta_bs = (int)this->getParentModule()->par("indice");
	//int indice_esta_bs = (int) this->getParentModule()->getIndex(); /////// CAMBIO_RED ///////
	int estaciones = (int) this->getParentModule()->par("numBS"); // Numero de estaciones base de la red (25)
	int usuarios = (int) this->getParentModule()->par("numUser"); // Numero de usuarios
	int pasarelas = (int) this->getParentModule()->par("numGW"); // Numero de BS-pasarelas
	/////// ENCAMINAMIENTO SEGUN EL CAMINO CON MENOR NUMERO DE SALTOS ///////
	for(int j=0; j<pasarelas; j++)
	{
		int min_index = Tabla->indices_rutas[ind_usr][0][j];
		if(Tabla->num_rutas[ind_usr][j]>1)
		{
			for(int i=0; i<Tabla->num_rutas[ind_usr][j]; i++)
			{
				////ev << " Saltos ruta num. " << i << " del usuario User[" << ind_usr << "]:" << Tabla->saltos_total_rutas[Tabla->indices_rutas[ind_usr][i][j]] << endl;
				//if(Tabla->saltos_total_rutas[Tabla->indices_rutas[ind_usr][i][j]] < Tabla->saltos_total_rutas[min_index])
				if(Tabla->peso_total_rutas[Tabla->indices_rutas[ind_usr][i][j]] < Tabla->peso_total_rutas[min_index])
				{
					min_index=Tabla->indices_rutas[ind_usr][i][j];
				}
			}
			opt_route[j]=min_index;
			////ev << " 	Numero de saltos de la ruta con menos saltos hasta " << nombres_bs[indices_pasarelas[j]] << ": " << Tabla->saltos_total_rutas[min_index] << endl;
		}
		else if(Tabla->num_rutas[ind_usr][j]==1)
		{
			////ev << " Numero de saltos de la ruta num. 0 del usuario User[" << ind_usr << "]:" << Tabla->saltos_total_rutas[Tabla->indices_rutas[ind_usr][0][j]] << endl;
			min_index = Tabla->indices_rutas[ind_usr][0][j];
			opt_route[j]=min_index;
			////ev << " 	Numero de saltos de la ruta con menos saltos hasta " << nombres_bs[indices_pasarelas[j]] << ": " << Tabla->saltos_total_rutas[min_index] << endl;
		}
		else if(Tabla->num_rutas[ind_usr][j]==0)
		{
			opt_route[j]= -1;
		}
	}

	for(int q=0; q<pasarelas; q++)
	{
		if(opt_route[q]!=-1)
		{
			Tabla->rutas_optimas_pasarelas[q][ind_usr] = Tabla->array_de_rutas[opt_route[q]];
			auxiliar_matrix[q][ind_usr] = Tabla->peso_total_rutas[opt_route[q]];
		}
	}
	if(ind_usr == usuarios-1 && indice_esta_bs==estaciones-1)
	{
		for(int k=0; k<pasarelas; k++)
		{
			//ev << endl;
			//ev << " **TABLA CON LA " << k << "-ESIMA RUTA OPTIMA (MENOR NUM. SALTOS) HASTA LA PASARELA " << nombres_bs[indices_pasarelas[k]]<< " PARA CADA USUARIO**" << endl;
			for(int a=0; a<usuarios; a++)
			{
				for(int b=0; b<columnas; b++)
				{
					//ev << " | " << Tabla->rutas_optimas_pasarelas[k][a][b];
				}
				//ev << " --> Peso: " << auxiliar_matrix[k][a] << endl;
			}
			//ev << endl;
		}
	}

	int aux1=0, first_index=0, first_onu=0;
	for(int h=0; h<pasarelas; h++)
	{
		if(opt_route[h]>0 && aux1==0)
		{
			aux1=1;
			first_index=opt_route[h];
			first_onu=indices_pasarelas[h];
			////ev << " Numero de saltos de la ruta con menos saltos hasta la " << nombres_bs[first_onu] << ": " << Tabla->saltos_total_rutas[opt_route[h]] << endl;
		}
		else if(opt_route[h]>0 && aux1!=0)
		{
			////ev << " Numero de saltos de la ruta con menos saltos hasta " << nombres_bs[indices_pasarelas[h]] << ": " << Tabla->saltos_total_rutas[opt_route[h]] << endl;
			//if(Tabla->saltos_total_rutas[opt_route[h]]<Tabla->saltos_total_rutas[first_index])
			if(Tabla->peso_total_rutas[opt_route[h]]<Tabla->peso_total_rutas[first_index])
			{
				first_index=opt_route[h];
				first_onu=indices_pasarelas[h];
			}
		}
	}
	////ev << " Ruta con menor numero de saltos --> " << nombres_bs[first_onu] << " --> Num: saltos: " << Tabla->saltos_total_rutas[first_index] << endl;

	Tabla->rutas_optimas[ind_usr] = Tabla->array_de_rutas[first_index];
	media_saltos = (double)(media_saltos + get_index_last_element(first_index) - 1);
	auxiliar_array[ind_usr] = Tabla->saltos_total_rutas[first_index];
	//auxiliar_array[ind_usr] = Tabla->peso_total_rutas[first_index];

	if(ind_usr == usuarios-1)
	//if(ind_usr == usuarios-1 && indice_esta_bs==estaciones-1)
	{
		//ev << endl;
		//ev << "		" << this->nombres_bs[indice_esta_bs] << endl;
		//ev << " **TABLA CON LA RUTA OPTIMA HASTA UNA ONU (MENOR NUMERO DE SALTOS) PARA CADA USUARIO**" << endl;
		for(int a=0; a<usuarios; a++)
		{
			for(int b=0; b<columnas; b++)
			{
				//ev << " | " << Tabla->rutas_optimas[a][b];
			}
			//ev << "  --> Saltos: " << auxiliar_array[a] << endl;
		}
		//ev << endl;
	}
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION RECALCULO_RUTA_DELAY_AWARE(INT IND_USR)--> (RE)CALCULA LA RUTA DE MENOR RETARDO PARA EL USUARIO DE INDICE 'IND_USR'	/
//													 Y ACTUALIZA SU ENTRADA CORRESPONDIENTE EN LA TABLA DE RUTAS OPTIMAS.		/
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void BS_mac::recalculo_ruta_delay_aware(int ind_usr)
{
	//int indice_esta_bs = (int)this->getParentModule()->par("indice");
	//int indice_esta_bs = (int) this->getParentModule()->getIndex(); /////// CAMBIO_RED ///////
	//int estaciones = (int) this->getParentModule()->par("numBS"); // Numero de estaciones base de la red (25)
	//int usuarios = (int) this->getParentModule()->par("numUser"); // Numero de usuarios
	int pasarelas = (int) this->getParentModule()->par("numGW"); // Numero de BS-pasarelas

	/////// ENCAMINAMIENTO SEGUN EL CAMINO CON MENOS RETARDO ///////
	for(int j=0; j<pasarelas; j++)
	{
		int min_index = Tabla->indices_rutas[ind_usr][0][j];
		if(Tabla->num_rutas[ind_usr][j]>1)
		{
			for(int i=0; i<Tabla->num_rutas[ind_usr][j]; i++)
			{
				////ev << " Retardo ruta num. " << i << " del usuario User[" << ind_usr << "]:" << Tabla->retardo_total_rutas[Tabla->indices_rutas[ind_usr][i][j]] << endl;
				if(Tabla->peso_total_rutas[Tabla->indices_rutas[ind_usr][i][j]] < Tabla->peso_total_rutas[min_index])
				{
					min_index=Tabla->indices_rutas[ind_usr][i][j];
				}
			}
			opt_route[j]=min_index;
			////ev << " 	Peso de la ruta con menos retardo hasta " << nombres_bs[indices_pasarelas[h]] << ": " << Tabla->retardo_total_rutas[min_index] << endl;
		}
		else if(Tabla->num_rutas[ind_usr][j]==1)
		{
			////ev << " Peso ruta num. 0 del usuario User[" << ind_usr << "]:" << Tabla->retardo_total_rutas[Tabla->indices_rutas[ind_usr][0][j]] << endl;
			min_index = Tabla->indices_rutas[ind_usr][0][j];
			opt_route[j]=min_index;
			////ev << " 	Peso de la ruta con menos retardo hasta " << nombres_bs[indices_pasarelas[h]] << ": " << Tabla->retardo_total_rutas[min_index] << endl;
		}
		else if(Tabla->num_rutas[ind_usr][j]==0)
		{
			opt_route[j]= -1;
		}
	}

	for(int q=0; q<pasarelas; q++)
	{
		if(opt_route[q]!=-1)
		{
			Tabla->rutas_optimas_pasarelas[q][ind_usr] = Tabla->array_de_rutas[opt_route[q]];
			auxiliar_matrix[q][ind_usr] = Tabla->peso_total_rutas[opt_route[q]];
		}
	}
	/*
	if(ind_usr == usuarios-1 && indice_esta_bs == estaciones-1 && actualizaciones == 0)
	{
		for(int k=0; k<pasarelas; k++)
		{
			//ev << endl;
			//ev << " **TABLA CON LA " << k << "-ESIMA RUTA OPTIMA (MENOR PESO) HASTA LA PASARELA " << nombres_bs[indices_pasarelas[k]]<< " PARA CADA USUARIO**" << endl;
			for(int a=0; a<usuarios; a++)
			{
				for(int b=0; b<columnas; b++)
				{
					//ev << " | " << Tabla->rutas_optimas_pasarelas[k][a][b];
				}
				//ev << " --> Peso: " << auxiliar_matrix[k][a] << endl;
			}
			//ev << endl;
		}

		//// IMPRESION TABLA 'ARRAY_DE_PESOS' (PESO DE CADA ENLACE) ////
			//ev << " TABLA CON EL PESO DE CADA ENLACE DE TODAS LAS RUTAS POSIBLES DESDE CADA USUARIO A UNA ONU" << endl;
			for(int a=0; a<max_fila; a++)
			{
				for(int b=0; b<columnas; b++)
				{
					//ev << " | " << Tabla->array_de_pesos[a][b];
				}
				//ev << " --> Total: " << Tabla->peso_total_rutas[a] << endl;
			}
			//ev << endl;
	}
	*/
	int aux1=0, first_index=0, first_onu=0;
	for(int h=0; h<pasarelas; h++)
	{
		if(opt_route[h]>0 && aux1==0)
		{
			aux1=1;
			first_index=opt_route[h];
			first_onu=indices_pasarelas[h];
			////ev << " Peso de la ruta con menos retardo hasta la " << nombres_bs[indices_pasarelas[h]] << ": " << Tabla->retardo_total_rutas[opt_route[h]] << endl;
		}
		else if(opt_route[h]>0 && aux1!=0)
		{
			////ev << " Peso de la ruta con menos retardo hasta la " << nombres_bs[indices_pasarelas[h]] << ": " << Tabla->retardo_total_rutas[opt_route[h]] << endl;
			if(Tabla->peso_total_rutas[opt_route[h]]<Tabla->peso_total_rutas[first_index])
			{
				first_index=opt_route[h];
				first_onu=indices_pasarelas[h];
			}
		}
	}
	////ev << " Ruta con menos retardo --> " << nombres_bs[indices_pasarelas[first_onu]] << " --> Retardo: " << Tabla->retardo_total_rutas[first_index] << endl;

	Tabla->rutas_optimas[ind_usr] = Tabla->array_de_rutas[first_index];
	media_saltos = (double)(media_saltos + get_index_last_element(first_index)-1);
	auxiliar_array[ind_usr] = Tabla->peso_total_rutas[first_index];
	/*
	if(ind_usr == usuarios-1 && indice_esta_bs==estaciones-1 && actualizaciones==0)
	//if(ind_usr == usuarios-1 && indice_esta_bs==estaciones-1)
	{
		//ev << endl;
		//ev << "		" << this->nombres_bs[indice_esta_bs] << endl;
		ev << " **TABLA CON LA RUTA OPTIMA HASTA UNA ONU (MENOR RETARDO) PARA CADA USUARIO**" << endl;
		for(int a=0; a<usuarios; a++)
		{
			for(int b=0; b<columnas; b++)
			{
				ev << " | " << Tabla->rutas_optimas[a][b];
			}
			ev << "  --> Peso (retardo): " << auxiliar_array[a] << " s."<< endl;
		}
		ev << endl;
		//actualizaciones++;
		ev << endl;
	//// IMPRESION TABLA 'ARRAY_DE_PESOS' (PESO DE CADA ENLACE) ////
		ev << " TABLA CON EL PESO DE CADA ENLACE DE TODAS LAS RUTAS POSIBLES DESDE CADA USUARIO A UNA ONU" << endl;
		for(int g=0; g<usuarios; g++)
		{
			for(int h=0; h<pasarelas; h++)
			{
				for(int i=0; i<Tabla->num_rutas[g][h]; i++)
				{
					for(int j=0; j<columnas; j++)
					{
						ev << " | " << Tabla->array_de_pesos[Tabla->indices_rutas[g][i][h]][j];
					}
					ev << " --> Total: " << Tabla->peso_total_rutas[Tabla->indices_rutas[g][i][h]] << endl;
				}
			}
		}
		ev << endl;
	}
	*/
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION ROUTING()--> ESTA FUNCION REALIZA LAS FUNCIONES DE ENCAMINAMIENTO AL INICIALIZARSE LA RED. 			/
//					   PRIMERO COMPUTA TODOS LOS CAMINOS POSIBLES DE UNA BS ORIGEN A UNA ONU-PASARELA DESTINO, 	/
//					   Y RELLENA TODAS LAS TABLAS INVOLUCRADAS.													/
//					   DESPUES, PARTIENDO DE ESTAS TABLAS, SE CALCULA LA RUTA OPTIMA PARA CADA BS EN FUNCION 	/
//					   DEL METODO DE ENCAMINAMIENTO ELEGIDO (SPRA, MHRA, DARA).									/
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void BS_mac::routing()
{
	// VARIABLES AUXILIARES //
	int indice_esta_bs = (int)this->getParentModule()->par("indice"); // Indice de esta BS
	//int indice_esta_bs = (int) this->getParentModule()->getIndex(); /////// CAMBIO_RED ///////
	int estaciones = (int) this->getParentModule()->par("numBS"); // Numero de estaciones base de la red
	int num_pasarelas = (int) this->getParentModule()->par("numGW"); // Numero de BSs pasarela
	int usuarios = (int) this->getParentModule()->par("numUser"); // Corresponde a las filas de la tabla rutas_optimas[][]
	int criterio = (int) this->getParentModule()->par("criterio"); // Criterio seguido para encaminar (SPRA, DARA, MHRA)
	cModule *red = (cModule*) this->getParentModule()->getParentModule(); // Referencia al modulo tipo cNetwork que corresponde a la red simulada
	cModule *nodo, *sig_nodo;
	cTopology::Node *topo_nodo;
	int enlaces;
	char* nombre_usuario;
	char* nombre_nodo;
	char* nombre_sig_nodo;
	double dist_user_bs = 0, distancia = 0, del_user_bs = 0.0, retardo = 0.0, wei_user_bs = 0.0, peso = 0.0;

	//// INICIO CALCULO RUTAS PARA CADA USUARIO (RELLENO TABLAS) ////
	// Para cada usuario, se meten tantas filas como posibles caminos hay hasta una onu pasarela.
	for(int a=0; a<usuarios; a++)
	{
		// La primera celda de todas las filas contiene el nombre del modulo del usuario inalambrico:
		nombre_usuario = (char*) red->getSubmodule("user",a)->getFullName();
		// Acceso remoto al modulo BS que atiende a cada usuario:
		nodo = red->getSubmodule("user",a)->gate("physicalUser$o")->getNextGate()->getOwnerModule();
		nombre_nodo = (char*) nodo->getFullName(); // Nombre de la BS con la esta conectado el User
		dist_user_bs = (double) red->getSubmodule("user",a)->gate("physicalUser$o")->getChannel()->par(2); // Obtencion de la distancia entre el User y la BS
		del_user_bs = (double) red->getSubmodule("user",a)->gate("physicalUser$o")->getChannel()->par(4); // Obtencion del retardo de propagacion del canal entre User y BS
		//wei_user_bs = (double) top.getNodeFor(red->getSubmodule(nombres_bs[a]))->getLinkOut(0)->getLocalGate()->getChannel()->par(3); // Obtencion del peso del canal entre User y BS
		topo_nodo = top.getNodeFor((cModule*)nodo); // Acceso al nodo del grafo (cTopology) top correspondiente al modulo BS con el que conecta el User
		enlaces = topo_nodo->getNumOutLinks(); // Numero de enlaces salientes del nodo BS con el que conecta el User
		int primero = get_index_first_empty_row(); // Obtencion del indice de la primera fila vacia de las tablas array_de_rutas[][], array_de_longitudes [][], array_de_pesos[][], array_de_retardos[][]
		// Si la BS con la que conecta el User es una pasarela, se graba en las tablas la ruta User->BS->ONU y se pasa al siguiente User
		// (los User que estan conectados a las BS pasarela encaminan su trafico directamente por ellas, aunque se ha decidido que estos User concretos no generen trafico)
		if(strcmp(topo_nodo->getLinkOut(1)->getRemoteNode()->getModule()->getName(),"onu")==0)
		{
			sig_nodo = topo_nodo->getLinkOut(1)->getRemoteNode()->getModule();
			nombre_sig_nodo = (char*)sig_nodo->getFullName();
			distancia = (double) (check_and_cast<cDelayChannel*>(topo_nodo->getLinkOut(1)->getLocalGate()->getChannel()))->par(2);
			retardo = (double) (check_and_cast<cDelayChannel*>(topo_nodo->getLinkOut(1)->getLocalGate()->getChannel()))->par(4);
			//peso = (double) (check_and_cast<cDelayChannel*>(topo_nodo->getLinkOut(1)->getLocalGate()->getChannel()))->par(3);
			peso = 0.0; // Se pone a 0 el peso entre BS y ONU pasarelas (los pesos, que sirven para elegir una ruta optima, solo proceden en los enlaces entre BSs, asi que los enlaces ente User y BS y entre ONu y BS tienen peso nulo).
			// RELLENO DE UNA NUEVA FILA EN LAS TABLAS //
			Tabla->array_de_rutas[primero][0] = nombre_usuario;
			Tabla->array_de_rutas[primero][1] = nombre_nodo;
			Tabla->array_de_rutas[primero][2] = nombre_sig_nodo;
			Tabla->array_de_longitudes[primero][0] = dist_user_bs;
			Tabla->array_de_longitudes[primero][1] = distancia;
			Tabla->array_de_retardos[primero][0] = del_user_bs;
			Tabla->array_de_retardos[primero][1] = retardo;
			Tabla->array_de_pesos[primero][0] = wei_user_bs;
			Tabla->array_de_pesos[primero][1] = peso;
		}
		else
		{
			// Si la BS con la que conecta el User no es una pasarela, se recorren todos los enlaces salientes de este nodo BS para ir incluyendo nuevas rutas para este User en las tablas
			for(int b=1; b<(int)topo_nodo->getNumOutLinks(); b++)
			{
				sig_nodo = topo_nodo->getLinkOut(b)->getRemoteNode()->getModule(); // Acceso remoto al modulo BS con el que conecta la BS asociada al User por el enlace saliente 'b'
				nombre_sig_nodo = (char*)sig_nodo->getFullName(); // Nombre de esta BS, siguiente nodo en la posible nueva ruta
				distancia = (double) (check_and_cast<cDelayChannel*>(topo_nodo->getLinkOut(b)->getLocalGate()->getChannel()))->par(2); // Distancia entre las 2 BSs
				retardo = (double) (check_and_cast<cDelayChannel*>(topo_nodo->getLinkOut(b)->getLocalGate()->getChannel()))->par(4); // Retardo de propagacion en el canal entre las 2 BSs
				peso = (double) (check_and_cast<cDelayChannel*>(topo_nodo->getLinkOut(b)->getLocalGate()->getChannel()))->par(3); // Peso del enlace entre las 2 BSs
				// RELLENO DE UNA NUEVA FILA EN LAS TABLAS //
				Tabla->array_de_rutas[primero+b-1][0] = nombre_usuario;
				Tabla->array_de_rutas[primero+b-1][1] = nombre_nodo;
				Tabla->array_de_rutas[primero+b-1][2] = nombre_sig_nodo;
				Tabla->array_de_longitudes[primero+b-1][0] = dist_user_bs;
				Tabla->array_de_longitudes[primero+b-1][1] = distancia;
				Tabla->array_de_retardos[primero+b-1][0] = del_user_bs;
				Tabla->array_de_retardos[primero+b-1][1] = retardo;
				Tabla->array_de_pesos[primero+b-1][0] = wei_user_bs;
				Tabla->array_de_pesos[primero+b-1][1] = peso;
			}
		}
	}
	// Una vez se han rellenado una serie de filas de las tablas con rutas entre un User, la BS asociada y el siguiente nodo con el que conecta dicha BS,
	// es muy probable que la mayoria de las rutas esten a medias (es decir, que aun no acaben en una ONU pasarela). Por eso, ahora se recorren todas las
	// filas ya rellenadas para intentar completar estas rutas (e incluso incluir nuevas, si procede). Pero este recorrido no se puede hacer una vez,
	// porque en cada iteracion del bucle se incluye un unico nodo mas a cada ruta, luego se necesitan unas cuantas iteraciones para garantizar que todas las
	// rutas se han completado.
	int bucle = 0;
	while(bucle<=7) //Para SFNet: bucle<=15 //Para SFNet_small: bucle<=10 //Para SFNet_small2: bucle<=7
	{
		// Variables auxiliares //
		int c=0, ultimo_nodo=0;
		char *elemento;
		char* nom_onu;
		double dist_onu=0.0, ret_onu=0.0, pes_onu=0.0;
		cModule *bs_intermedia;
		cTopology::Node *top_inter_bs;
		// Se recorren todas las filas de la tabla array_de_rutas[][]:
		while(c<filas)
		{
			// Si esta no es una fila vacia, se comprueba si termina o no en una ONU
			if(strcmp(Tabla->array_de_rutas[c][0]," ")!=0)
			{
				ultimo_nodo = get_index_last_element(c); // Ultimo nodo de la ruta, hasta el momento
				// Si la ruta no termina en una ONU, se accede al modulo BS cuyo nombre aparece en el ultimo nodo (hasta el momento) de la ruta
				// y se intenta completar esta ruta buscando los nodos con los que esta BS intermedia conecta (ONU u otra BS):
				if(strpbrk(Tabla->array_de_rutas[c][ultimo_nodo],"on") == NULL)
				{
					elemento = (char*) Tabla->array_de_rutas[c][ultimo_nodo]; // Nombre del ultimo elemento que aparece en la fila (se sabe que es una BS)
					bs_intermedia = (cModule*) red->getSubmodule(elemento); // Obtencion del modulo al que corresponde el nombre que aparece en la ultima posicion ocupada de la fila 'c'
					/*
					/////// CAMBIO_RED ///////
					int number = -1, cont_aux=0;
					while(cont_aux<estaciones && number<0)
					{
						if(strcmp(elemento,(char*)red->getSubmodule("bs",cont_aux)->getFullName())==0)
						{
							number = cont_aux;
						}
						cont_aux++;
					}
					bs_intermedia = (cModule*) red->getSubmodule("bs",number);
					/////// CAMBIO_RED ///////
					*/
					top_inter_bs = top.getNodeFor(bs_intermedia); // Acceso al nodo del grafo que corresponde al modulo de la BS intermedia
					// Si la "BS intermedia" es una pasarela, se incluye en la fila la info del enlace BS<-->ONU y se pasa a la siguiente fila:
					if(strcmp(top_inter_bs->getLinkOut(1)->getRemoteNode()->getModule()->getName(),"onu")==0)
					{
						nom_onu = (char*) top_inter_bs->getLinkOut(1)->getRemoteNode()->getModule()->getFullName();
						dist_onu = (double) (check_and_cast<cDelayChannel*>(top_inter_bs->getLinkOut(1)->getLocalGate()->getChannel()))->par(2);
						ret_onu = (double) (check_and_cast<cDelayChannel*>(top_inter_bs->getLinkOut(1)->getLocalGate()->getChannel()))->par(4);
						//pes_onu = (double) (check_and_cast<cDelayChannel*>(top_inter_bs->getLinkOut(1)->getLocalGate()->getChannel()))->par(3);
						pes_onu = 0.0;
						Tabla->array_de_rutas[c][ultimo_nodo+1] = nom_onu;
						Tabla->array_de_longitudes[c][ultimo_nodo] = dist_onu;
						Tabla->array_de_retardos[c][ultimo_nodo] = ret_onu;
						Tabla->array_de_pesos[c][ultimo_nodo] = pes_onu;
					}
					// Si la "BS intermedia" no es una pasarela, se recorren los enlaces salientes de este nodo y se intentan incluir nuevas rutas en la tabla
					else if(strcmp(top_inter_bs->getLinkOut(1)->getRemoteNode()->getModule()->getName(),"onu")!=0)
					{
						int primera_vez = 0, indice_auxiliar = 0;
						for(int d=1;d<(int)top_inter_bs->getNumOutLinks();d++)
						{
							cModule *aux_mod = top_inter_bs->getLinkOut(d)->getRemoteNode()->getModule();
							char *aux_nom  = (char*)aux_mod->getFullName();
							if(ya_aparece(aux_nom,c,ultimo_nodo)==0) // Si no aparece ya este elemento en la ruta, se mete
							{
								double dist_aux = (double) (check_and_cast<cDelayChannel*>(top_inter_bs->getLinkOut(d)->getLocalGate()->getChannel()))->par(2);
								double ret_aux = (double) (check_and_cast<cDelayChannel*>(top_inter_bs->getLinkOut(d)->getLocalGate()->getChannel()))->par(4);
								double pes_aux = (double) (check_and_cast<cDelayChannel*>(top_inter_bs->getLinkOut(d)->getLocalGate()->getChannel()))->par(3);
								if(primera_vez == 1)//Si no es la primera nueva ruta que se mete para este nodo:
								{
									indice_auxiliar = get_index_first_empty_row();//Se mete en la primera fila disponible
									Tabla->array_de_rutas[indice_auxiliar] = Tabla->array_de_rutas[c];
									Tabla->array_de_longitudes[indice_auxiliar] = Tabla->array_de_longitudes[c];
									Tabla->array_de_retardos[indice_auxiliar] = Tabla->array_de_retardos[c];
									Tabla->array_de_pesos[indice_auxiliar] = Tabla->array_de_pesos[c];
								}
								else if(primera_vez == 0)//Si es la primera nueva ruta que se mete para este nodo:
								{
									indice_auxiliar = c;//Se completa la ruta en la presente fila
									primera_vez = 1;
								}
								Tabla->array_de_rutas[indice_auxiliar][ultimo_nodo+1] = aux_nom;
								Tabla->array_de_longitudes[indice_auxiliar][ultimo_nodo] = dist_aux;
								Tabla->array_de_retardos[indice_auxiliar][ultimo_nodo] = ret_aux;
								Tabla->array_de_pesos[indice_auxiliar][ultimo_nodo] = pes_aux;
							}
						}
						if(primera_vez==0)
						{// Si no se ha a�adido nada a esta ruta incompleta, se elimina de la tabla
							for(int i=c; i<filas-1; i++)
							{
								Tabla->array_de_rutas[i] = Tabla->array_de_rutas[i+1];
								Tabla->array_de_longitudes[i] = Tabla->array_de_longitudes[i+1];
								Tabla->array_de_retardos[i] = Tabla->array_de_retardos[i+1];
								Tabla->array_de_pesos[i] = Tabla->array_de_pesos[i+1];
							}
							c--; // Si se ha borrado una fila, se evita que se incremente el valor de c en la siguiente iteracion
						}
					}
				}
			}
			c++;
		}
		bucle++;
	}
//// FIN CALCULO RUTAS PARA CADA USUARIO (RELLENO TABLAS) ////
	max_fila = get_index_first_empty_row(); // Obtencion del indice de la ultima fila ocupada en las tablas recien rellenadas

/////// INICIO RELLENO VECTORES DE LONGITUD, RETARDO, PESO Y SALTOS TOTALES DE TODAS LAS RUTAS POSIBLES ///////
	for(int a=0; a<max_fila; a++)
	{
		for(int b=0; b<columnas; b++)
		{
			Tabla->longitud_total_rutas[a] = Tabla->longitud_total_rutas[a] + Tabla->array_de_longitudes[a][b];
		}
	}
	for(int a=0; a<max_fila; a++)
	{
		for(int b=0; b<columnas; b++)
		{
			Tabla->retardo_total_rutas[a] = Tabla->retardo_total_rutas[a] + Tabla->array_de_retardos[a][b];
		}
	}
	for(int a=0; a<max_fila; a++)
	{
		for(int b=0; b<columnas; b++)
		{
			Tabla->peso_total_rutas[a] = Tabla->peso_total_rutas[a] + Tabla->array_de_pesos[a][b];
		}
	}
	for(int a=0; a<max_fila; a++)
	{
		Tabla->saltos_total_rutas[a] = get_index_last_element(a) - 1;
	}
/////// FIN RELLENO VECTORES DE LONGITUD, RETARDO Y SALTOS TOTALES DE TODAS LAS RUTAS POSIBLES ///////

	/*
	if(indice_esta_bs == estaciones-1)
	{
		ev << endl << " CRITERIO SEGUIDO PARA REALIZAR EL ENCAMINAMIENTO DE LOS MENSAJES DE LOS USUARIOS: " << endl;
		switch(criterio)
		{
		case 0:
			ev << " 	SHORTEST-PATH: Minimizar la distancia recorrida de origen a destino" << endl;
			break;
		case 1:
			ev << " 	MINIMUM-DELAY: Minimizar el retardo de propagacion de origen a destino" << endl;
			break;
		case 2:
			ev << " 	MINIMUM-HOP: Minimizar el numero de saltos total" << endl;
			break;
		}
	}
	*/

	// CALCULO DE LA RUTA MAS CORTA PARA CADA USUARIO, SEGUN EL CRITERIO ELEGIDO:
	auxiliar_array.resize(usuarios); // Redimensionamiento de un array y una matriz auxiliares para la siguiente operacion
	auxiliar_matrix.resize(num_pasarelas);
	for(int n=0;n<num_pasarelas;n++)
	{
		auxiliar_matrix[n].resize(usuarios,0.0);
	}
	opt_route.resize(num_pasarelas,-1);
	for(int r=0; r<usuarios; r++)
	{
		get_routes_for_each_user(r);
		switch(criterio)
		{
		case 0:
			calculo_ruta_shortest_path(r);
			break;
		case 1:
			recalculo_ruta_delay_aware(r);
			break;
		case 2:
			calculo_ruta_minimum_hop(r);
			break;
		}
	}

	//auxiliar_array.resize(0);
	for(int n=0;n<num_pasarelas;n++)
	{
		auxiliar_matrix[n].resize(0);
	}
	opt_route.resize(0);

	media_saltos = (double)(media_saltos/usuarios);
	////ev << " Numero medio de saltos en cada ruta elegida como optima: " << media_saltos << endl << endl;

	// REDIMENSIONAMOS LOS ARRAYS QUITANDO LAS FILAS QUE NO HACEN FALTA (ASI SE LIBERA ALGO DE ESPACIO EN MEMORIA):
	Tabla->saltos_total_rutas.resize(max_fila);
	Tabla->longitud_total_rutas.resize(max_fila);
	Tabla->retardo_total_rutas.resize(max_fila);
	Tabla->peso_total_rutas.resize(max_fila);
	Tabla->array_de_rutas.resize(max_fila);
	Tabla->array_de_longitudes.resize(max_fila);
	Tabla->array_de_retardos.resize(max_fila);
	Tabla->array_de_pesos.resize(max_fila);

	// IMPRESION DE LAS TABLAS (SOLO EN LA INICIALIZACION DE LA ULTIMA BS):
	if(indice_esta_bs == estaciones-1)
	{
		/////// INICIO DE LA IMPRESION DE LAS TABLAS ORDENANDO LAS FILAS SEGUN EL USER DE ORIGEN DE CADA RUTA ///////
		/*
		//ev << endl;
		//ev << " **RUTAS COMPLETAS ORDENADAS SEGUN EL USUARIO ORIGEN Y LA BS-PASARELA DESTINO**" << endl;
		for(int g=0; g<usuarios; g++)
		{
			for(int h=0; h<num_pasarelas; h++)
			{
				for(int i=0; i<Tabla->num_rutas[g][h]; i++)
				{
					for(int j=0; j<columnas; j++)
					{
						//ev << " | " << Tabla->array_de_rutas[Tabla->indices_rutas[g][i][h]][j];
					}
					//ev << " --> PESO: " << Tabla->peso_total_rutas[Tabla->indices_rutas[g][i][h]] << endl;
				}
			}
		}
		//ev << endl;
	//}
 //--->
		//// IMPRESION TABLA 'ARRAY_DE_LONGITUDES' (LONGITUD DE CADA ENLACE) ////
		//ev << " TABLA CON LA LONGITUD DE CADA ENLACE DE TODAS LAS RUTAS POSIBLES DESDE CADA USUARIO A UNA ONU" << endl;
		for(int g=0; g<usuarios; g++)
		{
			for(int h=0; h<num_pasarelas; h++)
			{
				for(int i=0; i<Tabla->num_rutas[g][h]; i++)
				{
					for(int j=0; j<columnas; j++)
					{
						//ev << " | " << Tabla->array_de_longitudes[Tabla->indices_rutas[g][i][h]][j];
					}
					//ev << " --> Total: " << Tabla->longitud_total_rutas[Tabla->indices_rutas[g][i][h]] << endl;
				}
			}
		}
		//ev << endl;
	//// IMPRESION TABLA 'ARRAY_DE_RETARDOS' (RETARDO DE CADA ENLACE) ////
		//ev << " TABLA CON EL RETARDO DE CADA ENLACE DE TODAS LAS RUTAS POSIBLES DESDE CADA USUARIO A UNA ONU" << endl;
		for(int g=0; g<usuarios; g++)
		{
			for(int h=0; h<num_pasarelas; h++)
			{
				for(int i=0; i<Tabla->num_rutas[g][h]; i++)
				{
					for(int j=0; j<columnas; j++)
					{
						//ev << " | " << Tabla->array_de_retardos[Tabla->indices_rutas[g][i][h]][j];
					}
					//ev << " --> Total: " << Tabla->retardo_total_rutas[Tabla->indices_rutas[g][i][h]] << endl;
				}
			}
		}
		//ev << endl;
	//// IMPRESION TABLA 'ARRAY_DE_PESOS' (PESO DE CADA ENLACE) ////
		//ev << " TABLA CON EL PESO DE CADA ENLACE DE TODAS LAS RUTAS POSIBLES DESDE CADA USUARIO A UNA ONU" << endl;
		for(int g=0; g<usuarios; g++)
		{
			for(int h=0; h<num_pasarelas; h++)
			{
				for(int i=0; i<Tabla->num_rutas[g][h]; i++)
				{
					for(int j=0; j<columnas; j++)
					{
						//ev << " | " << Tabla->array_de_pesos[Tabla->indices_rutas[g][i][h]][j];
					}
					//ev << " --> Total: " << Tabla->peso_total_rutas[Tabla->indices_rutas[g][i][h]] << endl;
				}
			}
		}
		//ev << endl;
		 */
		/////// FIN DE LA IMPRESION DE LAS TABLAS ORDENANDO LAS FILAS SEGUN EL USER DE ORIGEN DE CADA RUTA ///////

		/////// INICIO DE LA IMPRESION DE LAS TABLAS SIN NINGUN ORDEN CONCRETO, TAL COMO SE HAN RELLENADO ///////
		/*
			//// IMPRESION TABLA 'ARRAY_DE_RUTAS' (NOMBRES DE LOS NODOS) ////
		//ev << endl;
		//ev << " TABLA CON LOS NODOS INTERMEDIOS DE TODAS LAS RUTAS POSIBLES DESDE CADA USUARIO A UNA ONU" << endl;
		for(int a=0; a<max_fila; a++)
		{
			for(int b=0; b<columnas; b++)
			{
				//ev << " | " << Tabla->array_de_rutas[a][b];
			}
			//ev << " Total saltos: " << Tabla->saltos_total_rutas[a] << endl;
		}
		//ev << endl;

		//// IMPRESION TABLA 'ARRAY_DE_LONGITUDES' (LONGITUD DE CADA ENLACE) ////
		//ev << " TABLA CON LA LONGITUD DE CADA ENLACE DE TODAS LAS RUTAS POSIBLES DESDE CADA USUARIO A UNA ONU" << endl;
		for(int a=0; a<max_fila; a++)
		{
			for(int b=0; b<columnas; b++)
			{
				//ev << " | " << Tabla->array_de_longitudes[a][b];
			}
			//ev << " --> Total: " << Tabla->longitud_total_rutas[a] << endl;
		}
		//ev << endl;
	//// IMPRESION TABLA 'ARRAY_DE_RETARDOS' (RETARDO DE CADA ENLACE) ////
		//ev << " TABLA CON EL RETARDO DE CADA ENLACE DE TODAS LAS RUTAS POSIBLES DESDE CADA USUARIO A UNA ONU" << endl;
		for(int a=0; a<max_fila; a++)
		{
			for(int b=0; b<columnas; b++)
			{
				//ev << " | " << Tabla->array_de_retardos[a][b];
			}
			//ev << " --> Total: " << Tabla->retardo_total_rutas[a] << endl;
		}
		//ev << endl;
	//// IMPRESION TABLA 'ARRAY_DE_PESOS' (PESO DE CADA ENLACE) ////
		//ev << " TABLA CON EL PESO DE CADA ENLACE DE TODAS LAS RUTAS POSIBLES DESDE CADA USUARIO A UNA ONU" << endl;
		for(int a=0; a<max_fila; a++)
		{
			for(int b=0; b<columnas; b++)
			{
				//ev << " | " << Tabla->array_de_pesos[a][b];
			}
			//ev << " --> Total: " << Tabla->peso_total_rutas[a] << endl;
		}
		//ev << endl;
		/////// FIN IMPRESION TABLAS ///////
		*/
		/////// FIN DE LA IMPRESION DE LAS TABLAS SIN NINGUN ORDEN CONCRETO, TAL COMO SE HAN RELLENADO ///////
	}

	// VACIAMOS LOS ARRAYS CUYA INFORMACION NO VA A HACER MAS FALTA EN LA SIMULACION:
	//Tabla->saltos_total_rutas.clear();
	Tabla->longitud_total_rutas.clear();
	Tabla->retardo_total_rutas.clear();
	Tabla->array_de_longitudes.clear();
	Tabla->array_de_retardos.clear();
	//Tabla->array_de_pesos.clear();
	//Tabla->peso_total_rutas.clear();
}
