////////////////////////////////////////////////////////////////////////////////////
// PROYECTO FIN DE CARRERA															  											 	//
// TITULO: Disenyo de nuevas arquitecturas y algoritmos de gestin de recursos en	//
//		  	 redes de acceso FiWi  												  												//
// AUTOR: Ana Emma Lopez Mato																											//
// TUTOR: Noemi Merayo Alvarez																										//
// INGENIERIA TECNICA DE TELECOMUNICACIONES, SISTEMAS DE TELECOMUNICACION					//
// UNIVERSIDAD DE VALLADOLID													      											//
////////////////////////////////////////////////////////////////////////////////////

#ifndef __RED_WIRELESS_BS_MAC_H_
#define __RED_WIRELESS_BS_MAC_H_

#include <omnetpp.h>
#include <vector>
#include "BS_table.h"
#include "BS_queue.h"
#include "DATA_m.h"
/*
 * MODULO BS_mac:
 * Elabora la tabla de encaminamiento para todos los usuarios wireless, estableciendo la
 * ruta optima en funcion de un criterio elegido en el fichero omnetpp.ini, de entre los
 * tres posibles:
 * 0->Camino de menor longitud, 1->Camino de menor retardo, 2->Camino de menos saltos.
 * Asimismo, gestiona la multiplexacion en tiempo de la capacidad de cada enlace de radio
 * saliente del modulo BS (cada enlace correspondiente a un submodulo BS_queue).
 */
typedef std::vector<int> int_vector_t; // VECTOR DE TIPO ENTERO
typedef std::vector<cModule *> module_vector_t; // VECTOR DE TIPO cMODULE
typedef std::vector<char*> string_vector_t; // VECTOR DE TIPO CADENA DE CARACTERES
typedef std::vector<simtime_t> time_vector_t; // VECTOR DE TIPO CADENA DE CARACTERES
typedef std::vector<double> double_vector_t; // VECTOR DE TIPO DOUBLE
typedef std::vector< double_vector_t > double_matrix_t; // MATRIZ DE TIPO ENTERO

class BS_mac : public cSimpleModule
{
  protected:
    virtual void initialize();
    virtual void handleMessage(cMessage *msg);
    virtual void finish();
    virtual int get_index_first_empty_row();
    virtual int is_an_empty_row(int row);
    virtual int get_index_last_element(int row);
    virtual int ya_aparece(char* nombre_nodo, int fila, int indice);
    virtual int es_pasarela(cModule *node);
    virtual int conseguir_siguiente_puerta(int init_ind, int puertas, cModule *node);
    virtual int puerta_entrada_libre (cModule *node, int puertas);
    virtual int puerta_salida_libre (cModule *node, int puertas);
    virtual void calculo_ruta_shortest_path(int ind_usr);
    virtual void calculo_ruta_minimum_hop(int ind_usr);
    virtual void get_routes_for_each_user(int index);

  public:
    FILE * retardo_wireless;
    Analysis retardo;
    int ind_primero;
    int max_fila;
    int filas;
    int columnas;
    int used_queues;
    int prioridades;
    int esPasarela;
    int enlaces_intermedios;
    int extraction_method;
		int nodos_actualizados;
		int actualizaciones;
    double t_frame;
    double t_slot;
    double txrate;
    double rxrate;
    double eff_cap;
    double C_i;
    double media_saltos;
    long sumcolas;
    simtime_t time_frame;
    BS_table *Tabla;
    BS_queue *cola;
    BS_queue *queue;
    int_vector_t opt_route;
    int_vector_t index_used_queues;
    string_vector_t nombres_bs;
    int_vector_t indices_pasarelas;
		cTopology top;
		cTopology::Node *top_node;
		simtime_t timepacket;
		time_vector_t slot_tx_delay;
		double_vector_t auxiliar_array;
		double_matrix_t auxiliar_matrix;
		virtual void get_num_used_queues();
		virtual void extract_packets();
		virtual void relleno_arrays_retardos();
    virtual void recalculo_ruta_delay_aware(int ind_usr);
    virtual void routing();
};

#endif
