////////////////////////////////////////////////////////////////////////////////////
// PROYECTO FIN DE CARRERA															  											 	//
// TITULO: Disenyo de nuevas arquitecturas y algoritmos de gesti�n de recursos en	//
//		  	 redes de acceso FiWi  												  												//
// AUTOR: Ana Emma Lopez Mato																											//
// TUTOR: Noemi Merayo Alvarez																										//
// INGENIERIA TECNICA DE TELECOMUNICACIONES, SISTEMAS DE TELECOMUNICACION					//
// UNIVERSIDAD DE VALLADOLID													      											//
////////////////////////////////////////////////////////////////////////////////////
#include "BS_queue.h"
#include "BS_table.h"
#include "DATA_m.h"
#include <string.h>
/*
 * MODULO BS_queue:
 * Almacena en sus subcolas (tantas como prioridades de servicio) los paquetes cuyo
 * proximo destino es el que se indica en el parametro 'target', para luego transmitirlos
 * cuando se tiene el testigo, siguiendo el metodo de extraccion de paquetes de prioridad
 * estricta.
 * La gestion del paso del testigo se realiza de forma centralizada con el m�dulo BS_MAC.
 * En cada m�dulo compuesto BS hay tantos m�dulos BS_queue como BSs hay en la red.
 */
Define_Module(BS_queue)
;

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION INITIALIZE()--> ESTA FUNCION SE INVOCA DESPUES DE QUE OMNET++ HA PUESTO EN MARCHA LA RED, EN LA CUAL SE LEEN LOS      //
//						  PARIMETROS DEL MODULO Y SE INICIALIZAN TODAS DAS LAS VARIABLES DECLARADAS PARA ESTE MODULO SIMPLE, SE //
//						  ASIGNAN ESTRUCTURAS DE DATOS DINIMICOS Y SE ASIGNAN E INICIALIZAN LOS AUTOMENSAJES SI SON NECESARIOS  //
//						  PARA EL FUNCIONAMIENTO DE ESTE MODULO.			   													//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void BS_queue::initialize()
{
	// Se incluyen en la topologia los nodos que contienen el parametro 'aux' (Users, BSs y ONUs):
	top.extractByParameter("aux");
	// Se extrae el nodo correspondiente al modulo compuesto que contiene este submodulo:
	cModule *origen = (cModule*) this->getParentModule();
	top_node = top.getNodeFor(origen);

	//// DEFINIMOS EL PROCESO PARA PODER ENTRAR DESDE ESTE MODULO AL MODULO 'BS_table' ////
	Tabla = check_and_cast<BS_table *> (getParentModule()->getSubmodule("bs_table"));

	// Obtencion parametros del fichero omnetpp.ini:
	//int indice = (int) this->getParentModule()->par("indice"); // Indice de la BS
	//int indice = (int) this->getParentModule()->getIndex(); // Indice de la BS /////// CAMBIO_RED ///////
	int conexiones = (int) this->getParentModule()->par("conexiones"); // Numero de puertas de salida de la BS
	txrate = (double) this->par("tasabinaria"); // Tasa de transmision wireless (bps)
	prioridades = (int) this->getParentModule()->par("numcolas"); // Numero de colas (clases de servicio)

	// Inicializacion de variables:
	int links = top_node->getNumOutLinks(); // Total conexiones con usuario/onu/BSs
	int link_2 = links-1; //Total conexiones onu/BSs
	colas_salida = link_2; // Numero de posibles modulos BS_queue[] en uso
	colas_totales = conexiones-1; // Numero de modulos BS_queue que contiene en total cada BS
	used_bs_queues = 0;
	paquetes_recibidos = 0; // Cantidad de paquetes recibidos
	volcado = 0; // Variable auxiliar para entrar en la funcion VolcadoFicheros()
	contador = 0; // Variable auxiliar para gestionar la inactividad de una BS_queue
	remaining_bw = 0; // Ancho de banda sobrante en cada slot
	bw_left = 0; // Ancho de banda que aun queda para transmitir en cada slot
	bw_tx = 0; // Ancho de banda que se ha transmitido en un slot
	bw_tx_anterior = 1; // Ancho de banda que se ha transmitido en el solt anterior
	time_tx = 0; // Tiempo de slot consumido para transmitir
	time_resto = 0; // Tiempo de slot sobrante
	queue.resize(prioridades, 0); // Tantas subcolas como prioridades hay
	user_tamqueuepop.resize(prioridades, 0); // Longitud de los paquetes que se borran de cada cola
	user_tamqueueextract.resize(prioridades, 0); // Longitud de los paquetes que se checkean antes de sacarlos de cada cola
	user_tamextract.resize(prioridades, 0); // Longitud de los paquetes que se extraen de cada cola
	tamqueue.resize(prioridades, 0); // Longitud de cada una de las colas (inicialmente, 0 Bytes)
	sumcheckedpacket = 0; // Suma de los paquetes checkeados en el metodo de extraccion con prioridad estricta
	sumextractedpacket = 0; // Suma de los paquetes extraidos en el metodo de extraccion con prioridad estricta
	tamsumcola = 0; // Suma de la longitud de los paquetes que llegan a las colas de la BS_queue
	sumpaquete.resize(prioridades, 0); // Suma de los paquetes checkeados en las colas
	sumcolapop.resize(prioridades, 0); // Suma total de bytes extraidos de las colas
	bytes_cola.resize(prioridades, 0); //Bytes que quedan en las colas tras cada ciclo
	Testigo = 0; // Variable para la gestion de los tiempos de slot (TDMA)
	Interval = 0.0; // Intervalo de transmision permitido en cada slot (tiepo actual + tiempo de slot)

	// Si el enlace num.1 conecta con una ONU, es una pasarela:
	if(strcmp(top_node->getLinkOut(1)->getRemoteNode()->getModule()->getName(),"onu")==0)
	{
		esPasarela = 1;
		link_2 = link_2-prioridades;
		if(this->getIndex() == 0)
		{
			//ev << " La estacion base " << origen->getFullName() << " es una pasarela con la parte optica." << endl;

		}
	}
	else
	{
		esPasarela = 0;
		if (this->getIndex() == 0)
		{
			//ev << " La estacion base " << origen->getFullName() << " no es una pasarela con la parte optica." << endl;
		}
	}

	if (this->getIndex() == 0)
	{
		// Numero de interfaces de salida del BS que estan en uso:
		//ev << " Numero total de conexiones con otras estaciones base: " << link_2 << endl;
		//ev << " Total prioridades de servicio de los usuarios: " << prioridades << endl;
		// Cuantas conexiones externas con BSs/usuarios/ONU hay?
		//ev << " **CONEXIONES DE LA ESTACION BASE n." << indice << "** " << endl;
	}
	/*
	for (int j = 0; j < links; j++)
	{
		// NOTA: Indices del array de parametros de los canales inalambricos:
		//		 par(0): isEnabled (bool) --> POR DEFECTO
		//		 par(1): delay (s) --> NULO
		//		 par(2): longitud (double)
		//		 par(3): peso (double) --> Inicialmente 0.0, cambia con DARA.
		//		 par(4): retardo_propagacion (double) --> longitud/3�10e8.
		top_node->getLinkOut(j);
		if (this->getIndex() == 0)
		{
			char* nombre_nodo = (char*) top_node->getLinkOut(j)->getRemoteNode()->getModule()->getFullName();
			char* nombre_canal = (char*) (check_and_cast<cDelayChannel*>(top_node->getLinkOut(j)->getLocalGate()->getChannel()))->getComponentType()->getName();
			char* nombre_longitud = (char*) (check_and_cast<cDelayChannel*>(top_node->getLinkOut(j)->getLocalGate()->getChannel()))->par(2).getFullName();
			double longitud = (double) (check_and_cast<cDelayChannel*>(top_node->getLinkOut(j)->getLocalGate()->getChannel()))->par(2);
			double retardo = (double) (check_and_cast<cDelayChannel*>(top_node->getLinkOut(j)->getLocalGate()->getChannel()))->par(4);
			ev << " CANAL '" << nombre_canal << "': " << origen->getFullName() << "<-->" << nombre_nodo << endl;
			// Visualizacion
			ev << "  Longitud ('" << nombre_longitud << "'): " << longitud << " metros" << endl;
			ev << "  Retardo de propagacion: " << retardo << " segundos." << endl;
		}
	}
	*/
	// Si es una BS_queue[] que puede contener paquetes destinados a alguna BS/onu,
	// se asigna el target, esto es, el nombre de la BS/ONU destino de los paquetes
	// que va a almacenar:
	if (this->getIndex() < colas_salida)
	{
		int j = 1, fin = -1;
		char* nombre_cola;
		this->setTarget(" "); // Objetivo en blanco

		while (j < links && fin < 0)
		{
			nombre_cola = (char*) top_node->getLinkOut(j)->getRemoteNode()->getModule()->getFullName();
			if (this->getIndex() == 0 || se_repite(nombre_cola) == 0)
			{
				// Si es el modulo BS_queue[0] o no aparece ya este nodo como Target de otro modulo BS_queue[] de esta BS,
				// se asigna como Target a este modulo BS_queue[] el nombre del nodo al otro extremo del enlace con el que
				// se asocia este modulo BS_queue[]:
				this->setTarget(nombre_cola);
				fin = 1;
			}
			j++;
		}
		// Si al final del bucle, a este modulo BS_queue[] no se le ha asignado Target, entonces
		// es que no se va a usar:
		if (strcmp(this->getTarget(), " ") == 0)
		{
			//ev << " *COLA EN DESUSO* " << endl;
			uso = 0; // Esta cola no se va a usar nunca porque no se asocia con nigun enlace de salida de esta BS
			usado = 0; // Esta cola no esta activa
		}
		else
		{
			//ev << " Cola BS para recibir los mensajes destinados al nodo " << this->getTarget() << endl;
			uso = 1; // Esta cola puede ser usada porque corresponde a un enlace entre esta BS y otro nodo (este parametro siempre va a estar a 1)
			usado = 1; // Inicialmente se considera que esta cola esta siendo usada (si luego resulta que no recibe paquetes, puede ponerse a cero esta variable y considerarse cola no activa).
			if(esPasarela == 1 && getIndex()!=0)
			{
				usado = 0; // Si es una pasarela, en realidad solo se usa la cola que conecta con la ONU.
			}
		}
	}
	else
	{
		//ev << " *COLA EN DESUSO* " << endl;
		uso = 0;
		usado = 0;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION HANDLEMESSAGE(CMESSAGE *MSG)--> ESTA FUNCION SE INVOCA CON EL MENSAJE COMO PARAMETRO CADA VEZ QUE EL MODULO RECIBE UN //
//										  UN MENSAJE O PAQUETE. SE RECORRE EL CODIGO IMPLEMENTADO PARA DEVOLVER UN VALOR O  	//
//										  EJECUTAR UNA FUNCION DENTRO DEL MODULO SIMPLE. EL TIEMPO DE SIMULACION NO TRANSCURRE  //
//										  DENTRO DE LA LLAMADA DE LA FUNCION HANDLEMESSAGE(CMESSAGE *MSG) MIENTRAS RECORRE EL	//
//										  CODIGO IMPLEMENTADO.																	//                                                                         //
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void BS_queue::handleMessage(cMessage *msg) {
	int type = msg->getKind();
	switch (type)
	{
		case 12: // Paquete 'Datos' (mensajes de usuarios wireless)
			if(type==12)
			{
				DATA *msjdatos = check_and_cast<DATA*>(msg);
				if(uso == 0)
				{
					//ev << "ERROR!!!" << endl;
					delete msjdatos;
				}
				if(usado == 0 && uso == 1)
				{
					usado = 1;
					contador = 0;
				}

				////ev << " INSERCION del paquete DATA con Id. " << msjdatos->getId() << " en la cola de prioridad P" << msjdatos->getPriority() << endl;
				// Introducimos el paquete en la cola de prioridad correspondiente:
				tamqueue[msjdatos->getPriority()] = tamqueue[msjdatos->getPriority()] + msjdatos->getByteLength();
				msjdatos->setTime_enter_queue(simTime());
				queue[msjdatos->getPriority()].insert(msjdatos);
			}
		break;
		default:
			delete msg;
			//ev << " MENSAJE BORRADO" << endl;
			break;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION GETTARGET()--> DEVUELVE EL VALOR DEL PARAMETRO TARGET, VARIABLE GLOBAL DE BS_QUEUE /
//////////////////////////////////////////////////////////////////////////////////////////////
const char * BS_queue::getTarget() const {
	return target.c_str();
}
//////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION SETTARGET(CONST CHAR * TARGET)--> MODIFICA EL VALOR DE TARGET, VARIABLE GLOBAL DE BS_QUEUE /
//////////////////////////////////////////////////////////////////////////////////////////////////////
void BS_queue::setTarget(const char * target) {
	this->target = target;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION SEREPITE(CHAR* NOMBRE)--> COMPRUEBA SI LA CADENA 'nombre' COINCIDE CON EL TARGET DE ALGUN MODULO  /
//									BS_QUEUE ANTERIOR. DEVUELVE 0 SI ES LA PRIMERA VEZ QUE SE USA 'nombre'  /
//									Y 1 EN CASO CONTRARIO.													/
/////////////////////////////////////////////////////////////////////////////////////////////////////////////
int BS_queue::se_repite(char* nombre) {
	int k = 0, repetido = 0;
	while (k < this->getIndex() && repetido == 0)
	{
		BS_queue *cola_previa;
		cola_previa = check_and_cast<BS_queue*>(this->getParentModule()->getSubmodule("bs_queue", k));
		if (strcmp(nombre, cola_previa->getTarget()) == 0)
		{
			repetido = 1;
		}
		k++;
	}
	return repetido;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION STRICTPRIORITYQUEUE()--> METODO DE EXTRACCION DE PAQUETES EN LAS COLAS TENIENDO EN CUENTA LA PRIORIDAD. /														/
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void BS_queue::strictpriorityqueue() {
	//ev << " **Metodo de extraccion de paquetes en colas con prioridad estricta**" << endl;
	// Se recorren todas las colas 'cQueue' para ver si tienen paquetes y extraer los Bytes que se puedan:
	int i = 0, salir = -1; //
	while (i < prioridades && salir < 0) {
		// Se comprueba que la cola i tiene contiene paquetes y se compara el
		// ancho de banda total con la suma de todos los paqutes extraidos:
		if (queue[i].getLength() > 0 && bandwidth >= sumcheckedpacket) {
			// ... Si el BW es mayor que los bytes extraidos Y quedan paquetes en la cola i, se extraen paquetes:
			int acabar = -1; //
			while (queue[i].getLength() > 0 && acabar < 0) {
				// Llamada a la funcion 'chequeoPaquete()' para ver si podemos extraerlo:
				chequeoPaquete(i);
				//ev << " 	Suma de los paquetes chequeados: " << sumcheckedpacket << " Bytes" << endl;
				// Se comprueba el ancho de banda de la cola con la suma de los paquetes extraidos (y este ultimo):
				if (bandwidth >= sumcheckedpacket) {
					// ... si el BW es mayor, se extrae un paquete 'DATA' y se envia:
					//ev << " 	EXTRACCION y ENVIO de un paquete inalambrico de prioridad P" << i << " al siguiente nodo: " << this->getTarget() << endl;
					// Llamada a la funcion para extraer paquetes:
					envioPaquete(i);
				} else if (bandwidth < sumcheckedpacket) {
					// ... Si el BW es menor, no se extraen mas paquetes:
					//ev << " 	NO se extraen mas paquetes porque no entra en el ancho de banda de slot de la cola P" << i << endl;
					acabar = 1; //
					salir = 1; //
					// Salida del bucle que recorre los paquetes de la cola i
				}
			}
		} else if (bandwidth < sumcheckedpacket)
		{
			// ... Si el BW es menor que los bytes extraidos:
			tamsumcola = tamsumcola - sumextractedpacket; // RESTAMOS EL TAMANYO DE LOS PAQUETES QUE HEMOS EXTRAIDO A LA SUMA DEL TAMANYO DE LOS PAQUETES DE LAS COLAS
			//ev << " 	NO se envian mas paquetes porque con el ultimo que se ha checkeado se excede el ancho de banda disponible(" << bandwidth << ")." << endl;
			salir = 1; //
		}
		i++;
	}
	// Si se han recorrido todas las colas, y aun queda BW para transmitir:
	if (bandwidth >= sumcheckedpacket)
	{
		// Se restan los bytes de los paquetes extraidos a la suma de bytes en las colas:
		tamsumcola = tamsumcola - sumextractedpacket;
		//ev << " 	NO se envian mas paquetes porque las colas estan vacias." << endl;
	}
}

/////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION CENTRALIZEDMETHOD()--> METODO CENTRALIZADO DE EXTRACCION DE PAQUETES EN LAS COLAS.	/														/
/////////////////////////////////////////////////////////////////////////////////////////////////
void BS_queue::centralizedmethod() {
	//ev << " **Metodo centralizado de extraccion de paquetes de las colas**" << endl;
	double capacidad_real = (double) bandwidth/prioridades;
	// Se recorren todas las colas 'cQueue' para ver si tienen paquetes y extraer los Bytes que se puedan:
	int i = 0, salir = -1; //
	while (i < prioridades && salir < 0)
	{
		sumpaquete[i] = 0; // Suma de los paquetes checkeados en las colas
		sumcolapop[i] = 0; // Suma de los bytes extraidos de las colas
		// Se comprueba que la cola i tiene contiene paquetes y se compara el
		// ancho de banda total con la suma de todos los paqutes extraidos:
		if (queue[i].getLength() > 0 && capacidad_real >= sumpaquete[i])
		{
			// ... Si el BW es mayor que los bytes extraidos Y quedan paquetes en la cola i, se extraen paquetes:
			int acabar = -1; //
			while (queue[i].getLength() > 0 && acabar < 0)
			{
				// Llamada a la funcion 'chequeoPaquete()' para ver si podemos extraerlo:
				chequeoPaquete(i);
				//ev << " 	Suma de los paquetes chequeados en la cola P" << i << ": " << sumpaquete[i] << " Bytes" << endl;
				// Se comprueba el ancho de banda de la cola con la suma de los paquetes extraidos (y este ultimo):
				if (capacidad_real >= sumpaquete[i])
				{
					// ... si el BW es mayor, se extrae un paquete 'DATA' y se envia:
					//ev << " 	EXTRACCION y ENVIO de un paquete inalambrico de prioridad P" << i << " al siguiente nodo: " << this->getTarget() << endl;
					// Llamada a la funcion para extraer paquetes:
					envioPaquete(i);
				}
				else if (capacidad_real < sumpaquete[i])
				{
					// ... Si el BW es menor, no se extraen mas paquetes:
					//ev << " 	NO se extraen mas paquetes porque no entra en el ancho de banda de slot de la cola P" << i << endl;
					acabar = 1;
					// Salida del bucle que recorre los paquetes de la cola i
				}
			}
		}
		else if (capacidad_real < sumpaquete[i])
		{
			// ... Si el BW es menor que los bytes extraidos:
			tamsumcola = tamsumcola - sumextractedpacket; // RESTAMOS EL TAMANYO DE LOS PAQUETES QUE HEMOS EXTRAIDO A LA SUMA DEL TAMANYO DE LOS PAQUETES DE LAS COLAS
			//ev << " 	NO se envian mas paquetes porque con el ultimo que se ha checkeado se excede el ancho de banda disponible para esta subcola (" << capacidad_real << ")." << endl;
			salir = 1; //
		}
		i++;
	}
	// Si se han recorrido todas las colas, y aun queda BW para transmitir:
	if (bandwidth >= sumcheckedpacket)
	{
		// Se restan los bytes de los paquetes extraidos a la suma de bytes en las colas:
		tamsumcola = tamsumcola - sumextractedpacket;
		//ev << " 	NO se envian mas paquetes porque las colas estan vacias." << endl;
	}
	else if (bandwidth < sumcheckedpacket)
	{
		tamsumcola = tamsumcola - sumextractedpacket; // RESTAMOS EL TAMANYO DE LOS PAQUETES QUE HEMOS EXTRAIDO A LA SUMA DEL TAMANYO DE LOS PAQUETES DE LAS COLAS
		//ev << " 	NO se envian mas paquetes porque con el ultimo que se ha checkeado se excede el ancho de banda disponible(" << bandwidth << ")." << endl;
	}
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION DELETEELEMENT(CMESSAGE *MSG, INT PRIORITY)--> ELIMINA UN PAQUETE DE LA COLA CON PRIORIDAD 'priority'.	/														/
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void BS_queue::deleteElement(cMessage *msg, int priority) {
	//ETHERNETmsg *msgdata = check_and_cast<ETHERNETmsg*>(msg);
	DATA *msgdata = check_and_cast<DATA*> (msg);
	// Extraccion del primer paquete ETHERNET de la cola de la prioridad indicada:
	//msgdata = (ETHERNETmsg *)queue[priority].pop();
	msgdata = (DATA *) queue[priority].pop();
	// Obtencion de la longitud del paquete extraido:
	user_tamqueuepop[priority] = (long) msgdata->getByteLength();
	// Resta de los bytes del paquete borrado al total de bytes de esta cola:
	tamqueue[priority] = tamqueue[priority] - user_tamqueuepop[priority];
	// Borrado del paquete 'DATA':
	//ev << " MENSAJE 'DATA' (Id." << msgdata->getId() << ") BORRADO" << endl;
	delete msgdata;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION CHEQUEOPAQUETE(INT INDEX)--> CHEQUEA LOS PAQUETES AL PRINCIPIO DE LA COLA DE PRIORIDAD 'index'	/
//									   PARA SABER SU LONGITUD EN BYTES.										/
/////////////////////////////////////////////////////////////////////////////////////////////////////////////
void BS_queue::chequeoPaquete(int index)
{
	//ETHERNETmsg *msgdata;
	DATA *msgdata;
	// Obtencion de informacion del primer paquete de la cola 'index', sin extraerlo:
	//msgdata = (ETHERNETmsg *)queue[index].front();
	msgdata = (DATA *) queue[index].front();
	// Almacenamiento en la posicion 'index' de la variable 'user_tamqueueextract' de la longitud del paquete:
	user_tamqueueextract[index] = (long) msgdata->getByteLength();
	// Suma de los paquetes chequeados:
	//sumcheckedpacket = sumcheckedpacket + user_tamqueueextract[index];
	sumcheckedpacket = sumcheckedpacket + (long) msgdata->getByteLength();
	sumpaquete[index] = sumpaquete[index] + (long) msgdata->getByteLength();
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION ENVIOPAQUETE(INT PRIORITY)--> EXTRAE UN PAQUETE DEL PRINCIPIO DE LA COLA DE PRIORIDAD 'priority'. /
/////////////////////////////////////////////////////////////////////////////////////////////////////////////
void BS_queue::envioPaquete(int priority)
{
	//ETHERNETmsg *msgdata;
	DATA *msgdata;
	int num_puertas = (int) this->getParentModule()->gateSize("physicalBS");
	double distancia = 0, Ddistancia = 0;

	// Extraccion del primer paquete de la cola de indice 'priority':
	take(msgdata = (DATA *) queue[priority].pop());
	////ev << " 	Identificador del paquete: " << msgdata->getId() << endl;
	////ev << " 	Longitud del paquete: " << msgdata->getByteLength() << " Bytes" << endl;
	// Almacenamiento en la posicion 'priority' de la variable 'user_tamextract' de la longitud del paquete:
	user_tamextract[priority] = msgdata->getByteLength();
	// Almacenamiento en la posicion 'priority' de la variable 'user_tamqueuepop' de la longitud del paquete:
	user_tamqueuepop[priority] = msgdata->getByteLength();
	// Resta de la longitud del paquete extraido a la longitud de la cola 'priority':
	tamqueue[priority] = tamqueue[priority] - msgdata->getByteLength();
	// Suma de la longitud del paquete a los paquetes extraidos de esta cola:
	sumextractedpacket = sumextractedpacket + msgdata->getByteLength();
	////ev << " 	Suma de los bytes extraidos en total en este slot: " << sumextractedpacket << " Bytes" << endl;
	// Suma total de los bytes de los paquetes extraidos de la cola i, y visualizacion:
	sumcolapop[priority] = sumcolapop[priority] + msgdata->getByteLength();
	////ev << " 	Suma de los bytes extraidos de la cola P" << priority << " en total: " << sumcolapop[priority] << " Bytes" << endl;
	// Actualizacion del ancho de banda disponible durante este slot:
	bw_left = bw_left - msgdata->getByteLength();
	////ev << " 	Durante este slot, aun pueden transmitirse " << bw_left << " Bytes" << endl;

	// Busqueda del destino de este paquete 'DATA':
	char* nombre_puerta;
	cModule *destino;
	int x = 0, indice = -1;
	while (x < num_puertas && indice < 0) {
		if (this->getParentModule()->gate("physicalBS$o", x)->isConnected())
		{
			nombre_puerta = (char*) this->getParentModule()->gate("physicalBS$o", x)->getNextGate()->getOwnerModule()->getFullName();
			if (strcmp(this->getTarget(), nombre_puerta) == 0)
			{
				destino = this->getParentModule()->gate("physicalBS$o", x)->getNextGate()->getOwnerModule();
				indice = x;
			}
		}
		x++;
	}
	msgdata->setOrigen(this->getParentModule()->getFullName());
	distancia = (double) this->getParentModule()->gate("physicalBS$o", indice)->getChannel()->par(2);
	msgdata->setDestino(destino->getFullName());
	if(strcmp(destino->getModuleType()->getName(),"BS")==0)
	{
		msgdata->setDestAddress(destino->par("indice"));
		//msgdata->setDestAddress(destino->getIndex()); /////// CAMBIO_RED ///////
	}
	else
	{
		msgdata->setDestAddress(destino->getIndex());
	}
	Ddistancia = distancia + (double) msgdata->getMetros();
	msgdata->setMetros(Ddistancia);
	int saltos = msgdata->getSaltos() + 1;
	msgdata->setSaltos(saltos);
	send(msgdata, "bsqueueInOut$o");
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION VOLCADOFICHEROS()--> ESCRIBE EN UNO O VARIOS FICHEROS DE TEXTO LOS DATOS DE MAYOR INTERES DE LA	/
//							   SIMULACION, PARA SU POSTERIOR ANALISIS.										/
/////////////////////////////////////////////////////////////////////////////////////////////////////////////
void BS_queue::volcadoFicheros() {
	volcado = 1;
	if (usado == 1) // Si es una cola que esta siendo usada, esto es, se considera activa:
	{
		// 'LONGITUD_COLAS': NUMERO TOTAL DE PAQUETES PRESENTES EN CADA SUBMODULO BS_QUEUE[], ASOCIADO A UN ENLACE SALIENTE ACTIVO, DE CADA BS.
		Longitud_colas = fopen("results/Longitud_colas.txt", "a+");
		for (int p=0; p<prioridades; p++)
		{
			fprintf(Longitud_colas, "%s.BS_queue[%i] -> Paquetes: %li \t Bytes: %li\n", this->getParentModule()->getFullName(), this->getIndex(), (long int) queue[p].getLength(), (long int) tamqueue[p]);
		}
		fclose(Longitud_colas);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION FINISH()--> ESTA FUNCION SE INVOCA CUANDO LA SIMULACION HA TERMINADO CON EXITO SIN QUE SE PRODUZCA NINGUN ERROR. 		//
//					  LO USAMOS PARA LA RECOGIDA DE ESTADASTICAS Y VISUALIZACION POR PANTALLA O MEDIANTE UN ARCHIVO.            //
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void BS_queue::finish() {
	volcadoFicheros();
}
