////////////////////////////////////////////////////////////////////////////////////
// PROYECTO FIN DE CARRERA															  											 	//
// TITULO: Disenyo de nuevas arquitecturas y algoritmos de gestin de recursos en	//
//		  	 redes de acceso FiWi  												  												//
// AUTOR: Ana Emma Lopez Mato																											//
// TUTOR: Noemi Merayo Alvarez																										//
// INGENIERIA TECNICA DE TELECOMUNICACIONES, SISTEMAS DE TELECOMUNICACION					//
// UNIVERSIDAD DE VALLADOLID													      											//
////////////////////////////////////////////////////////////////////////////////////

#ifndef __RED_WIRELESS_BS_QUEUE_H_
#define __RED_WIRELESS_BS_QUEUE_H_

#include <omnetpp.h>
#include <vector>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <tr1/memory>
#include "analysis.h"
#include "BS_table.h"
#include "DATA_m.h"
/*
 * MODULO BS_queue:
 * Almacena en sus subcolas (tantas como prioridades de servicio) los paquetes cuyo
 * proximo destino es el que se indica en el parametro 'target', para luego transmitirlos
 * cuando se tiene el testigo, siguiendo el metodo de extraccion de paquetes de prioridad
 * estricta.
 */
typedef std::vector<int> int_vector_t; // VECTOR DE TIPO ENTERO
typedef std::vector<long> long_vector_t; // VECTOR DE TIPO ENTERO LARGO
typedef std::vector<cQueue> cQueue_vector_t; // VECTOR DE TIPO COLA
class BS_queue : public cSimpleModule
{
  private:
	FILE * Parametros_slot_BSqueue;
	FILE * Longitud_colas;

  protected:
	opp_string target;
    virtual void initialize();
    virtual void handleMessage(cMessage *msg);
    virtual int se_repite(char* nombre);
    virtual void finish();

  public:
    virtual const char * getTarget() const;
    virtual void setTarget(const char * target);
    int esPasarela; // Es esta BS una pasarela optico-wireless?
    int uso; // Puede usarse esta cola?
    int usado; // Se usa realmente esta cola?
    int Testigo;
    simtime_t Interval;
    int colas_salida;
    int colas_totales;
    int prioridades;
    double txrate;
    double t_slot;
    int used_index;
    simtime_t tiempo_slot;
    simtime_t timepacket;
    double t_frame;
    simtime_t tiempo_frame;
    int used_bs_queues;
	BS_table *Tabla;
	cQueue_vector_t queue; // Array de cubcolas cQueue, una por prioridad
	cTopology top;
	cTopology::Node *top_node;

	// Ancho de banda (en Bytes) que se permite transmitir en cada posesion de testigo:
	double bandwidth;
	// Ancho de banda (en Bytes) que no se puede aprovechar en cada posesion de testigo:
	double remaining_bw;
	// Ancho de banda (en Bytes) que aun se puede transmitir durante un slot:
	double bw_left;
	// Ancho de banda (en Bytes) que se ha transmitido durante un slot:
	double bw_tx;
	// Ancho de banda (en Bytes) que se ha transmitido durante el slot anterior:
	double bw_tx_anterior;
	// Tiempo (segundos) que se ha usado para transmitir durante un slot:
	double time_tx;
	// Tiempo (segundos) que NO se ha aprovechado para transmitir durante un slot:
	double time_resto;

	long paquetes_recibidos;
	long paquetes_colas;
	long_vector_t tamqueue; // Longitud de cada cola
	long_vector_t user_tamqueuepop; // Longitud de los paquetes borrados de las colas
	long_vector_t user_tamqueueextract; // Longitud de los paquetes checkeados previa extracci�n de las colas
	long_vector_t user_tamextract; // Longitud de los paquetes extraidos de las colas

	long sumcheckedpacket; // Suma de los paquetes CHECKEADOS en el metodo de extraccion con prioridad estricta
	long sumextractedpacket; // Suma de los paquetes EXTRAIDOS en el metodo de extraccion con prioridad estricta
	long tamsumcola; // Suma de la longitud de los paquetes LLEGADOS a las colas de la BS_queue
	long_vector_t sumpaquete; // Suma de los paquetes checkeados en las colas
	long_vector_t sumcolapop; // Suma total de bytes extraidos de las colas
	long_vector_t bytes_cola; //Bytes que quedan en las colas tras cada ciclo
	double_vector_t media; // Media
	// Funcion para eliminar un paquete del inicio de una cola cQueue:
	virtual void deleteElement(cMessage *msg, int priority);
	// Metodos de extraccion de paquetes de los usuarios desde las colas de las BSs:
	virtual void strictpriorityqueue();
	// Metodos de extraccion de paquetes de los usuarios desde las colas de las BSs:
	virtual void centralizedmethod();
    // Funcion para chequear los paquetes del principio de cada cola para saber su longitud en Bytes:
	void chequeoPaquete(int index);
	// Funcion para extraer los paquetes del principio de cada cola:
	void envioPaquete(int priority);
	virtual void volcadoFicheros();
	int volcado;
	int contador;
};

#endif
