////////////////////////////////////////////////////////////////////////////////////
// PROYECTO FIN DE CARRERA															  											 	//
// TITULO: Disenyo de nuevas arquitecturas y algoritmos de gesti�n de recursos en	//
//		  	 redes de acceso FiWi  												  												//
// AUTOR: Ana Emma Lopez Mato																											//
// TUTOR: Noemi Merayo Alvarez																										//
// INGENIERIA TECNICA DE TELECOMUNICACIONES, SISTEMAS DE TELECOMUNICACION					//
// UNIVERSIDAD DE VALLADOLID													      											//
////////////////////////////////////////////////////////////////////////////////////

#include <omnetpp.h>
#include <vector>
#include <time.h>
#include <string.h>
#include "BS_rx_tx.h"
#include "BS_table.h"
#include "BS_queue.h"
#include "BS_mac.h"
#include "LSA_m.h"
#include "DATA_m.h"
/*
 * MODULO BS_rx_tx:
 * Recibe los mensajes procedentes de modulos externos (usuarios/BSs) para enviarlos al
 * submodulo BS_queue correspondiente, asi como envia hacia el modulo externo pertinente
 * los mensajes de los submodulos BS_queue.
 */
Define_Module(BS_rx_tx);

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION INITIALIZE()--> ESTA FUNCION SE INVOCA DESPUES DE QUE OMNET++ HA PUESTO EN MARCHA LA RED, EN LA CUAL SE LEEN LOS      //
//						  PARIMETROS DEL MODULO Y SE INICIALIZAN TODAS DAS LAS VARIABLES DECLARADAS PARA ESTE MODULO SIMPLE, SE //
//						  ASIGNAN ESTRUCTURAS DE DATOS DINIMICOS Y SE ASIGNAN E INICIALIZAN LOS AUTOMENSAJES SI SON NECESARIOS  //
//						  PARA EL FUNCIONAMIENTO DE ESTE MODULO.			   													//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void BS_rx_tx::initialize()
{
	// INICIALIZACION DE LAS VARIABLES:
	indice_esta_bs = (int)this->getParentModule()->par("indice");
	//indice_esta_bs = (int)this->getParentModule()->getIndex(); /////// CAMBIO_RED ///////
	retardo_mensaje = 0; // Retardo con el que llega cada paquete a este modulo
	umbral_retardo = 0.025; // En el paper de DARA, se fija el retardo maximo de un paquete en 25ms.
	paquetes_borrados_dara = 0; // Acumula el numero de paquetes borrados por superar el umbral de retardo al implementar DARA
	aux_counter = 0; // Variable auxiliar para seguir la llegada de 100000 paquetes a cada BS
	full_bs_queues = 0; // Acumula el numero de BSs que ya han recibido 100000 paquetes (util para parar la simulacion cuando han llegado 100000 paquetes a todas las BSs, o a las BSs pasarela)
	paquetes_recibidos = 0; // Acumula el total de paquetes recibidos por esta BS
	bytes_recibidos = 0; // Acumula el total de bytes recibidos por esta BS
	tx_rate = (double) this->getParentModule()->par("tasabinaria"); // Tasa de transmision de la BS
	BSs = (int) par("numBS"); // Numero de BSs de la red
	pasarelas = (int) par("numGW"); // Numero de BSs-pasarela
	prioridades = (int) this->getParentModule()->par("numcolas"); // Numero de prioridades en el trafico inalambrico
	control = check_and_cast<BS_mac *>(getParentModule()->getSubmodule("bs_mac")); // Acceso remoto al modulo BS_mac de esta BS
	mean_link_load = 0.0; // Media de la carga de recpecion de paquetes de cada enlace (normalizada a la capacidad de enlace)
	mean_alt_link_load = 0.0; // Media de la carga de recepcion de paquetes de enlace (normalizada a la tasa binaria)
	mean_node_load = 0.0; // Media de la carga de transmision de paquetes de cada enlace (normalizada a la capacidad de enlace)
	mean_alt_node_load = 0.0; // Media de la carga de transmision de paquetes de cada enlace (normalizada a la tasa binaria)

	/////// Topologia de la red ///////
	// Se incluyen en la topologia los nodos que contienen el parametro 'aux' (Users, BSs y ONUs):
	top.extractByParameter("aux");
	// Si esta bien, debe ser: t.getNumNodes() = numBS+numOnu
	// Se define como nodo del grafo la BS a la que pertenece esta cola:
	nodeOut = top.getNodeFor(getParentModule());
	if(strcmp(nodeOut->getLinkOut(1)->getRemoteNode()->getModule()->getName(),"onu")==0)
	{
		esPasarela=1;
	}
	else
	{
		esPasarela=0;
	}
	links = (int)nodeOut->getNumOutLinks(); // Total de enlaces salientes de esta BS
	bs_links = links-1;
	if(esPasarela == 1)
	{
		// Si esta BS es una pasarela, al numero conexiones con otras BSs hay que restarle el numero de conexiones con la ONU pasarela
		// (tantas como prioridades) para que represente realmente el numero de BSs con las que conecta la pasarela
		bs_links = bs_links-prioridades;
		// Si esta es una BS pasarela, la tasa de transmision de paquetes hacia la ONU pasarela es la de un enlace optico:
		tx_rate = (double) this->getParentModule()->par("txrate");
		tx_rate = tx_rate/10; // 100 Mbps, como las ONUs
	}

	// Impresion por pantalla de las conexiones salientes de esta BS:
	//ev << " Numero de enlaces salientes de este nodo: "<< links << endl;
	//ev << " 	De los cuales conectan con otras BSs: "<< bs_links << endl;
	/*for(int a=0; a<links; a++)
	{
		ev << " 	Enlace: " << a
				<< " --> Puerta interna: " << nodeOut->getLinkOut(a)->getLocalGate()->getOwnerModule()->gate(nodeOut->getLinkOut(a)->getLocalGate()->getName(),nodeOut->getLinkOut(a)->getLocalGate()->getIndex())->getPreviousGate()->getFullName()
				<< " --> Puerta externa: " << nodeOut->getLinkOut(a)->getLocalGate()->getFullName()
				<< " --> Puerta remota externa: " << nodeOut->getLinkOut(a)->getRemoteGate()->getFullName()
				<< " --> Nodo: " << nodeOut->getLinkOut(a)->getRemoteNode()->getModule()->getFullName() << endl;
	}*/
	total_enlaces = 0; // Numero total de enlaces entre BSs de la red (el valor autentico de esta variable es e lque se obtiene al inicializar la ultima BS, por eso al hacer finish se da ese valor obtenido por la ultima BS a todas las demas)
	for(int i=0; i<BSs; i++)
	{
		total_enlaces = total_enlaces + check_and_cast<BS_rx_tx *>(this->getParentModule()->getParentModule()->getSubmodule(control->nombres_bs[i])->getSubmodule("bs_rx_tx"))->bs_links;
		//total_enlaces = total_enlaces + check_and_cast<BS_rx_tx *>(this->getParentModule()->getParentModule()->getSubmodule("bs",i)->getSubmodule("bs_rx_tx"))->bs_links; /////// CAMBIO_RED ///////
	}
	////ev << " Total enlaces entre nodos de la red mallada: " << total_enlaces << endl;

	/*
	//ev << " Numero de nodos de la topologia: " << top.getNumNodes() << endl;
	for(int b=0; b<(int)top.getNumNodes(); b++)
	{
		//ev << " Nodo de indice " << b << " en la topologia: " << top.getNode(b)->getModule()->getFullName() << endl;
	}
	*/
	// REDIMENSIONAMIENTO DE LOS VECTORES:
	retardo_medio.resize(links); // Retardo medio de cada enlace de esta BS
	arrived_packets.resize(links, 0); // Paquetes recibidos por cada enlace de esta BS (se resetea en cada periodo LSA para recalcular la tasa de llegadas en cada periodo)
	arrived_bytes.resize(links, 0); // Bytes recibidos por cada enlace de esta BS (sirve para calcular la carga de los enlaces al hacer finish)
	sent_packets.resize(links, 0); // Paquetes enviados por cada enlace de esta BS
	sent_bytes.resize(links, 0); // Bytes enviados por cada enlace de esta BS (sirve para calcular la carga de los enlaces al hacer finish)
	link_load.resize(links, 0); // Carga de rx de paquetes de cada enlace de esta BS, normalizada a la capacidad de enlace
	alt_link_load.resize(links, 0); // Carga de rx de paquetes de cada enlace de esta BS, normalizada a la tasa de transmision
	node_load.resize(links, 0); // Carga de rx de paquetes de cada enlace de esta BS, normalizada a la capacidad de enlace
	alt_node_load.resize(links, 0); // Carga de rx de paquetes de cada enlace de esta BS, normalizada a la tasa de transmision
	int long_paq = (int) this->getParentModule()->par("longitud_media_paquete");
	int bits_paquete = 8*long_paq;
	service_rate = (double) 1/bits_paquete; // Tasa de servicio de los nodos
	effective_capacity = (double)tx_rate/bs_links; // Capacidad efectiva (capacidad de canal entre enlaces salientes a otras BSs)
	////ev << " Capacidad efectiva de los enlaces de este nodo: " << effective_capacity << " bps" << endl;
	////ev << " Tasa de servicio de la red inalambrica: " << service_rate << endl;

	// SEGUN EL CRITERIO DE ENCAMINAMIENTO ELEGIDO EN OMNETPP.INI...
	int criterio = (int)this->getParentModule()->par("criterio");
	if(criterio == 1) // SI SE CONSIDERA CAMINO OPTIMO EL DE MENOR RETARDO (DARA):
	{
		packet_intensity.resize(links, 0);
		previous_estimated_lambda.resize(links, 0);
		aux_samples.resize(links,0);
		//gateway_arrival_rate.resize(pasarelas,0);
		estimated_lambda = 0;

		// Envio de automensaje para crear el primer LSA:
		LSA_period = (simtime_t) 7.5; // Periodos entre LSAs recogidos en el paper: 7.5s, 15s, 30s, 60s.
		//ev << " Periodo entre envios de LSAs: " << SIMTIME_DBL(LSA_period) << endl;
		//cMessage *msg = new cMessage ("Creacion LSA",3); /////// ACTIVAR ///////
		//scheduleAt(LSA_period,msg); /////// ACTIVAR ///////
		contador_LSAs = 0;
		prev_LSAs = 0;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION HANDLEMESSAGE(CMESSAGE *MSG)--> ESTA FUNCION SE INVOCA CON EL MENSAJE COMO PARAMETRO CADA VEZ QUE EL MODULO RECIBE UN //
//										  UN MENSAJE O PAQUETE. SE RECORRE EL CODIGO IMPLEMENTADO PARA DEVOLVER UN VALOR O  	//
//										  EJECUTAR UNA FUNCION DENTRO DEL MODULO SIMPLE. EL TIEMPO DE SIMULACION NO TRANSCURRE  //
//										  DENTRO DE LA LLAMADA DE LA FUNCION HANDLEMESSAGE(CMESSAGE *MSG) MIENTRAS RECORRE EL	//
//										  CODIGO IMPLEMENTADO.																	//                                                                         //
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void BS_rx_tx::handleMessage(cMessage *msg)
{
	int type = msg->getKind();
	switch(type)
	{
		case 3:
			// LLEGA UN AUTOMENSAJE QUE INDICA UN NUEVO ENVIO DE LSA, CON IDENTIFICADOR = 3
			if(type==3)
			{
				// PASOS A SEGUIR:
				// 1. Se calculan los parametros a meter en el LSA (con la funcion 'calculate_weight_params()').
				// 2. Se crea el LSA y se insertan los datos.
				// 3. Se env�a el LSA por todas las puertas conectadas con BSs.
				//ev << " Total mensajes LSA recibidos en el periodo anterior: " << prev_LSAs << endl;
				// 1:
				int first_link_index = links - bs_links;
				for(int i=first_link_index; i<links; i++) // Se descuentan los enlaces que no son entre BSs.
				{
					if(strpbrk(nodeOut->getLinkOut(i)->getRemoteNode()->getModule()->getFullName(),"b")!=NULL)
					{
						calculate_weight_params(i);
					}
				}
				/*
				// Si esta BS es una pasarela, calculo lambda_j (tasa de llegadas de paquetes a la pasarela j):
				// NOTA: Enlace pasarela -> 1 (el primero es para el usuario asociado)
				if(strcmp(nodeOut->getLinkOut(1)->getRemoteNode()->getModule()->getName(),"onu")==0)
				{
					gateway_arrival_rate[indice_esta_bs] = packet_intensity[1];
				}
				// NOTA: En el paper de DARA, lambda_j es la tasa de llegadas a la ONU pasarela (downstream)
				*/
				// 2:
				LSA *advertisement = new LSA("Link_State_Advertisement",9);
				advertisement->setSrcAddress(indice_esta_bs);
				advertisement->setSrcName(this->getParentModule()->getFullName());
				advertisement->setNum_links(links);
				advertisement->setRemote_gatesArraySize(links);
				advertisement->setRemote_nodesArraySize(links);
				advertisement->setArrival_rateArraySize(links);
				advertisement->setRcvd_pcktsArraySize(links);
				advertisement->setLink_loadArraySize(links);
				advertisement->setEffective_link_capacity((double)control->eff_cap); // eff_cap; //
				advertisement->setService_rate((double)service_rate);
				advertisement->setTimestamp(simTime());
				for(int i=0; i<links; i++)
				{
					advertisement->setRemote_nodes(i,nodeOut->getLinkOut(i)->getRemoteNode()->getModule()->getFullName());
					advertisement->setRemote_gates(i,(int)nodeOut->getLinkOut(i)->getRemoteGate()->getIndex());
					// NOTA: Si el extremo remoto es un usuario, la puerta es unica, y el indice que devuelve es 0.
					advertisement->setArrival_rate(i,(double)packet_intensity[i]);
					advertisement->setLink_load(i,(double)link_load[i]);
					advertisement->setRcvd_pckts(i,arrived_packets[i]);
				}
				// VACIAMOS LOS VECTORES CADA PERIODO LSA:
				packet_intensity.clear();
				link_load.clear();
				arrived_packets.clear();

				// 3:
				for(int i=first_link_index; i<links; i++)
				{
					//if(strpbrk(advertisement->getRemote_nodes(i),"b")!=NULL){
						//ev << " Envio LSA al nodo " << advertisement->getRemote_nodes(i) << endl; //ev-comentar
						////ev << " por la puerta " << nodeOut->getLinkOut(i)->getLocalGate()->getOwnerModule()->gate(nodeOut->getLinkOut(i)->getLocalGate()->getName(),nodeOut->getLinkOut(i)->getLocalGate()->getIndex())->getPreviousGate()->getFullName() << endl;
						LSA *nuevo = check_and_cast<LSA *>(advertisement->dup());
						send(nuevo,nodeOut->getLinkOut(i)->getLocalGate()->getOwnerModule()->gate(nodeOut->getLinkOut(i)->getLocalGate()->getName(),nodeOut->getLinkOut(i)->getLocalGate()->getIndex())->getPreviousGate());
						//}
				}
				scheduleAt(simTime()+LSA_period, msg);
				delete advertisement;
			}
			break;

		case 9:
			// LLEGA UN LSA DE OTRO NODO, CON IDENTIFICADOR = 9
			if(type==9)
			{
				if(esPasarela == 0)
				{
					contador_LSAs++;
					//ev << " Mensajes LSA recibidos hasta el momento en este periodo: " << contador_LSAs << endl;
					// PASOS A SEGUIR:
					// 1. Se extraen los parametros del enlace por el que ha entrado el LSA (�nicamente).
					// 2. Se calcula el peso de ese enlace.
					// 3. Se actualiza el retardo del canal asociado a ese enlace, con el peso.
					// 4. Se actualiza la tabla de retardos de cada ruta y el vector de retardos totales.
					// 5. Se actualiza la tabla de rutas optimas para cada usuario.

					// 1:
					LSA *anuncio = check_and_cast<LSA*>(msg);
					//ev << " Longitud del mensaje LSA: " << anuncio->getByteLength() << " Bytes." << endl;
					// Busqueda del indice del enlace en el sentido Emisor_LSA-->Receptor_LSA:
					int indice=-1;
					for(int i=1; i<(int)anuncio->getNum_links(); i++)
					{
						////ev << " Nodo que se conecta con " << anuncio->getSrcName() << " por el enlace " << i << ": " << anuncio->getRemote_nodes(i) << endl;
						if(strcmp(anuncio->getRemote_nodes(i),this->getParentModule()->getFullName())==0)
						{
							indice = i;
						}
					}
					// Busqueda del indice del enlace en el sentido Receptor_LSA-->Emisor_LSA:
					int index=-1;
					for(int i=1; i<links; i++)
					{
						char* name = (char*) nodeOut->getLinkOut(i)->getRemoteNode()->getModule()->getFullName();
						////ev << " Nodo que se conecta con esta BS por el indice " << i << ": " << name << endl;
						if(strcmp(name,anuncio->getSrcName())==0)
						{
							index = i;
						}
					}
					// 2-3:
					if(indice!=-1)
					{
						calculate_delay_weight(indice, index, anuncio);
					}
					// �OJO! Como los dos siguientes pasos pueden suponer mucha carga computacional,
					//		 realizarlos solo cuando llega el ultimo LSA esperado.
					if(contador_LSAs==bs_links)
					{
						check_and_cast<BS_mac *>(this->getParentModule()->getParentModule()->getSubmodule("bs0")->getSubmodule("bs_mac"))->nodos_actualizados++;
						//check_and_cast<BS_mac *>(this->getParentModule()->getParentModule()->getSubmodule("bs",0)->getSubmodule("bs_mac"))->nodos_actualizados++; /////// CAMBIO_RED ///////
						int nodos = BSs - pasarelas;
						BS_mac *control_aux;
						if(check_and_cast<BS_mac *>(this->getParentModule()->getParentModule()->getSubmodule("bs0")->getSubmodule("bs_mac"))->nodos_actualizados == nodos)
						//if(check_and_cast<BS_mac *>(this->getParentModule()->getParentModule()->getSubmodule("bs",0)->getSubmodule("bs_mac"))->nodos_actualizados == nodos) /////// CAMBIO_RED ///////
						{
							for(int n=0; n<BSs; n++)
							{
								control_aux = check_and_cast<BS_mac *>(this->getParentModule()->getParentModule()->getSubmodule(control->nombres_bs[n])->getSubmodule("bs_mac"));
								//control_aux = check_and_cast<BS_mac *>(this->getParentModule()->getParentModule()->getSubmodule("bs",n)->getSubmodule("bs_mac")); /////// CAMBIO_RED ///////
								// 4:
								control_aux->relleno_arrays_retardos();
								for(int i=0; i<(int)par("numUser"); i++)
								{
									// 5:
									control_aux->recalculo_ruta_delay_aware(i);
								}
							}
							check_and_cast<BS_mac *>(this->getParentModule()->getParentModule()->getSubmodule("bs0")->getSubmodule("bs_mac"))->nodos_actualizados = 0;
							//check_and_cast<BS_mac *>(this->getParentModule()->getParentModule()->getSubmodule("bs",0)->getSubmodule("bs_mac"))->nodos_actualizados = 0; /////// CAMBIO_RED ///////
						}
						prev_LSAs = contador_LSAs;
						contador_LSAs = 0;
					}
				}
				// Si es pasarela, como no necesita enviar trafico a otras BSs, lo del encaminamiento wireless le da un poco igual,
				// asi que borra los LSAs que recibe (pero es importante que los reciba, y que los envie)
				else if(esPasarela == 1)
				{
					delete msg;
				}
			}
			break;

		case 12:
			// LLEGA UN PAQUETE 'DATA' CON IDENTIFICADOR = 12
			if(type==12)
			{
				DATA *msjdatos = check_and_cast<DATA*>(msg);
				cModule *destino;
				/*if(simTime() >= 200) // SI LLEVAMOS 200s DE SIMULACION...
				{
					endSimulation(); // LLAMAMOS A LA FUNCION TERMINAR SIMULACION
					callFinish(); // LLAMAMOS A LA FUNCION FINALIZAR
				}*/
				// Si esta BS que recibe el paquete es una pasarela, este se borra (o se envia a la ONU pasarela, segun se configure la red),
				// previo calculo y almacenamiento de informacion referente al nuevo paquete:
				if(esPasarela == 1)
				{
					// Si el mensaje viene de uno de los modulos BS_queue[] de esta BS, hacia el exterior:
					if(strcmp(msjdatos->getOrigen(),this->getParentModule()->getFullName())==0)
					{
						msjdatos->setTime_exit_bs(simTime());
						//ev << " Envio de paquete inalambrico, de prioridad P" << msjdatos->getPriority() << " e Id." << msjdatos->getId() << endl;
						//ev << " 	Longitud del mensaje: " << msjdatos->getByteLength() << "B" << endl;

						simtime_t retardo = simTime() - msjdatos->getTimestamp();
						retardo_mensaje = SIMTIME_DBL(retardo);
						//ev << " Retardo del mensaje destinado al nodo " << msjdatos->getDestino() << ": " << retardo_mensaje << "s" << endl;
						//simtime_t d_queue = simTime() - msjdatos->getTime_enter_queue();
						//ev << " 	Retardo de encolamiento: " << d_queue << "s" << endl;

						double hops = (double) msjdatos->getSaltos();
						saltos.analyze(hops);
						double meters = (double) msjdatos->getMetros();
						longitud.analyze(meters);

						//int link = get_index_link(1);
						int link = 1;
						sent_packets[link]++;
						////ev << " Total paquetes enviados: " << sent_packets[link] << endl;
						sent_bytes[link] = sent_bytes[link] + msjdatos->getByteLength();
						// Conexion directa con la ONU:
						destino = (cModule*) nodeOut->getLinkOut(1)->getRemoteNode()->getModule();
						//delete msjdatos; // BORRADO DEL PAQUETE

						// Envio del paquete al modulo rx/tx,considerando el retardo de transmision:
						double datarate = (double) msjdatos->getByteLength()*8/tx_rate; // CALCULAMOS EL PARAMETRO datarate QUE INDICA LA TASA DE TRANSMISION DEL PAQUETE INALAMBRICO
						simtime_t tasa = (simtime_t)datarate;
						control->timepacket = control->timepacket + tasa;
						//ev << " 	Retardo de transmision: " << control->timepacket << "s" << endl;
						int puerta = 1 + msjdatos->getPriority(); // IMPORTANTISIMO  SI numcolas > 1
						sendDelayed(msjdatos, control->timepacket, "bsRxTx$o", puerta);
						// NOTA: Tal como esta configurada la red en el fichero *.ned, el enlace de indice 1
						//			 de toda BS pasarela conecta con la cola de prioridad P0 de la ONU pasarela asociada.
						//			 Ahora bien, si el numero de clases de servicio en la parte inalambrica es mayor que 1,
						//			 hay que asegurarse de que el trafico se envia a la cola adecuada de la ONU:
						//			 Si msjdatos tiene prioridad Pn, la BS pasarela ha de enviarlo al modulo ON_SISTqueue[n],
						//			 con el que coencta por la puerta 1+n.

						/*
						double tiempo_slot = (double) check_and_cast<BS_queue *>(msjdatos->getSenderGate()->getOwnerModule())->t_slot;
						int ind_cola = (int) check_and_cast<BS_queue *>(msjdatos->getSenderGate()->getOwnerModule())->used_index;
						double offset = (double) tiempo_slot*ind_cola;
						offset_time = (simtime_t) offset;
						//ev << " 	Retardo de slot: " << offset_time << "s" << endl;
						control->slot_tx_delay[ind_cola] = control->slot_tx_delay[ind_cola] + tasa;
						//ev << " 	Retardo de transmision: " << control->slot_tx_delay[ind_cola] << "s" << endl;
						sendDelayed(msjdatos, offset_time + control->slot_tx_delay[ind_cola], "bsRxTx$o", 1);
						*/
					}
					// Si el mensaje viene de fuera, hacia alguno de los modulos BS_queue[] de esta BS:
					else if(strcmp(msjdatos->getOrigen(),this->getParentModule()->getFullName())!=0)
					{
						paquetes_recibidos++;
						bytes_recibidos = bytes_recibidos + msjdatos->getByteLength();
						//ev << " Paquetes recibidos en esta BS: " << paquetes_recibidos << endl;
						/*
						// PARA PARAR LA SIMULACION CUANDO LLEGAN 100000 PAQUETES A CADA UNA DE LAS BSs PASARELA:
						if(paquetes_recibidos>=100000 && aux_counter==0)
						{
							aux_counter = 1;
							full_bs_queues = 1;
						}
						double total_paquetes = 0;
						for(int i=0; i<(int)control->indices_pasarelas.size(); i++)
						{
							total_paquetes = total_paquetes + check_and_cast<BS_rx_tx *>(this->getParentModule()->getParentModule()->getSubmodule(control->nombres_bs[control->indices_pasarelas[i]])->getSubmodule("bs_rx_tx"))->full_bs_queues;
						}
						if(total_paquetes == (int)control->indices_pasarelas.size())
						{
							endSimulation(); // LLAMAMOS A LA FUNCION TERMINAR SIMULACION
							callFinish(); // LLAMAMOS A LA FUNCION FINALIZAR
						}
						*/
						/*
						// PARA PARAR LA SIMULACION CUANDO LLEGAN 100000 PAQUETES A LA BS PASARELA 'bs7'
						// (CUANDO SOLO SE GENERAN PAQUETES ENTRE bs0 Y bs7, ruta de mayor retardo):
						if(indice_esta_bs==5 && paquetes_recibidos>=100000)
						{
							endSimulation(); // LLAMAMOS A LA FUNCION TERMINAR SIMULACION
							callFinish(); // LLAMAMOS A LA FUNCION FINALIZAR
						}
						*/
						/*
						// PARA PARAR LA SIMULACION CUANDO LLEGAN 100000 PAQUETES A CADA UNA DE LAS BSs DE LA RED:
						if(paquetes_recibidos >=100000 && aux_counter==0)
						{
							aux_counter=1;
							full_bs_queues = 1;
						}
						double total_paquetes = 0;
						for(int i=0; i<BSs; i++)
						{
							total_paquetes = total_paquetes + check_and_cast<BS_rx_tx *>(this->getParentModule()->getParentModule()->getSubmodule(control->nombres_bs[i])->getSubmodule("bs_rx_tx"))->full_bs_queues;
						}
						if(total_paquetes == BSs)
						{
							endSimulation(); // LLAMAMOS A LA FUNCION TERMINAR SIMULACION
							callFinish(); // LLAMAMOS A LA FUNCION FINALIZAR
						}*/
						/*
						// PARA PARAR LA SIMULACION CUANDO LLEGAN 100000 PAQUETES EN TOTAL A LAS BSs PASARELA:
						double total_paquetes = 0;
						for(int i=0; i<(int)control->indices_pasarelas.size(); i++)
						{
							total_paquetes = total_paquetes + check_and_cast<BS_rx_tx *>(this->getParentModule()->getParentModule()->getSubmodule(control->nombres_bs[control->indices_pasarelas[i]])->getSubmodule("bs_rx_tx"))->paquetes_recibidos;
						}
						*/
						//ev << " Recepcion de un paquete inalambrico, de prioridad P" << msjdatos->getPriority() << endl;
						simtime_t retardo = simTime() - msjdatos->getTimestamp();
						retardo_mensaje = SIMTIME_DBL(retardo);
						//ev << " Retardo total del mensaje procedente del nodo " << msjdatos->getOrigen() << ": " << retardo_mensaje << "s" << endl;
						retardo_pasarela.analyze(retardo_mensaje);
						(check_and_cast<BS_mac *>(this->getParentModule()->getParentModule()->getSubmodule("bs0")->getSubmodule("bs_mac")))->retardo.analyze(retardo_mensaje);
						//(check_and_cast<BS_mac *>(this->getParentModule()->getParentModule()->getSubmodule("bs",0)->getSubmodule("bs_mac")))->retardo.analyze(retardo_mensaje); /////// CAMBIO_RED ///////

						//double tasa_aux = (double) this->getParentModule()->par("tasabinaria");
						double tasa_aux = (double) check_and_cast<BS_mac *>(msjdatos->getSenderModule()->getParentModule()->getSubmodule("bs_mac"))->txrate;
						carga = (double) (bytes_recibidos * 8 / (SIMTIME_DBL(simTime()) * tasa_aux));
						media_carga.analyze((double)carga);
						//ev << " Carga de este nodo pasarela: " << carga << endl;

						double hops = (double) msjdatos->getSaltos();
						//ev << " Nodos intermedios (BSs) por los que ha pasado el paquete: " << hops << endl;

						double meters = (double) msjdatos->getMetros();
						//ev << " Distancia (metros) recorrida por el paquete: " << meters << endl;

						saltos.analyze(hops);
						longitud.analyze(meters);
						/*
						if(total_paquetes >= 100000)
						{
							endSimulation(); // LLAMAMOS A LA FUNCION TERMINAR SIMULACION
							callFinish(); // LLAMAMOS A LA FUNCION FINALIZAR
						}
						*/
						////ev << " Mensaje recibido por la puerta " << msjdatos->getArrivalGate()->getFullPath() << endl;
						// Indice, en el grafo, del enlace por el que ha llegado el paquete:
						//int link = get_index_link(msjdatos->getArrivalGate()->getIndex());
						int link = msjdatos->getArrivalGate()->getIndex();
						////ev << " Correspondiente al enlace " << link << " del grafo" << endl;

						// Se cambia almacena el retardo del ultimo paquete para calcular el retardo medio del enlace por el que vino:
						retardo_medio[link].analyze(retardo_mensaje);
						arrived_packets[link]++;
						arrived_bytes[link] = arrived_bytes[link] + msjdatos->getByteLength();
						//delete msjdatos; // BORRADO DEL PAQUETE
						send(msjdatos,"queueRxTx$o",0); // ALMACENAMIENTO DEL PAQUETE EN LA COLA CORRESPONDIENTE, PARA SU POSTERIOR ENVIO A LA ONU PASARELA
					}
				}
				// Si esta BS que recibe el paquete no es una pasarela...
				else if(esPasarela == 0)
				{
					char* nombre_origen = (char*) msjdatos->getOrigen();
					char* nombre_destino = (char*) msjdatos->getDestino();
					// Si el mensaje viene de un modulo BS_queue[] de esta BS, hacia otra BS:
					if(strcmp(nombre_origen,this->getParentModule()->getFullName())==0)
					{
						msjdatos->setTime_exit_bs(simTime());
						//ev << " Envio de paquete inalambrico, de prioridad P" << msjdatos->getPriority() << " e Id." << msjdatos->getId() << endl;
						simtime_t retardo = simTime() - msjdatos->getTimestamp();
						retardo_mensaje = SIMTIME_DBL(retardo);
						//ev << " Retardo del mensaje destinado al nodo " << msjdatos->getDestino() << ": " << retardo_mensaje << "s" << endl;
						//simtime_t d_queue = simTime() - msjdatos->getTime_enter_queue();
						//ev << " 	Retardo de encolamiento: " << d_queue << "s" << endl;

						destino = this->getParentModule()->getParentModule()->getSubmodule(nombre_destino);
						//destino = this->getParentModule()->getParentModule()->getSubmodule("bs", msjdatos->getDestAddress()); /////// CAMBIO_RED ///////

						// INICIO BUSQUEDA INDICE PUERTA DE SALIDA:
						int gate_size = (int) this->gateSize("bsRxTx");
						int i=0;
						encontrado = -1;
						while(i<gate_size && encontrado<0)
						{
							if(this->gate("bsRxTx$o",i)->getNextGate()->isConnected())
							{
								cModule *nodo;
								nodo = (cModule*) this->gate("bsRxTx$o",i)->getNextGate()->getNextGate()->getOwnerModule();
								if(strcmp(destino->getFullName(),nodo->getFullName())==0)
								{
									encontrado = i;
								}
							}
							i++;
						}
						// FIN BUSQUEDA INDICE PUERTA DE SALIDA.

						//ev << " 	Longitud del mensaje: " << msjdatos->getByteLength() << "B" << endl;
						//int link = get_index_link(encontrado);
						int link = encontrado;
						sent_packets[link]++;
						////ev << " Total paquetes enviados: " << sent_packets[link] << endl;
						sent_bytes[link] = sent_bytes[link] + msjdatos->getByteLength();

						// Envio del paquete al modulo rx/tx,considerando el retardo de transmision:
						double datarate = (double) msjdatos->getByteLength()*8/control->txrate; // CALCULAMOS EL PARAMETRO datarate QUE INDICA LA TASA DE TRANSMISION DEL PAQUETE INALAMBRICO
						simtime_t tasa = (simtime_t)datarate;
						/*
						double tiempo_slot = (double) check_and_cast<BS_queue *>(msjdatos->getSenderGate()->getOwnerModule())->t_slot;
						int ind_cola = (int) check_and_cast<BS_queue *>(msjdatos->getSenderGate()->getOwnerModule())->used_index;
						double offset = (double) tiempo_slot*ind_cola;
						offset_time = (simtime_t) offset;
						//ev << " 	Retardo de slot: " << offset_time << "s" << endl;
						control->slot_tx_delay[ind_cola] = control->slot_tx_delay[ind_cola] + tasa;
						//ev << " 	Retardo de transmision: " << control->slot_tx_delay[ind_cola] << "s" << endl;
						*/
						control->timepacket = control->timepacket + tasa;
						//ev << " 	Retardo de transmision: " << control->timepacket << "s" << endl;
						insert_Packet(msjdatos); // Envio del paquete al siguiente nodo, pero comprobando antes que cabe en la cola del modulo BS_queue[] en el que se tiene que almacenar (distincion entre los metodos de insercion centralizado y de prioridad de colas)
						//sendDelayed(msjdatos, offset_time + control->slot_tx_delay[ind_cola], "bsRxTx$o", encontrado);
					}
					// Si el mensaje viene de fuera, hacia alguno de los modulos BS_queue[] de esta BS:
					else if(strcmp(nombre_origen,this->getParentModule()->getFullName())!=0)
					{
						/* Control de admision segun el retardo estimado en el algoritmo DARA */
						int criterio = (int)this->getParentModule()->par("criterio");
						// SI SE USA EL ALGORITMO DARA:
						if(criterio == 1)
						{
							// Suma de los bits presentes en las colas "activas" de esta BS:
							long longitud_cola = 0;
							for(int i=0; i<control->used_queues; i++)
							{
								longitud_cola = longitud_cola + (long) check_and_cast<BS_queue *>(this->gate("queueRxTx$o",control->index_used_queues[i])->getNextGate()->getOwnerModule())->tamqueue[msjdatos->getPriority()];
							}
							double siguiente_ciclo = (double) fmod(SIMTIME_DBL(simTime()),control->t_frame);
							//ev << " Tiempo que falta hasta que llegue el siguiente ciclo de transmision: " << siguiente_ciclo << endl;
							double bits_previos = (double) 8*longitud_cola;
							//ev << " Bits presentes en las colas de esta BS: " << bits_previos << endl;
							bits_previos = bits_previos + (double) msjdatos->getBitLength();
							double bits_ciclo = (double) tx_rate*control->t_frame;
							////ev << " Bits por cada ciclo de transmision: " << bits_ciclo << endl;
							double ciclos_previos = (double) floor(bits_previos/bits_ciclo);
							////ev << " Numero de ciclos necesarios para enviar todos los bits de las colas (y el nuevo paquete): " << ciclos_previos << endl;
							double resto_ciclo = (double) fmod(bits_previos,bits_ciclo);
							////ev << " Bits que se transmitirian en el mismo ciclo que este paquete (este incluido): " << resto_ciclo << endl;
							double tiempo_previo = (double) ciclos_previos*control->t_frame + (double) resto_ciclo/tx_rate;
							//ev << " Tiempo que tardaria este paquete en llegar al siguiente nodo: " << tiempo_previo << endl;
							double retardo_paquete = SIMTIME_DBL(simTime() - msjdatos->getTimestamp());
							//ev << " Retardo actual del paquete: " << retardo_paquete << endl;
							double retardo_estimado = siguiente_ciclo + tiempo_previo; //+ retardo_paquete;
							//ev << " Retardo de transferencia estimado para este paquete: " << retardo_estimado << endl;

							/* Hasta aqui se ha calculado el retardo de enlace estimado para este paquete hasta el siguiente nodo
							 * Si se quiere estimar el retardo end-to-end, se hace lo siguiente:
							 * �OJO! Ahora si que se sumaria el retardo real previo del paquete */
							int aux_length_route = (int) msjdatos->getRutaArraySize();
							int intermediateBSs = aux_length_route - 2; // Los dos ultimos nodos son la BS pasarela y la ONU pasarela, que no cuentan en el encaminamiento por la red mesh.
							int aux_ind = 0;
							// Si en la ruta pasa por mas BSs intermedias aparte de esta misma, se calcula el retardo de enlace para cada una
							// (salvo ara esta, que ya se ha calculado ahora) y al final se suman todos los retardos de los enlaces para tener el end-to-end
							// (sumarle el retardo previo del paquete para que sea end-to-end de verdad).
							if(intermediateBSs > 1)
							{
								for(int i=0; i<intermediateBSs; i++)
								{
									if(strcmp((char*)msjdatos->getRuta(i),(char*)this->getParentModule()->getFullName())==0)
									{
										aux_ind = i+1;
									}
								}
								for(int i=aux_ind; i<intermediateBSs; i++)
								{
									BS_mac *aux_mac;
									aux_mac = check_and_cast<BS_mac *>(this->getParentModule()->getParentModule()->getSubmodule(msjdatos->getRuta(i))->getSubmodule("bs_mac"));
									longitud_cola = 0;
									siguiente_ciclo = 0;
									bits_previos = 0;
									bits_ciclo = 0;
									ciclos_previos = 0;
									resto_ciclo = 0;
									tiempo_previo = 0;
									for(int j=0; j<(int)aux_mac->used_queues; j++)
									{
										longitud_cola = longitud_cola + (long) check_and_cast<BS_queue *>(aux_mac->gate("bsmacInOut$o",aux_mac->index_used_queues[j])->getNextGate()->getOwnerModule())->tamqueue[msjdatos->getPriority()];
									}
									// Tiempo que faltara para el proximo ciclo de tx cuando este paquete llegue a este nodo:
									siguiente_ciclo = (double) fmod(SIMTIME_DBL(simTime())+retardo_estimado,aux_mac->t_frame);
									// Se estima que los paquetes que habra en las colas de esta BS seran los mismos que ahora,
									// entre los que se envien y se reciban en este tiempo:
									bits_previos = (double) 8*longitud_cola;
									//ev << " Bits presentes en las colas de esta BS: " << bits_previos << endl;
									bits_previos = bits_previos + (double) msjdatos->getBitLength();
									bits_ciclo = (double) tx_rate*aux_mac->t_frame;
									ciclos_previos = (double) floor(bits_previos/bits_ciclo);
									resto_ciclo = (double) fmod(bits_previos,bits_ciclo);
									tiempo_previo = (double) ciclos_previos*aux_mac->t_frame + (double) resto_ciclo/tx_rate;
									//ev << " Tiempo que tardaria este paquete en llegar al siguiente nodo: " << tiempo_previo << endl;
									retardo_estimado = siguiente_ciclo + tiempo_previo;
								}
							}
							// Una vez se ha estimado el retardo para todos los enlaces intermedios, se sume el retardo previo:
							retardo_estimado = retardo_estimado + retardo_paquete;
							//ev << " Retardo end-to-end estimado para este paquete:" << retardo_estimado << endl;
							// SI SE SUPERA EL UMBRAL DE RETARDO:
							if(retardo_estimado >= umbral_retardo)
							{
								//ev << " Mayor de 25 ms --> BORRADO!" << endl;
								paquetes_borrados_dara++;
								delete msjdatos;
							}
							// SI NO SE SUPERA EL UMBRAL DE RETARDO:
							else
							{
								//ev << " Menor de 25 ms --> SE INSERTA" << endl;
								/////// PRUEBA ///////
								paquetes_recibidos++;
								bytes_recibidos = bytes_recibidos + msjdatos->getByteLength();
								//ev << " Paquetes recibidos en esta BS: " << paquetes_recibidos << endl;
								/*
								// PARA PARAR LA SIMULACION CUANDO LLEGAN 100000 PAQUETES A CADA UNA DE LAS BSs DE LA RED:
								if(paquetes_recibidos >=100000 & aux_counter==0)
								{
									aux_counter=1;
									full_bs_queues = 1;
								}
								double total_paquetes = 0;
								for(int i=0; i<BSs; i++)
								{
									total_paquetes = total_paquetes + check_and_cast<BS_rx_tx *>(this->getParentModule()->getParentModule()->getSubmodule(control->nombres_bs[i])->getSubmodule("bs_rx_tx"))->full_bs_queues;
								}
								if(total_paquetes == BSs)
								{
									endSimulation(); // LLAMAMOS A LA FUNCION TERMINAR SIMULACION
									callFinish(); // LLAMAMOS A LA FUNCION FINALIZAR
								}
								*/
								//ev << " Recepcion de un paquete inalambrico, de prioridad P" << msjdatos->getPriority() << endl;
								simtime_t retardo = simTime() - msjdatos->getTimestamp();
								retardo_mensaje = SIMTIME_DBL(retardo);
								//ev << " Retardo total del mensaje procedente del nodo " << msjdatos->getOrigen() << ": " << retardo_mensaje << "s" << endl;
								//ev << " Mensaje recibido por la puerta " << msjdatos->getArrivalGate()->getFullPath() << endl;
								// Indice, en el grafo, del enlace por el que ha llegado el paquete:
								//int link = get_index_link(msjdatos->getArrivalGate()->getIndex());
								int link = msjdatos->getArrivalGate()->getIndex();
								//ev << " Correspondiente al enlace " << link << " del grafo" << endl;
								// Se cambia almacena el retardo del ultimo paquete para calcular el retardo medio del enlace por el que vino:
								retardo_medio[link].analyze(retardo_mensaje);
								arrived_packets[link]++;
								////ev << " Total paquetes recibidos: " << arrived_packets[link] << endl;
								arrived_bytes[link] = arrived_bytes[link] + msjdatos->getByteLength();

								// INICIO BUSQUEDA INDICE BS_QUEUE[]:
								// Vamos a ver la ruta en el mensaje:
								int h=0, siguiente = -1;
								while(h<(int)msjdatos->getRutaArraySize() && siguiente<0)
								{
									char* nodo_ruta = (char*) msjdatos->getRuta(h);
									if(strcmp(msjdatos->getDestino(),nodo_ruta)==0)
									{
										siguiente = h+1;
									}
									h++;
								}
								char* nodo_siguiente = (char*) msjdatos->getRuta(siguiente);
								// Si el 'target' de la cola coincide con el nombre del nodo destino, ahi va:
								int long_puerta = (int) this->gateSize("queueRxTx"); // Tantas como modulos BS_queue
								int j=0, found = -1;
								while(j<long_puerta && found<0)
								{
									cola_bs = check_and_cast<BS_queue *>(this->gate("queueRxTx$o",j)->getNextGate()->getOwnerModule()); // ENTRAMOS Y CHEQUEAMOS EL MODULO BS_QUEUE
									if(strcmp(nodo_siguiente,cola_bs->getTarget())==0)
									{
										found = j;
									}
									j++;
								}
								// FIN BUSQUEDA INDICE BS_QUEUE[]

								// --> Envio del paquete al modulo BS_queue[] que toca:
								send(msjdatos,"queueRxTx$o",found);
								//ev << " Mensaje entregado a la cola " << cola_bs->getFullName() << ", para luego enviarlo al nodo " << cola_bs->getTarget() << endl;
							}
							/////// PRUEBA ///////
					}
					// SI NO SE USA EL ALGORITMO DARA:
					else
					{
						paquetes_recibidos++;
						bytes_recibidos = bytes_recibidos + msjdatos->getByteLength();
						//ev << " Paquetes recibidos en esta BS: " << paquetes_recibidos << endl;
						/*
						// PARA PARAR LA SIMULACION CUANDO LLEGAN 100000 PAQUETES A CADA UNA DE LAS BSs DE LA RED:
						if(paquetes_recibidos >=100000 & aux_counter==0)
						{
							aux_counter=1;
							full_bs_queues = 1;
						}
						double total_paquetes = 0;
						for(int i=0; i<BSs; i++)
						{
							total_paquetes = total_paquetes + check_and_cast<BS_rx_tx *>(this->getParentModule()->getParentModule()->getSubmodule(control->nombres_bs[i])->getSubmodule("bs_rx_tx"))->full_bs_queues;
						}
						if(total_paquetes == BSs)
						{
							endSimulation(); // LLAMAMOS A LA FUNCION TERMINAR SIMULACION
							callFinish(); // LLAMAMOS A LA FUNCION FINALIZAR
						}
						*/
						//ev << " Recepcion de un paquete inalambrico, de prioridad P" << msjdatos->getPriority() << endl;
						simtime_t retardo = simTime() - msjdatos->getTimestamp();
						retardo_mensaje = SIMTIME_DBL(retardo);
						//ev << " Retardo total del mensaje procedente del nodo " << msjdatos->getOrigen() << ": " << retardo_mensaje << "s" << endl;
						//ev << " Mensaje recibido por la puerta " << msjdatos->getArrivalGate()->getFullPath() << endl;
						// Indice, en el grafo, del enlace por el que ha llegado el paquete:
						//int link = get_index_link(msjdatos->getArrivalGate()->getIndex());
						int link = msjdatos->getArrivalGate()->getIndex();
						//ev << " Correspondiente al enlace " << link << " del grafo" << endl;
						// Se cambia almacena el retardo del ultimo paquete para calcular el retardo medio del enlace por el que vino:
						retardo_medio[link].analyze(retardo_mensaje);
						arrived_packets[link]++;
						////ev << " Total paquetes recibidos: " << arrived_packets[link] << endl;
						arrived_bytes[link] = arrived_bytes[link] + msjdatos->getByteLength();

						// INICIO BUSQUEDA INDICE BS_QUEUE:
						// Vamos a ver la ruta en el mensaje:
						int h=0, siguiente = -1;
						while(h<(int)msjdatos->getRutaArraySize() && siguiente<0)
						{
							char* nodo_ruta = (char*) msjdatos->getRuta(h);
							if(strcmp(msjdatos->getDestino(),nodo_ruta)==0)
							{
								siguiente = h+1;
							}
							h++;
						}
						char* nodo_siguiente = (char*) msjdatos->getRuta(siguiente);
						// Si el 'target' de la cola coincide con el nombre del nodo destino, ahi va:
						int long_puerta = (int) this->gateSize("queueRxTx"); // Tantas como modulos BS_queue
						int j=0, found = -1;
						while(j<long_puerta && found<0)
						{
							cola_bs = check_and_cast<BS_queue *>(this->gate("queueRxTx$o",j)->getNextGate()->getOwnerModule()); // ENTRAMOS Y CHEQUEAMOS EL MODULO BS_QUEUE
							if(strcmp(nodo_siguiente,cola_bs->getTarget())==0)
							{
								found = j;
							}
							j++;
						}
						// FIN BUSQUEDA INDICE BS_QUEUE[]

						// --> Envio del paquete al modulo BS_queue[] que toca:
						send(msjdatos,"queueRxTx$o",found);
						//ev << " Mensaje entregado a la cola " << cola_bs->getFullName() << ", para luego enviarlo al nodo " << cola_bs->getTarget() << endl;
					}
				}
			}
		}
			break;
		default:
			delete msg;
			break;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION FINISH()--> ESTA FUNCION SE INVOCA CUANDO LA SIMULACION HA TERMINADO CON EXITO SIN QUE SE PRODUZCA NINGUN ERROR. 		//
//					  LO USAMOS PARA LA RECOGIDA DE ESTADASTICAS Y VISUALIZACION POR PANTALLA O MEDIANTE UN ARCHIVO.            //
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void BS_rx_tx::finish()
{
	// Asignacion a la variable total_enlaces de su valor real (el que se ha calculado en la ultima BS en la inicializacion)
	total_enlaces = check_and_cast<BS_rx_tx *>(this->getParentModule()->getParentModule()->getSubmodule(control->nombres_bs[BSs - 1])->getSubmodule("bs_rx_tx"))->total_enlaces;
	//total_enlaces = check_and_cast<BS_rx_tx *>(this->getParentModule()->getParentModule()->getSubmodule("bs",BSs - 1)->getSubmodule("bs_rx_tx"))->total_enlaces; /////// CAMBIO_RED ///////

	//ev << " Total enlaces entre nodos de la red mallada: " << total_enlaces << endl;
	double total_average_delay = calculate_average_transfer_delay();

	// 'RETARDO_NODOS_WIRELESS': FICHERO CON EL RETARDO MEDIO EN CADA ENLACE SALIENTE DE CADA BS DE LA SUBRED WIRELESS.
	retardo_intermedio=fopen("results/retardo_nodos_wireless.txt", "a+");
	if(esPasarela==1)
	{
		fprintf(retardo_intermedio,"BS[%i] (pasarela)\n", indice_esta_bs);
	}
	else
	{
		fprintf(retardo_intermedio,"BS[%i] (NO pasarela)\n", indice_esta_bs);
	}
	for (int i=0; i<links; i++)
	{
		fprintf(retardo_intermedio, "\t Retardo del enlace wireless %i: %g\n", i, (double)retardo_medio[i].average());
	}
	if(indice_esta_bs == BSs-1)
	{
		fprintf(retardo_intermedio, " TOTAL RETARDO MEDIO DE TRANSFERENCIA EN LOS ENLACES INTERMEDIOS DE LA SUBRED WIRELESS: %g\n", total_average_delay);
	}
	fclose(retardo_intermedio);

	// 'PAQUETES_RECIBIDOS_BS': FICHERO CON EL NUMERO TOTAL DE PAQUETES Y BYTES RECIBIDOS EN CADA BS.
	paquetes_por_bs = fopen("results/paquetes_recibidos_bs.txt", "a+");
	if(esPasarela==1)
	{
		fprintf(paquetes_por_bs,"BS[%i] (pasarela)\n", indice_esta_bs);
	}
	else
	{
		fprintf(paquetes_por_bs,"BS[%i] (NO pasarela)\n", indice_esta_bs);
	}
	int paqs = check_and_cast<BS_table *>(this->getParentModule()->getSubmodule("bs_table"))->paquetes_borrados;
	fprintf(paquetes_por_bs,"\t Total Paquetes:\t%g \t\t Total Bytes:\t%g \t\t Paquetes Borrados:\t%i\n",(double) paquetes_recibidos, (double) bytes_recibidos, paqs);
	fclose(paquetes_por_bs);

	// Llamada a la funcion que calcula la carga de cada enlace:
	calculate_link_load();
	// 'WIRELESS_LINK_LOAD': FICHERO CON EL VALOR MEDIO DE LA CARGA DE CADA ENLACE ENTRANTE Y SALIENTE EN CADA MODULO BS.
	wireless_link_loads=fopen("results/wireless_link_load.txt", "a+");
	if(esPasarela==1)
	{
		fprintf(wireless_link_loads,"BS[%i] (pasarela)\n", indice_esta_bs);
	}
	else
	{
		fprintf(wireless_link_loads,"BS[%i] (NO pasarela)\n", indice_esta_bs);
	}
	int aux_number_links = links - bs_links;
	for (int i=aux_number_links-1; i<links; i++)
	{
		fprintf(wireless_link_loads, "\t Enlace %i:\n",i);
		fprintf(wireless_link_loads, "\t\t Paquetes recibidos: %g\n",(double)arrived_packets[i]);
		fprintf(wireless_link_loads, "\t\t\t Carga normalizada a la capacidad efectiva: %g\n",(double)link_load[i]);
		fprintf(wireless_link_loads, "\t\t\t Carga normalizada a la tasa de transmision: %g\n",(double)alt_link_load[i]);
		fprintf(wireless_link_loads, "\t\t Paquetes enviados: %g\n",(double)sent_packets[i]);
		fprintf(wireless_link_loads, "\t\t\t Carga normalizada a la capacidad efectiva: %g\n",(double)node_load[i]);
		fprintf(wireless_link_loads, "\t\t\t Carga normalizada a la tasa de transmision: %g\n",(double)alt_node_load[i]);
	}
	if(indice_esta_bs == BSs-1)
	{
		fprintf(wireless_link_loads, " MEDIA CARGA RX ENLACES NORMALIZADA A CAP_EFF: %g\n", (double)mean_link_load);
		fprintf(wireless_link_loads, " MEDIA CARGA RX ENLACES NORMALIZADA A RX_RATE: %g\n", (double)mean_alt_link_load);
		fprintf(wireless_link_loads, " MEDIA CARGA TX ENLACES NORMALIZADA A CAP_EFF: %g\n", (double)mean_node_load);
		fprintf(wireless_link_loads, " MEDIA CARGA TX ENLACES NORMALIZADA A TX_RATE: %g\n", (double)mean_alt_node_load);
	}
	fclose(wireless_link_loads);

	if(esPasarela==1)
	{
		// 'CARGA_PASARELAS': FICHERO CON LA CARGA DE ENTRADA A CADA BS PASARELA.
		carga_pasarela = fopen("results/carga_pasarelas.txt", "a+");
		fprintf(carga_pasarela,"BS [%i] (pasarela)\n", indice_esta_bs);
		fprintf(carga_pasarela,"Carga total de llegadas: %g \n", (double)media_carga.average());
		fclose(carga_pasarela);

		// 'DATOS_RECIBIDOS_PASARELAS': FICHEROS CON EL TOTAL DE PAQUETES Y BYTES RECIBIDOS EN CADA BS PASARELA.
		datos_recibidos = fopen("results/datos_recibidos_pasarela.txt", "a+");
		fprintf(datos_recibidos,"BS[%i] (pasarela)\n", indice_esta_bs);
		fprintf(datos_recibidos,"\t Paquetes recibidos en total: %g\n",(double) paquetes_recibidos);
		fprintf(datos_recibidos,"\t Bytes recibidos en total: %g\n",(double) bytes_recibidos);
		fclose(datos_recibidos);

		// 'RETARDO_WIRELESS': RETARDO MEDIO DE TODOS LOS PAQUETES RECIBIDOS EN CADA BS PASARELA, MAS EL RETARDO MEDIO DE TODOS LOS PAQUETES
		//					   QUE HAN LLEGADO EN TOTAL A TODAS LAS BSs PASARELA (DATO INSERTADO DESDE EL MODULO bs0.BS_MAC).
		retardo_wireless_pasarelas=fopen("results/retardo_wireless.txt", "a+");
		fprintf(retardo_wireless_pasarelas, "\tRetardo wireless BS[%i]: %g\n", indice_esta_bs, (double)retardo_pasarela.average());
		fclose(retardo_wireless_pasarelas);

		// 'SALTOS_METROS_BS_PASARELAS': FICHERO CON EL NUMERO MEDIO DE SALTOS Y DE LONGITUD RECORRIDOS POR CADA PAQUETE
		//								 WIRELESS RECIBIDO EN CADA BS PASARELA.
		saltos_metros = fopen("results/saltos_metros_bs_pasarelas.txt", "a+");
		fprintf(saltos_metros,"BS [%i]\n", indice_esta_bs);
		fprintf(saltos_metros,"\t Media saltos paquetes wireless: %g\n",(double) saltos.average());
		fprintf(saltos_metros,"\t Media longitud recorrida: %g\n",(double) longitud.average());
		fclose(saltos_metros);
	}

}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION INSERT_PACKET(CMESSAGE *MSG)--> INSERCION DE UN MENSAJE EN LA COLA DEL MODULO BS_QUEUE DE LA SIGUIENTE BS POR LA QUE HA DE PASAR		/
//										  EL PAQUETE, SEGUN EL METODO ELEGIDO EN EL ARCHIVO DE CONFIGURACION (COLAS SEPARADAS O CON PRIORIDAD)	/
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void BS_rx_tx::insert_Packet(cMessage *msg)
{
	DATA *msjdatos = check_and_cast<DATA*>(msg);
	int h=0, sig = -1;
	while(h<(int)msjdatos->getRutaArraySize() && sig<0)
	{
		char* nodo_ruta = (char*) msjdatos->getRuta(h);
		if(strcmp(msjdatos->getDestino(),nodo_ruta)==0){	sig = h+1;	}
		h++;
	}
	char* nodo_siguiente = (char*) msjdatos->getRuta(sig);
	int longitud_puerta = (int) this->gateSize("queueRxTx"); // Tantas como modulos BS_queue
	// Si el 'target' de la cola coincide con el nombre del nodo destino, ahi va:
	int j=0, found = -1;
	while(j<longitud_puerta && found<0)
	{
		cola_bs = check_and_cast<BS_queue *>(this->getParentModule()->getParentModule()->getSubmodule(msjdatos->getDestino())->getSubmodule("bs_queue",j));
		//cola_bs = check_and_cast<BS_queue *>(this->getParentModule()->getParentModule()->getSubmodule("bs",msjdatos->getDestAddress())->getSubmodule("bs_queue",j)); /////// CAMBIO_RED ///////
		if(strcmp(nodo_siguiente,cola_bs->getTarget())==0)
		{
			found = j;
		}
		j++;
	}
	double buffer = (double) this->getParentModule()->par("buffer_size"); // Longitud total del buffer (Bytes)
	////ev << " Buffer de cada BS: " << buffer << " Bytes." << endl;
	//buffer = buffer/cola_bs->colas_salida;///////
	////ev << " Buffer de cada 'BS_queue': " << buffer << " Bytes." << endl;

	long suma_total_user = 0;
	long tamsumqueue_user = 0;
	int metodo = (int) this->getParentModule()->par("user_insercionmethod_separatequeue0_priorityqueue1");
	switch(metodo)
	{
		case 0: // Metodo de insercion de paquetes en colas separadas
			// Comprobacion de si hay espacio suficiente o no en el buffer de esta cola:
			// Si hay espacio:
			if(buffer/prioridades >= (cola_bs->tamqueue[msjdatos->getPriority()] + msjdatos->getByteLength()))
			{
				// Envio del paquete "msjdatos" a la siguiente BS:
				sendDelayed(msjdatos, control->timepacket, "bsRxTx$o", encontrado);
				//int ind_cola = (int) check_and_cast<BS_queue *>(msjdatos->getSenderGate()->getOwnerModule())->used_index;
				//sendDelayed(msjdatos, offset_time + control->slot_tx_delay[ind_cola], "bsRxTx$o", encontrado);
			}
			// Si no hay espacio:
			else if(buffer/prioridades < (cola_bs->tamqueue[msjdatos->getPriority()] + msjdatos->getByteLength()))
			{
				// Borrado del mensaje "msjdatos" al no poderlo insertar en su cola por estar llena:
				//ev << " MENSAJE 'DATA' (Id." << msjdatos->getId() << ") BORRADO" << endl;
				delete msjdatos;
				// INCREMENTO DE LA VARIABLE PARA CONTROLAR EL BORRADO DE PAQUETES EN LA PARTE INALAMBRICA:
				check_and_cast<BS_table *>(this->getParentModule()->getSubmodule("bs_table"))->paquetes_borrados++;
			}
			break;
		case 1: // Metodo de insercion de paquetes en colas por prioridades
			// Suma de todas las subcolas del modulo BS_QUEUE correspondiente:
			for(int i=0; i<prioridades; i++)
			{
				// Los bytes en todas las cQueue (tantas como prioridades) se almacenan en 'suma_total_user':
				suma_total_user = suma_total_user + cola_bs->tamqueue[i];
			}
			// La variable 'tamsumqueue_user' almacena la suma total de los bytes de todos los paquetes de la bs_queue
			// y la longitud del paquete a enviar:
			tamsumqueue_user = suma_total_user + msjdatos->getByteLength();
			// Comprobacion de si hay espacio suficiente en el buffer total (todas las colas):
			// Si los bytes de todas las colas no superan la longitud del buffer total:
			if(buffer >= tamsumqueue_user)
			{
				// Envio del paquete "msjdatos" a la siguiente BS:
				//ev << "Mensaje enviado." << endl;
				sendDelayed(msjdatos, control->timepacket, "bsRxTx$o", encontrado);
				//int ind_cola = (int) check_and_cast<BS_queue *>(msjdatos->getSenderGate()->getOwnerModule())->used_index;
				//sendDelayed(msjdatos, offset_time + control->slot_tx_delay[ind_cola], "bsRxTx$o", encontrado);
			}
			// Si la suma de los bytes de todas las colas supera la longitud del buffer total:
			else if(buffer < tamsumqueue_user)
			{
				// Recorremos todas las colas, de menor a mayor prioridad, para borrar alguno de sus paquetes previos
				// y dejar sitio asi al nuevo paquete:
				int i = prioridades-1, fuera = -1;
				while(i>=0 && fuera < 0)
				{
					// Si el nuevo paquete es de la misma prioridad que la de la cola que se esta checkeando (en principio, la mas baja), se borra directamente:
					if(msjdatos->getPriority() == i)
					{
						//ev << " MENSAJE 'DATA' (Id." << msjdatos->getId() << ") BORRADO" << endl;
						// Resta de los bytes del paquete borrado a la suma total de todas las colas:
						tamsumqueue_user = tamsumqueue_user - msjdatos->getByteLength();
						delete msjdatos; // BORRADO DEL PAQUETE 'msjdatos'
						// INCREMENTO DE LA VARIABLE PARA CONTROLAR EL BORRADO DE PAQUETES EN LA PARTE INALAMBRICA:
						check_and_cast<BS_table *>(this->getParentModule()->getSubmodule("bs_table"))->paquetes_borrados++;
						// Salida del bucle que recorre las colas
						fuera = 1;
					}
					// Si el nuevo paquete es de mayor prioridad que la de la cola que se esta checkeando
					// Y esta cola NO esta vacia, se borran paquetes suyos hasta poder meter el nuevo:
					else if(msjdatos->getPriority() < i && cola_bs->queue[i].getLength()>0)
					{
						int salir = -1;
						// Recorremos todos los paquetes de esta cola, desde el mas antiguo:
						while (cola_bs->queue[i].getLength()>0 && salir < 0)
						{
							// Llamada a la funcion del modulo 'bs_queue' para borrar el primer paquete de la cola:
							cola_bs->deleteElement(msjdatos,i);
							// INCREMENTO DE LA VARIABLE PARA CONTROLAR EL BORRADO DE PAQUETES EN LA PARTE INALAMBRICA:
							check_and_cast<BS_table *>(this->getParentModule()->getSubmodule("bs_table"))->paquetes_borrados++;

							// Resta de los bytes del paquete extraido a la variable 'tamsumqueue_user':
							tamsumqueue_user = tamsumqueue_user - cola_bs->user_tamqueuepop[i];
							// Si hay espacio suficiente en cola para insertar el nuevo paquete...
							if(buffer >= tamsumqueue_user)
							{
								// Envio del mensaje a la siguiente BS:
								//ev << "Mensaje enviado." << endl;
								sendDelayed(msjdatos, control->timepacket, "bsRxTx$o", encontrado);
								//int ind_cola = (int) check_and_cast<BS_queue *>(msjdatos->getSenderGate()->getOwnerModule())->used_index;
								//sendDelayed(msjdatos, offset_time + control->slot_tx_delay[ind_cola], "bsRxTx$o", encontrado);

								// Salida del bucle que recorre los paquetes de esta cola
								salir = 1;
								fuera = 1;
							}
							// Si NO hay espacio suficiente en cola para insertar el nuevo paquete...
							else if(buffer < tamsumqueue_user)
							{
								////ev << " 	 Como es MAYOR que la longitud TOTAL del buffer (" << buffer
								//		<< " bytes) --> LIBERACION DE MAS ESPACIO EN LAS COLAS" << endl;
							}
							// Pasamos a la siguiente iteracion
						}
						if(buffer > tamsumqueue_user)
						{
							// Salida del bucle que recorre todas las colas
							fuera = 1;
						}
					}
					else if(msjdatos->getPriority() < i && cola_bs->queue[i].getLength()==0)
					{
						////ev << "  	No quedan paquetes en la cola P" << i << endl;
					}
					i--;
				}
			}
			break;
		default:
			delete msjdatos;
			break;
	}
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION DELAY_WEIGHT(INT INDICE1, INT INDICE2, CMESSAGE *MSG)--> CALCULA EL PESO DEL ENLACE 'INDICE' A PARTIR DE LOS PARAMETROS		/
//																			 ADJUNTADOS EN EL MENSAJE 'MSG' (LINK STATE ANNOUNCEMENT), Y		/
//																			 ACTUALIZA EL RETARDO DEL CANAL ASOCIADO A ESTE ENLACE CON EL PESO.	/
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void BS_rx_tx::calculate_delay_weight(int indice1, int indice2, cMessage *msg)
{
	// Para cada enlace i-->j, el peso se calcula con la siguiente formula:
		// Wi = 1/(mhu*Ci) + 1/(2*mhu*Ci) + rho_i/(mhu*Ci - lambda_est_i)
		// mhu = service_rate
		// Ci = Capacidad efectiva del enlace i (Capacidad_total/num_enlaces_activos)
		// lambda_i = anuncio->getArrival_rate(j)
		// rho_i = anuncio->getLink_load(j) = lambda_i/(mhu*Ci)
		// Estimated_Lambda_i(Tn) = [3*alpha/(3*alpha+1)]*[Lambda_i(Tn)+Estimated_Lambda(Tn-1)]*S^(1-alpha/1+alpha)
		// Carga < 0.3 -> alpha = 1 -> Estimated_Lambda_i(Tn) = (3/4)*[Lambda_i(Tn)+Estimated_Lambda(Tn-1)]
		// Carga >= 0.3 -> alpha->inf -> Estimated_Lambda_i(Tn) = [Lambda_i(Tn)+Estimated_Lambda(Tn-1)]*1/S
		// NOTA: En el paper del algoritmo DARA pone que "S" es el numero de muestras usadas para las predicciones, pero no dice NADA MAS.
		//			 Tras buscar info de la Media Movil Ponderada (WMA, Weighted Moving Average), que es lo que dice dicho informe que hace este algoritmo,
		//			 he acabado poniendo en la formula de la WMA:
		//			 S = Tiempo_simulacion_total/Periodo_entre_envios_LSAs = Numero de veces que se actualizan los pesoso de los enlaces a lo largo de toda la simulacion.
		// 			 Y adem�s, para que tenga mas peso la Lambda del ultimo periodo LSA, multiplico la ultima Lambda por el numero de veces que se ha entrado en este metodo, hasta el momento.


	LSA *anuncio = check_and_cast<LSA*>(msg);
	// indice1 --> Indice del enlace en el array del nodo remoto
	// indice2 --> Indice del enlace en el array de este nodo

	//cTopology::Node *rem_node = top.getNodeFor(this->getParentModule()->getParentModule()->getSubmodule(anuncio->getSrcName()));
	//cTopology::Node *rem_node = top.getNodeFor(this->getParentModule()->getParentModule()->getSubmodule("bs",anuncio->getSrcAddress())); /////// CAMBIO_RED ///////

	//ev << " CALCULO DEL PESO DEL ENLACE DE " << this->getParentModule()->getFullName() << " A " << anuncio->getSrcName() << " + ACTUALIZACION DEL RETARDO DEL CANAL" << endl;
		////ev << " Modulo emisor del LSA: " << anuncio->getSenderModule()->getParentModule()->getFullName() << " = Nodo origen del LSA: " << anuncio->getSrcName() << endl;
	//double paquetes = anuncio->getRcvd_pckts(indice1);
	//ev << " 	-> Paquetes enviados desde este nodo al nodo " << anuncio->getSrcName() << ": " << paquetes << endl; //ev-comentar
	double lambda = anuncio->getArrival_rate(indice1);
	//ev << " 	-> Tasa de llegadas a " << anuncio->getSrcName() << ": " << lambda << "pqt/s" << endl; //ev-comentar
	//ev << " 		--> Tasa de llegadas estimada en el periodo anterior: " << previous_estimated_lambda[indice2] << "pqt/s" << endl; //ev-comentar
	//double diferencia = lambda - previous_estimated_lambda[indice2];
	//ev << " 		--> Diferencia con la tasa de llegadas estimada previamente: " << diferencia << endl; //ev-comentar
	double rho = anuncio->getLink_load(indice1); // rho = lambda/(service_rate*capacity);
	//ev << " 	-> Carga: " << rho << endl; //ev-comentar
	//double capacity = anuncio->getEffective_link_capacity();
	double capacity = control->eff_cap; // eff_cap; //
	//ev << " 	-> Capacidad efectiva: " << capacity << "bps" << endl;
	double tx_delay, slot_sync_delay, queueing_delay, weight;
	// Suma de la Lambda reportada en el LSA del nodo vecino y la Lambda estimada para este enlace en el periodo anterior
	// (se pondera lambda*aux_samples[indice2] para que pese mas el valor real obtenido en el ultimo LSA):
	aux_samples[indice2]++;
	//double sum_real_lambda_and_prev_estim_lambda = aux_samples[indice2]*lambda + previous_estimated_lambda[indice2];
	double sum_real_lambda_and_prev_estim_lambda = lambda + previous_estimated_lambda[indice2];

	/*
	double aux_sum = 0;
	for(int i=1; i<=aux_samples[indice2]; i++)
	{
		aux_sum += i;
	}
	//ev << " Numero de veces que se ha ejecutado el algoritmo DARA: " << aux_samples[indice2] << endl;
	//ev << " Sumatorio: " << aux_sum << endl;
	*/
	double carga_nodo = (double) this->getParentModule()->par("carga_nodo");
	if(carga_nodo < 0.6) // Para cargas bajas-moderadas (se puede poner 0.5)...
	{
		double three_quarters = (double) 3/4;
		estimated_lambda = (double)three_quarters*sum_real_lambda_and_prev_estim_lambda;
	}
	else // Para el resto...
	{
		double T = 200;
		//double inv_paquetes = (double) 1/aux_samples[indice2];
		//estimated_lambda = (double)inv_paquetes*sum_real_lambda_and_prev_estim_lambda;
		//estimated_lambda = (double)sum_real_lambda_and_prev_estim_lambda*aux_samples[indice2]/aux_sum;
		estimated_lambda = (double)sum_real_lambda_and_prev_estim_lambda*SIMTIME_DBL(LSA_period)/T;
	}
	//ev << " 		--> Tasa de llegadas estimada para el siguiente periodo: " << estimated_lambda << "pqt/s" << endl; //ev-comentar
	tx_delay = (double) 1/(service_rate*capacity);
	//ev << " 		--> Retardo de transmision: " << tx_delay << "s" << endl; //ev-comentar
	slot_sync_delay = (double) tx_delay/2;
	//ev << " 		--> Retardo de sincronizacion de slot: " << slot_sync_delay << "s" << endl; //ev-comentar
	queueing_delay = (double) rho/(service_rate*capacity - estimated_lambda);
	//queueing_delay = (double) rho/(service_rate*capacity - lambda); // SIN LA WMA
	//ev << " 		--> Retardo de encolamiento: " << queueing_delay << "s" << endl; //ev-comentar
	//double prev_peso = (check_and_cast<cDelayChannel*>(nodeOut->getLinkOut(indice2)->getLocalGate()->getChannel()))->par(3);
	//ev << " 	* Antiguo peso del canal: " << prev_peso << endl; //ev-comentar
	// CALCULO DEL NUEVO PESO DEL ENLACE:
	weight = (double) (tx_delay + slot_sync_delay + queueing_delay);
	weight = (double) fabs(weight);
	// NOTA: Si se quiere tener en cuenta tambien el retardo de propagacion (que, aunque sea peque�o, es lo que
	//		 diferencia a los enlaces no usados, pues mhu y Ci es comun a todos los del mismo nodo), se haria
	//		 lo siguiente:
		//double length = (check_and_cast<cDelayChannel*>(rem_node->getLinkOut(indice1)->getLocalGate()->getChannel()))->par(2);
		////ev << " 		--> Longitud del canal: " << length << "m" << endl;
		//double propagation_delay = (double) (check_and_cast<cDelayChannel*>(nodeOut->getLinkOut(indice2)->getLocalGate()->getChannel()))->par(4);
		////ev << " 		--> Retardo de propagacion: " << propagation_delay << "s" << endl;
		//weight = (double)(weight + propagation_delay);

	// ACTUALIZACION DEL RETARDO DEL CANAL EN EL SENTIDO BS_rx_LSA --> BS_tx_LSA:
	// NOTA: Enlace BS_rx_LSA --> BS_tx_LSA <==> Enlace BS_tx_paquetes --> BS_rx_paquetes
	(check_and_cast<cDelayChannel*>(nodeOut->getLinkOut(indice2)->getLocalGate()->getChannel()))->par(3) = (double)weight;
	// �OJO! El enlace puede tener pesos asimetricos (distinto peso en un sentido que en otro), luego
	// 		 �no se cambia en el sentido BS_tx_LSA --> BS_rx_LSA!
	//(check_and_cast<cDelayChannel*>(rem_node->getLinkOut(indice1)->getLocalGate()->getChannel()))->par(3) = (double)weight;

	//double retardo = (check_and_cast<cDelayChannel*>(nodeOut->getLinkOut(indice2)->getLocalGate()->getChannel()))->par(3);
	//ev << " 	* Nuevo peso del canal: " << retardo << endl; //ev-comentar

	//double retardo_2 = (check_and_cast<cDelayChannel*>(rem_node->getLinkOut(indice1)->getLocalGate()->getChannel()))->par(3);
	//ev << " 		--> Peso del canal en el otro sentido: " << retardo_2 << endl;

	previous_estimated_lambda[indice2] = estimated_lambda;
	delete anuncio;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION GET_INDEX_LINK(INT INDEX)--> ACTUALIZA EL RETARDO DE CADA ENLACE DE LA RED, SEGUN EL VALOR PROMEDIO	/
//							 		   DEL MISMO ENTRE TODOS LOS PAQUETES QUE LLEGAN POR EL (PESO). 			/
//									   DEVUELVE EL INDICE DEL ENLACE EN EL GRAFO DE LA CLASE cTopology.		    /
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int BS_rx_tx::get_index_link(int index)
{
	int indice=-1;
	// Se habilitan como enlaces los asociados a puertas de salida de la BS que estan conectadas:
	for(int i=0; i<links; i++)
	{
		//ev << " Indice enlace: " << nodeOut->getLinkOut(i)->getLocalGate()->getIndex() << endl;
		//ev << " Indice puerta: " << this->gate("bsRxTx$o",i)->getIndex() << endl;
		if(nodeOut->getLinkOut(i)->getLocalGate()->getIndex()==index)
		{
			indice = i;
		}
	}
	return indice;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION CALCULATE_WEIGHT_PARAMS(INT IND)--> CALCULA LOS PARAMETROS QUE SE ENVIAN EN EL LSA, NECESARIOS PARA CALCULAR	/
//							 		          EL RETARDO DE TRANSMISION (PESO DEL ENLACE).								/
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void BS_rx_tx::calculate_weight_params(int ind)
{
	////ev << " CALCULO PARAMETROS PARA ENVIAR EN EL LSA:" << endl;
	// packet_intensity -> Tasa de llegadas de paquetes (lambda) -> [paquetes/segundos]
	// service_rate -> Tasa de servicio del enlace (mhu) -> [1/bits]
	// link_load -> Carga del enlace (rho) -> [u.n.]
	// effective_capacity -> Capacidad efectiva del enlace (Capacidad/num_enlaces) -> [bits/segundo]
		// En "User_traffic":
		// --> lambda = tx_rate/(packetSize_user*8); --> Paquetes/segundo (teorico)
		// --> total_load_user = (double) (suma_total_bytes * 8/(simTime()*tx_rate)) --> Adimensional (practico)

	// Calculo lambda_i (tasa de llegadas del enlace i):

	packet_intensity[ind] = arrived_packets[ind]/SIMTIME_DBL(LSA_period); //ACTUAL
	//packet_intensity[ind] = arrived_packets[ind]/SIMTIME_DBL(simTime());
	//ev << " 	-> Tasa de llegadas de paquetes del enlace " << ind << ": " << packet_intensity[ind] << " paq/seg" << endl;
	////ev << " Enlace que conecta con " << nodeOut->getLinkOut(ind)->getRemoteNode()->getModule()->getFullName() << endl;

	double aux_eff_cap = check_and_cast<BS_mac *>(nodeOut->getLinkOut(ind)->getRemoteNode()->getModule()->getSubmodule("bs_mac"))->eff_cap; // eff_cap; //

	// Calculo rho_i (carga o utilizacion del enlace i):
	link_load[ind] = packet_intensity[ind]/(service_rate*aux_eff_cap); // rho_i
	//link_load[ind] = packet_intensity[ind]/(service_rate*effective_capacity); // rho_i
	//ev << " 	-> Uso del enlace " << ind << ": " << link_load[ind] << endl;
	//double prueba = arrived_bytes[ind]*8/(SIMTIME_DBL(simTime())*effective_capacity);
		//double prueba = arrived_bytes[ind]*8/(SIMTIME_DBL(simTime())*tx_rate);
	////ev << " Carga del enlace " << ind << ": " << prueba << endl;
		// NOTA: He comprobado que "prueba = packet_intensity[ind]", pero es que son la misma f�rmula.
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION CALCULATE_LINK_LOAD()--> CALCULA LA CARGA DE ENVIO/RECEPCION DE PAQUETES DE CADA ENLACE, NORMALIZADAS TANTO	/
//								   A LA CAPACIDAD EFECTIVA DEL CANAL COMO A LA CAPACIDAD TOTAL DE LA RED.				/
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void BS_rx_tx::calculate_link_load()
{
	int aux_number_links = links - bs_links;
	for(int i=aux_number_links; i<links; i++)
	{
		// C_i = Tasa de transmision de radio entre el numero total de enlaces salientes de cada nodo, conectando con otras BSs
		// txrate = Tasa de transmision de radio (suele coincidir con eff_cap, que la divide entre el numero de enlaces salientes activos)
		double aux_eff_cap = check_and_cast<BS_mac *>(nodeOut->getLinkOut(i)->getRemoteNode()->getModule()->getSubmodule("bs_mac"))->C_i;
		double aux_txrate = check_and_cast<BS_mac *>(nodeOut->getLinkOut(i)->getRemoteNode()->getModule()->getSubmodule("bs_mac"))->txrate;
		link_load[i] = (double) (arrived_bytes[i]*8)/(SIMTIME_DBL(simTime())*aux_eff_cap);
		alt_link_load[i] = (double) (arrived_bytes[i]*8)/(SIMTIME_DBL(simTime())*aux_txrate);
		node_load[i] = (double) (sent_bytes[i]*8)/(SIMTIME_DBL(simTime())*control->C_i);
		alt_node_load[i] = (double) (sent_bytes[i]*8)/(SIMTIME_DBL(simTime())*control->txrate);
	}
	for(int j=aux_number_links; j<links; j++)
	{
		mean_link_load = mean_link_load + link_load[j];
		mean_alt_link_load = mean_alt_link_load + alt_link_load[j];
		mean_node_load = mean_node_load + node_load[j];
		mean_alt_node_load = mean_alt_node_load + alt_node_load[j];
	}
	mean_link_load = (double)(mean_link_load/total_enlaces);
	mean_alt_link_load = (double)(mean_alt_link_load/total_enlaces);
	mean_node_load = (double)(mean_node_load/total_enlaces);
	mean_alt_node_load = (double)(mean_alt_node_load/total_enlaces);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION CALCULATE_AVERAGE_TRANSFER_DELAY()--> CALCULA EL RETARDO MEDIO DE TRANSFERENCIA EN LOS ENLACES ENTRE BSs DE TODA LA PARTE			/
//												INALAMBRICA Y DEVUELVE ESTE VALOR PARA REGISTRARLO EN EL ARCHIVO "RETARDO_NODOS_WIRELESS".	/
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
double BS_rx_tx::calculate_average_transfer_delay()
{
	double total_retardo = 0, media = 0;
	int total_enlaces_bs = 0;
	for(int i=0; i<BSs; i++)
	{
		cTopology::Node *aux_nodo = top.getNodeFor(this->getParentModule()->getParentModule()->getSubmodule(control->nombres_bs[i]));
		total_enlaces_bs = check_and_cast<BS_rx_tx *>(this->getParentModule()->getParentModule()->getSubmodule(control->nombres_bs[i])->getSubmodule("bs_rx_tx"))->links;
		//cTopology::Node *aux_nodo = top.getNodeFor(this->getParentModule()->getParentModule()->getSubmodule("bs",i)); /////// CAMBIO_RED ///////
		//total_enlaces_bs = check_and_cast<BS_rx_tx *>(this->getParentModule()->getParentModule()->getSubmodule("bs",i)->getSubmodule("bs_rx_tx"))->links; /////// CAMBIO_RED ///////

		for(int j=0; j<total_enlaces_bs; j++)
		{
			if (strpbrk(aux_nodo->getLinkOut(j)->getRemoteNode()->getModule()->getFullName(),"b") != NULL)
			{
				double retardo_auxiliar = (double)check_and_cast<BS_rx_tx *>(this->getParentModule()->getParentModule()->getSubmodule(control->nombres_bs[i])->getSubmodule("bs_rx_tx"))->retardo_medio[j].average();
				// retardo_auxiliar = (double)check_and_cast<BS_rx_tx *>(this->getParentModule()->getParentModule()->getSubmodule("bs",i)->getSubmodule("bs_rx_tx"))->retardo_medio[j].average(); /////// CAMBIO_RED ///////
				total_retardo = total_retardo + retardo_auxiliar;
				/////// PARA HACER LA MEDIA CONTABILIZANDO SOLO LOS ENLACES "ACTIVOS"
				if(retardo_auxiliar == 0.0)
				{
					total_enlaces = total_enlaces - 1;
				}
				///////
			}
		}
	}
	//ev << " Total enlaces activos entre BSs de la red: " << total_enlaces << endl;
	media = (double) (total_retardo/total_enlaces);
	////ev << " Total enlaces entre BSs de la red mallada: " << total_enlaces << endl;
	////ev << " Total retardo medio de todos estos enlaces: " << total_retardo << " s." << endl;
	////ev << " MEDIA: " << media << " s." << endl;
	return media;
}
