////////////////////////////////////////////////////////////////////////////////////
// PROYECTO FIN DE CARRERA															  											 	//
// TITULO: Disenyo de nuevas arquitecturas y algoritmos de gestin de recursos en	//
//		  	 redes de acceso FiWi  												  												//
// AUTOR: Ana Emma Lopez Mato																											//
// TUTOR: Noemi Merayo Alvarez																										//
// INGENIERIA TECNICA DE TELECOMUNICACIONES, SISTEMAS DE TELECOMUNICACION					//
// UNIVERSIDAD DE VALLADOLID													      											//
////////////////////////////////////////////////////////////////////////////////////

#include "BS_table.h"
/*
 * MODULO BS_table:
 * Contiene las diferentes tablas empleadas por el modulo BS para realizar el
 * encaminamiento de los paquetes:
 * array_de_rutas: Tabla con el nombre de los nodos que forman todas las rutas posibles.
 * array_de_longitudes: Tabla con la longitud (metros) de todas las rutas posibles.
 * array_de_retardos: Tabla con el retardo (segundos) de todas las rutas posibles.
 * rutas_optimas: Tabla con la ruta optima para cada usuario inalambrico.
 */
Define_Module(BS_table);

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION INITIALIZE()--> ESTA FUNCION SE INVOCA DESPUES DE QUE OMNET++ HA PUESTO EN MARCHA LA RED, EN LA CUAL SE LEEN LOS      //
//						  PARIMETROS DEL MODULO Y SE INICIALIZAN TODAS DAS LAS VARIABLES DECLARADAS PARA ESTE MODULO SIMPLE, SE //
//						  ASIGNAN ESTRUCTURAS DE DATOS DINIMICOS Y SE ASIGNAN E INICIALIZAN LOS AUTOMENSAJES SI SON NECESARIOS  //
//						  PARA EL FUNCIONAMIENTO DE ESTE MODULO.			   													//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void BS_table::initialize()
{
		paquetes_borrados = 0; // Total paquetes borrados por esta BS
	//// INICIALIZACION TABLA CON LAS RUTAS MAS CORTAS (BS_table.cc) ////
		int usuarios = (int) this->getParentModule()->par("numUser");
		int num_pasarelas = (int) this->getParentModule()->par("numGW");
		int connections = (int) this->getParentModule()->par("conexiones");
		int estaciones = (int) this->getParentModule()->par("numBS");
		filas = estaciones*connections*num_pasarelas*usuarios*2;
		columnas = 2 + estaciones;
		array_de_longitudes.resize(filas);
		array_de_retardos.resize(filas);
		array_de_pesos.resize(filas);
		array_de_rutas.resize(filas);
		saltos_total_rutas.resize(filas,0);
		longitud_total_rutas.resize(filas,0); // Vector con la longitud de User a ONU de cada ruta
		retardo_total_rutas.resize(filas,0.0); // Vector con el retardo de propagacion de User a ONU de cada ruta
		peso_total_rutas.resize(filas,0.0);
		rutas_optimas.resize(usuarios);
		// CREAMOS LAS 4 TABLAS GENERALES MEDIANTE EL BUCLE FOR
		for(int m=0;m<filas;m++)
		{
			// Tabla con las distancias de los enlaces de todos los caminos posibles
			array_de_longitudes[m].resize(columnas,0);
			// Tabla con los retardos de los enlaces de todos los caminos posibles
			array_de_retardos[m].resize(columnas,0.0);
			// Tabla con los pesos de los enlaces de todos los caminos posibles:
			array_de_pesos[m].resize(columnas,0.0);
			// Tabla con los nombres de los nodos que forman todos los caminos posibles
			array_de_rutas[m].resize(columnas," ");
		}
		// CREAMOS LA TABLAS CON LA RUTA OPTIMA PARA CADA USUARIO MEDIANTE EL BUCLE FOR
		for(int n=0;n<usuarios;n++)
		{
			rutas_optimas[n].resize(columnas," ");
		}
		rutas_optimas_pasarelas.resize(num_pasarelas);
		// CREAMOS LAS TABLAS CON LOS INDICES DE LAS POSIBLES RUTAS PARA CADA USUARIO, Y EL NUMERO DE RUTAS:
		indices_rutas.resize(usuarios);
		num_rutas.resize(usuarios);
		for(int m=0;m<usuarios;m++)
		{
			num_rutas[m].resize(num_pasarelas,0);
		}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION HANDLEMESSAGE(CMESSAGE *MSG)--> ESTA FUNCION SE INVOCA CON EL MENSAJE COMO PARAMETRO CADA VEZ QUE EL MODULO RECIBE UN //
//										  UN MENSAJE O PAQUETE. SE RECORRE EL CODIGO IMPLEMENTADO PARA DEVOLVER UN VALOR O  	//
//										  EJECUTAR UNA FUNCION DENTRO DEL MODULO SIMPLE. EL TIEMPO DE SIMULACION NO TRANSCURRE  //
//										  DENTRO DE LA LLAMADA DE LA FUNCION HANDLEMESSAGE(CMESSAGE *MSG) MIENTRAS RECORRE EL	//
//										  CODIGO IMPLEMENTADO.																	//                                                                         //
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void BS_table::handleMessage(cMessage *msg)
{
	/* NO SE UTILIZA */
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//FUNCION FINISH()--> ESTA FUNCION SE INVOCA CUANDO LA SIMULACION HA TERMINADO CON EXITO SIN QUE SE PRODUZCA NINGUN ERROR. 		//
//					  LO USAMOS PARA LA RECOGIDA DE ESTADASTICAS Y VISUALIZACION POR PANTALLA O MEDIANTE UN ARCHIVO.            //
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void BS_table::finish()
{
	/* NO SE UTILIZA */
}
